/*****************************************************************************/
#ifdef COMMENTS_WITH_COMMENTS
/*
                                  alamode.c

Browser based HTTPDMON-like (on steroids) WASD monitoring application.  WASD
has produced a Spanish sounding application (MonDeSi) so why not another with a
French flavour?  Although rather than meaning "fashionable" (cannot be said of
VMS or WASD), this application's name is intended to be reminscent of the North
American usage, "on the top or side".  Not being much of a dessert person I
might even prefer the cookery usage of "meats braised with vegetables in wine"
but that's too much of a stretch :-)  Whatever the force-fit it seemed less
prosaic than WASDmon.

This code is the data-gathering back-end.  The front-end is JavaScript driven. 
The front-end connects via WebSocket to this script and it provides
JSON formatted data back to the front-end for processing and display.

This application has been designed specifically against WASD v10.4.0 and later
but could be built on ealier systems with the attendant loss of some features
(e.g. server managed process count).

As this application does not lock the global section when reading values there
is the possibility of inconsistent resultant data.  Atomic reads are OK (e.g. a
longword or quadword (on non-VAX).  Data with relationships (i.e. peak
processing and current processing count) are an example because of the
processing occuring between the individual reads.  Another is the URI of the
latest request data (read character by character and therefore subject to
overwriting during the copy).  Rather than introduce the expense of a lock to
mitigate the risk, for an observational application like this it is considered
an acceptable tradeoff.

This application accesses the WASD global lock environment used to coordinate
and control WASD instances within a single system and across a cluster.  As a
consequnce of participating in the control lock environment directives sent to
the WASD envionment will report additional "instances" being notified.

  $ httpd/do=map
  %HTTPD-I-DO, 2 instances notified; KLAATU::WASD:80-70, KLAATU::WASD:80

The "instances" relating to alamode and those to servers should easily be
differentiated by process name.

The application is intended to communicate browser<->script using WebSocket but
can be used CGI employing XHR (XMLHttpRequest()) and a long-poll CGI response. 
Multiple JSON structures can be incorporated into (a single write and) a single
response stream by delimitting each with a sentinal sequence that cannot occur
naturally in a legal JSON structure.  This allows individual JSON structures to
be parsed from the stream.  The sequence """ (three consecutive quotation
characters) is used as the sentinal.  The same technique is used for long-poll
CGI-responses, and in WebSocket writes (to reduce the number of individual
I/Os).


MONITOR DISPLAY
---------------
A node is displayed as a text panel on the left and bar graphs on the right. 
These display the instantaneous (current) values for the various resources and
are updated dynamically at the interval when the data collector script detects
a change in values.  Each bar graph section contains a number representing the
current value and a number in the far right the maximum value reached while
monitoring.  All values are average per-second where applicable.   Important
element of the display have assocaited exaplanatory tooltips.


AUTHORISATION
-------------
This application provides substantial insight (at some level) into a site's
Web server profile and activity, as well as to the detail of some requests.
The application insists on authorised access and unless this is configured will
not permit access.

This can be a simple rule providing unrestricted access (where the username
becomes "WORLD")

  [WORLD]
  /cgi*-bin/alamode* read,https:

If the username is WORLD then authenticated requests have the username and
request URI display obfuscated (i.e. *********** out).

or when authentication is actually exercised

  ["A La Mode"=VMS]
  /cgi*-bin/alamode* read,https:

or further constrained by specified users

  ["A La Mode"=VMS]
  /cgi*-bin/alamode* ~USER1,~USER2,read,https:


CONFIGURATION LOGICALS
----------------------
Logical names defined /SYSTEM or in a table accessable to the scripting process.

ALAMODE_ALERT_403     integer HTTP 403 status per minute before alert
ALAMODE_ALERT_5NN     integer HTTP 500,501... status per minute before alert
ALAMODE_NI            network devices to be monitored (e.g. EWA1:,EWA2:)


PRIVILEGES
----------
SHARE       if other utilities (e.g. MONDESI) with channels allocated to NI
SYSPRV      access to global section (WASD accounting data)
SYSLCK      access WASD lock data (system/cluster IPC and coordination)
WORLD       access server process information (non-script account)

$ INSTALL ADD CGI-BIN:[000000]ALAMODE.EXE /PRIVILEGE=(SHARE,SYSPRV,SYSLCK,WORLD)


COPYRIGHT
---------
Copyright (C) 2014,2015 Mark G.Daniel
This program comes with ABSOLUTELY NO WARRANTY.
This is free software, and you are welcome to redistribute it under the
conditions of the GNU GENERAL PUBLIC LICENSE, version 3, or any later version.
http://www.gnu.org/licenses/gpl.txt


VERSION HISTORY
---------------
11-APR-2015  MGD  v1.0.1, add BUFIO and DIRIO to instance data
                          sys$getuai() to establish initial BYTLM (per JPP)
                          bugfix; proxy delta data into CollectAccounting()
                          bugfix; sys$gettim_prec defined IA64 V8.4 (per JPP)
                          Happy 62nd Wedding Anniversary Mum and Dad
29-MAR-2015  MGD  v1.0.0, initial release
01-FEB-2014  MGD  initial development
*/
#endif /* COMMENTS_WITH_COMMENTS */
/*****************************************************************************/

#define SOFTWAREVN "1.0.1"
/*                  ^^^^^ don't forget to update ALAMODE.JS compliance! */
#define SOFTWARENM "ALAMODE"
#define SOFTWARECR "Copyright (C) 2014,2015 Mark G.Daniel"
#ifdef __ALPHA
#  define SOFTWAREID SOFTWARENM " AXP-" SOFTWAREVN
#endif
#ifdef __ia64
#  define SOFTWAREID SOFTWARENM " IA64-" SOFTWAREVN
#endif
#ifdef __VAX
#  define SOFTWAREID SOFTWARENM " VAX-" SOFTWAREVN
#endif

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <netdb.h>

#include <cvtdef.h>
#include <descrip.h>
#include <dvidef.h>
#include <dvsdef.h>
#include <iodef.h>
#include <jpidef.h>
#include <lckdef.h>
#include <libdtdef.h>
#include <lnmdef.h>
#include <prvdef.h>
#include <secdef.h>
#include <ssdef.h>
#include <starlet.h>
#include <stsdef.h>
#include <syidef.h>
#include <uaidef.h>

#include "spidef.h"
#include "wslib.h"

#ifdef INCLUDE_WASD_H
#include INCLUDE_WASD_H
#else
#include "../httpd/wasd.h"
#endif

/* nomenclature rethink post 10.4.0 */
#ifdef WASDMON_SUPPORT_1040 
#define ALAMODE_SUPPORT_1040 
#endif

#ifndef ALAMODE_SUPPORT_1040
#error "minimum build version is WASD v10.4.0"
#endif

/** #define USE_INT64 0 **/
#include "simpler64.h"

#define VMSok(x) ((x) & STS$M_SUCCESS)
#define VMSnok(x) !(((x) & STS$M_SUCCESS))

#define FI_LI "ALAMODE", __LINE__
#define EXIT_FI_LI(status) \
{ printf ("Status:500\r\n\r\n[%s:%d]\r\n", FI_LI); exit(status); }

/* maximum number of network devices */
#define NI_DEVICE_MAX 8

/* puts three successive quotations into the stream - impossible JSON data */
#define JSON_SENTINAL "\"\"\""

/* number of these status codes per minute */
#define ALERT_STATUS_403 15
#define ALERT_STATUS_5NN 30

int  AlertStatusCode403,
     AlertStatusCodeGroup5,
     ConnectedCount,
     DataViaCGI,
     Debug,
     InstanceEnvNumber = 1,  /* default group number */
     IsCgiPlus,
     NodesAcross,
     RefreshInterval = 2,    /* second(s) */
     UsageCount,
     ViewWidth;

ulong  ProvideDataSeconds,
       ExitTimeStamp,
       SpiEfn,
       TimeStamp;

ulong  CollectBinTime [2],
       CurrentBinTime [2],
       ProvideBinTime [2],
       StartBinTime [2];
ushort  CurrentNumTime [7];

long NeedPrivMask [2] = { PRV$M_SHARE | PRV$M_SYSLCK |
                          PRV$M_SYSPRV | PRV$M_WORLD, 0 };

float  DeltaSeconds = 999999.0;

char  *CgiHttpRefererPtr,
      *CgiHttpUserAgentPtr,
      *CgiRemoteAddrPtr,
      *CgiRemoteUserPtr,
      *CgiRequestUriPtr,
      *CgiScriptNamePtr,
      *CgiQueryStringPtr,
      *CgiPlusEofPtr = NULL,
      *CgiPlusEotPtr = NULL,
      *CgiPlusEscPtr = NULL;

char  NetIntDev [NI_DEVICE_MAX*24],
      RefererHost [128];

/* sys$gettim_prec() arrived with IA64 V8.4 */
#if defined(sys$gettim_prec)
/* may also produce "%LINK-I-UDFSYM, GETTIM_PREC" which can safely be ignored */
int (*GetTimFunction)() = sys$gettim_prec;
#else
int (*GetTimFunction)() = sys$gettim;
#endif

/*******************/
/* data structures */
/*******************/

typedef struct STRUCT_IOSBLK
{
   ushort  iosb$w_status;
   ushort  iosb$w_bcnt;
   ulong  iosb$l_reserved;
}
IOSBLK;

#define IMAG_NAME_SIZE 127
#define UIC_STRING_MAX 63
#define JPI_USER_NAME_MAX 12
#define JPI_PRCNAM_MAX 15

typedef struct STRUCT_PROCESS
{
   ulong  JpiCpuId,
          JpiPrib,
          JpiPri,
          JpiWsExtent,
          JpiPpgCnt,
          JpiGpgCnt,
          JpiWsPeak,
          JpiPagFilCnt,
          JpiAstCnt,
          JpiAstLm,
          JpiBioCnt,
          JpiBioLm,
          JpiDioCnt,
          JpiDioLm,
          JpiEnqCnt,
          JpiEnqLm,
          JpiWsQuota,
          JpiPgFlQuota,
          JpiVirtPeak,
          JpiFilLm,
          JpiFilCnt,
          JpiBytLm,
          JpiBytCnt,
          JpiPrcLm,
          JpiPrcCnt,
          JpiTqLm,
          JpiTqCnt,
          JpiMultiThread,
          JpiKtCount,
          JpiPageFlts,
          JpiBufIo,
          JpiDirIo,
          JpiVolumes,
          PageFaults,
          ProcessId,
          PrevPageFaults;

   ulong  ConBinTime[2],
          CpuBinTime[2];

   char  *ProcessState;

   char  AstFlags [16],
         JpiPrcNam[JPI_PRCNAM_MAX+1];
}
PROCESS_DATA;


typedef struct STRUCT_CLIENT_DATA
{
   ulong  ClientCount,
          ProvideDetailData,
          ProvideInstanceData,
          ProvideProxyData,
          ProvideRequestData,
          ProvideScriptData,
          ProvideWebDavData,
          WorldAccess;

   char  InputBuffer [256],
         OutputBuffer [4096];

   struct WsLibStruct  *WsLibPtr;

} CLIENT_DATA;

typedef struct STRUCT_SYSTEM_DATA
{
   int  DetailBufferLength,
        InstanceClusterCount,
        InstanceBufferLength,
        InstanceNodeCount,
        InstanceStartupCount,
        PrevMinute,
        PrevInstanceStartupCount,
        ProxyBufferLength,
        RequestBufferLength,
        ScriptBufferLength,
        StaticBufferLength,
        RuntimeBufferLength,
        SummaryBufferLength,
        SystemDelta,
        UsageCount,
        WebDavBufferLength;

   ulong  ActiveCpuCount,
          AvailCpuCount,
          BufferedIO,
          CpuBusyPercent,
          CpuUserPercent,
          CpuTicksBusy,
          CpuTicksUser,
          ClusterMember,
          ClusterNodes,
          DirectIO,
          InstanceNextPid,
          JpiPid,
          FcpRead,
          FcpWrite,
          MemInUse,
          MemInUseMax,
          MemSize,
          PageSize,
          ProcessCount;

   ulong  BlockReadSize,
          BlockWriteSize,
          BytesPerSecAve,
          BytesPerSecMax,
          BytesPerSecMin,
          BytesRawPerSec,
          CurrentConnected,
          ConnectPeak,
          CurrentProcessing,
          CurrentScript,
          CurrentScriptCGIplus,
          CurrentScriptRTE,
          CurrentThrottled,
          CurrentThrottleBusyMetric,
          CurrentThrottleProcessing,
          CurrentThrottleQueued,
          CurrentWebSockets,
          DclProcessCurrent,
          CurrentDECnetCGI,
          CurrentDECnetOSU,
          CurrentDECnetTasks,
          MethodConnect,
          MethodCopy,
          MethodDelete,
          MethodGet,
          MethodHead,
          MethodLock,
          MethodMkCol,
          MethodMove,
          MethodOptions,
          MethodPost,
          MethodPropFind,
          MethodPropPatch,
          MethodPut,
          MethodSsh,
          MethodTrace,
          MethodUnLock,
          NetBytesPerSec,
          NetBytesRxPerSec,
          NetBytesTxPerSec,
          NetDatagramRx,
          NetDatagramTx,
          NetReadErr,
          NetWriteErr,
          PrevStatusCode403,
          PrevStatusCodeGroup5,
          ProcessingPeak,
          TotalRequest,
          SslCount,
          StatusCode403,
          TooBusy;

   ulong  DeltaDclCrePrc,
          DeltaDclDelPrc,
          PrevTotalDclHardLimit,
          TotalDclCrePrc,
          TotalDclCrePrcDetach,
          TotalDclCrePrcPersona,
          TotalDclCrePrcPersonaDefault,
          TotalDclCrePrcPersonaInvUser,
          TotalDclCrePrcPersonaPrvUser,
          TotalDclDelPrc,
          TotalDclForceX,
          TotalDclHardLimit,
          TotalDclProctor;

   ulong  DeltaAllScript,
          DeltaCgi,
          DeltaCgiPlus,
          DeltaCgiPlusReused,
          DeltaDECnet,
          DeltaDECnetReused,
          DeltaDECnetCgi,
          DeltaDECnetCgiReused,
          DeltaDECnetOsu,
          DeltaDECnetOsuReused,
          DeltaFile,
          DeltaFile304,
          DeltaOther,
          DeltaProxy,
          DeltaRequest,
          DeltaRequestFail,
          DeltaRte,
          DeltaRteReused,
          DeltaScriptCgi,
          DeltaThrottled,
          DeltaThrottleBusy,
          DeltaThrottleQueued,
          DeltaWebDavRead,
          DeltaWebDavTotal,
          DeltaWebDavWrite;

   ulong  TotalAdmin,
          TotalAllOther,
          TotalAllScript,
          TotalAutoScript,
          TotalCgi,
          TotalCgiPlus,
          TotalCgiPlusReused,
          TotalDclCommand,
          TotalDECnet,
          TotalDECnetCgi,
          TotalDECnetCgiReused,
          TotalDECnetReused,
          TotalDECnetOsu,
          TotalDECnetOsuReused,
          TotalDclScript,
          TotalDir,
          TotalFile,
          TotalFile304,
          TotalIsMap,
          TotalMenu,
          TotalNoModule,
          TotalProxy,
          TotalPut,
          TotalRte,
          TotalRteReused,
          TotalSsi,
          TotalThrottled,
          TotalThrottleBusy,
          TotalThrottleQueued,
          TotalUpdate,
          TotalWebDav,
          TotalWebDavLock,
          TotalWebDavOther,
          TotalWebDavRead,
          TotalWebDavWrite,
          TotalWebSocket;

   ulong  ProxyCacheEnabled,
          ProxyCacheErrorCount,
          ProxyCacheFreeBlocks,
          ProxyCacheSpace,
          ProxyCacheTotalBlocks,
          ProxyCacheRead,
          ProxyCacheRead304,
          ProxyCacheUsedBlocks,
          ProxyCacheWrite,
          ProxyDeltaCacheRead,
          ProxyDeltaCacheWrite,
          ProxyDeltaNetwork,
          ProxyDeltaTotal,
          ProxyEnabled,
          ProxyLookupCache,
          ProxyLookupDNS,
          ProxyLookupError,
          ProxyLookupLiteral,
          ProxyNetwork,
          ProxyRequestNoCache,
          ProxyResponseNoCache,
          ProxyTotal,
          ProxyTunnel,
          ProxyTunnelCurrent;

   ulong  WebDavAgentAppleCount,
          WebDavAgentBsdCount,
          WebDavAgentMicrosoftCount,
          WebDavAgentLinuxCount,
          WebDavAgentOtherCount,
          WebDavAgentSunCount,
          WebDavEnabled,
          WebDavLockingEnabled,
          WebDavMetaReadAttemptCount,
          WebDavMetaReadCount,
          WebDavMetaWriteAttemptCount,
          WebDavMetaWriteCount,
          WebDavQuotaEnabled,
          WebDavRequestCount,
          WebDavXmlParseCount;

   ulong  BootBinTime [2],
          BlocksRawRx [2],
          BlocksRawTx [2],
          BytesRawRx [2],
          BytesRawTx [2],
          InstanceNodePid [INSTANCE_MAX],
          ProxyBlocksRawRx [2],
          ProxyBlocksRawTx [2],
          ProxyBytesCacheRx [2],
          ProxyBytesCacheTx [2],
          ProxyBytesCacheRxTx [2],
          ProxyBytesNoCacheRx [2],
          ProxyBytesNoCacheTx [2],
          ProxyBytesNoCacheRxTx [2],
          ProxyBytesRawRx [2],
          ProxyBytesRawTx [2],
          ProxyBytesRawRxTx [2],
          ProxyBytesRxTx [2],
          StatusCodeGroup [6],
          UpBinTime [2],
          WebDavBytesRawRx [2],
          WebDavBytesRawTx [2];

   ulong  NetIntBlocksRx [2],
          NetIntBlocksTx [2],
          NetIntBlocksRxTx [2],
          NetIntBytesRx [2],
          NetIntBytesTx [2],
          NetIntBytesRxTx [2],
          NetIntErrorsHard [2],
          NetIntErrorsSoft [2];

   char  *ProxyCacheDeviceNamePtr;

   char  ArchName [15+1],
         DetailBuffer [2048],
         HttpdVersion [15+1],
         HwName [60+1],
         InstanceBuffer [4096],
         NodeName [15+1],
         ProxyBuffer [2048],
         RequestBuffer [2048],
         ScriptBuffer [2048],
         StaticBuffer [2048],
         RuntimeBuffer [256],
         SummaryBuffer [2048],
         VmsVersion [8+1],
         WebDavBuffer [2048];
}
SYSTEM_DATA;

SYSTEM_DATA  SystemData;

int  HttpdGblSecLength;
HTTPD_GBLSEC  *gblptr,
              *HttpdGblSecPtr;
ACCOUNTING_STRUCT  *accptr;
PROXY_ACCOUNTING_STRUCT  *pacptr;

#define INETACP$C_TRANS 2
#define INETACP_FUNC$C_GETHOSTBYNAME 1
#define INETACP_FUNC$C_GETHOSTBYADDR 2
$DESCRIPTOR (TcpIpDeviceDsc, "UCX$DEVICE");

/***********************/
/* function prototypes */
/***********************/

void AddClient ();
void CollectAccounting ();
void CollectInstanceData ();
void CollectNetIntData ();
int CollectProcessData (ulong, PROCESS_DATA*);
void CollectProxyData ();
void CollectScriptData ();
int CollectStaticData ();
int CollectSPI ();
void CollectWebDavData ();
int HavePriv (ulong*);
void JsonAlert (CLIENT_DATA*, char*, ...);
void JsonDo (CLIENT_DATA*, char*, ...);
void JsonDetailData ();
void JsonInstanceData ();
void JsonRequestData (CLIENT_DATA*, BOOL);
void JsonRuntime ();
void JsonScriptData ();
void JsonSummaryData ();
void JsonWebDavData ();
char* JsonNumber (int, ulong);
char* LookupHostName (ulong);
int MinimumWASD (char*);
char* MungeTime (ulong*);
void ProvideData ();
void ReadClient (struct WsLibStruct*);
void RemoveClient (struct WsLibStruct*);
void ReportInstanceControl (struct lksb*);
/* number 2 to avoid a clash in wasd.h */
char* SysTrnLnm2 (char*, char*, int);
int WebSockAvailable ();
void XhrResponse ();

#ifdef __VAX
#define NO_SNPRINTF 1
int snprintf (char*, __size_t, const char*, ...);
#endif

/*****************************************************************************/
/*
AST delivery is disabled during client acceptance and the add-client function
is deferred using an AST to help minimise the client setup window with a
potentially busy WebSocket application.
*/

main (int argc, char *argv[])

{
   int  status;
   char  *cptr, *sptr, *zptr;

   /*********/
   /* begin */
   /*********/

   if (argc > 1)
   {
      if (!strcasecmp (argv[1], "/VERSION"))
      {
         fprintf (stdout, "%%ALAMODE-I-VERSION, %s %s\n",
                  SOFTWAREID, WsLibVersion());
         exit (SS$_NORMAL);
      }
      if (!strcasecmp (argv[1], "/DBUG")) Debug = true;
   }

   /* don't want the C-RTL fiddling with the carriage control */
   if (!Debug) stdout = freopen ("SYS$OUTPUT", "w", stdout, "ctx=bin");

   if (!HavePriv (NeedPrivMask)) EXIT_FI_LI (SS$_NOPRIV);

#if !defined(_VAX) && !defined(__ALPHA)
   /* cater for "%LINK-I-UDFSYM, GETTIM_PREC" on unsupported platform */
   if (GetTimFunction == NULL) GetTimFunction = sys$gettim;
#endif

   GetTimFunction (&StartBinTime);

   WsLibInit ();

   /* no clients is fifteen minutes in seconds */
   WsLibSetLifeSecs (15*60);

   if (!SpiEfn)
      if (VMSnok (status = lib$get_ef (&SpiEfn)))
         EXIT_FI_LI (status);

   MapGlobalSection ();
   gblptr = HttpdGblSecPtr;
   accptr = &gblptr->Accounting;
   pacptr = &gblptr->ProxyAccounting;

   if (cptr = SysTrnLnm2 ("ALAMODE_ALERT_403", NULL, 0))
      AlertStatusCode403 = atoi(cptr);
   if (!AlertStatusCode403) AlertStatusCode403 = ALERT_STATUS_403;

   if (cptr = SysTrnLnm2 ("ALAMODE_ALERT_5NN", NULL, 0))
      AlertStatusCodeGroup5 = atoi(cptr);
   if (!AlertStatusCodeGroup5) AlertStatusCodeGroup5 = ALERT_STATUS_5NN;

   CollectStaticData ();
   CollectInstanceData ();
   ReportInstanceControl (NULL);

   JsonStaticData ();
   ProvideData ();

   if (Debug) for(;;) sys$hiber();

   IsCgiPlus = WsLibIsCgiPlus();

   for (;;)
   {
      /* with CGIplus this call will block waiting for the next request */
      WsLibCgiVar ("");

      /********************/
      /* request received */
      /********************/

      UsageCount++;

      if (!MinimumWASD ("10.4.0"))
      {
         fprintf (stdout, "Status: 500\r\n\r\nMinimum WASD v10.4.0!\n");
         exit (SS$_NORMAL);
      }

      CgiHttpRefererPtr = WsLibCgiVar("HTTP_REFERER");
      CgiHttpUserAgentPtr = WsLibCgiVar("HTTP_USER_AGENT");
      CgiQueryStringPtr = WsLibCgiVar("QUERY_STRING");
      CgiRemoteAddrPtr = WsLibCgiVar("REMOTE_ADDR");
      CgiRemoteUserPtr = WsLibCgiVar("REMOTE_USER");
      CgiRequestUriPtr = WsLibCgiVar("REQUEST_URI");
      CgiScriptNamePtr = WsLibCgiVar("SCRIPT_NAME");

      /* derive the URL scheme and host from the HTTP referrer field */
      zptr = (sptr = RefererHost) + sizeof(RefererHost)-1;
      cptr = CgiHttpRefererPtr;
      while (*cptr && *cptr != '/' && sptr < zptr) *sptr++ = *cptr++;
      if (*cptr == '/' && sptr < zptr) *sptr++ = *cptr++;
      if (*cptr == '/' && sptr < zptr) *sptr++ = *cptr++;
      while (*cptr && *cptr != '/' && sptr < zptr) *sptr++ = *cptr++;
      *sptr = '\0';

      if (!CgiRemoteUserPtr[0])
         fprintf (stdout, "Status: 500\r\n\r\nUnauthorised!\n");
      else
      if (MATCH6(CgiQueryStringPtr,"data=1"))
      {
         /********/
         /* data */
         /********/

         if (DataViaCGI = (WsLibCgiVarNull ("WEBSOCKET_INPUT") == NULL))
         {
            XhrResponse ();
            /* ProvideData() now drives it and should never ... */
            for(;;) sys$hiber();
         }

         if (!IsCgiPlus)
            fprintf (stdout, "Status: 500\r\n\r\nWebSocket must be CGIplus!\n");
         else
         if (!WebSockAvailable())
            fprintf (stdout, "Status: 500\r\n\r\nWebSocket unavailable!\n");
         else
            AddClient ();
      }
      else
      {
         /********/
         /* page */
         /********/

         if (WebSockAvailable())
         {
            if (!IsCgiPlus)
            {
               if (!strncasecmp (cptr = CgiRequestUriPtr, "/cgi-bin/", 9))
               {
                  /* redirect to "standard" CGIplus path */
                  fprintf (stdout,
"Status: 302\r\n\
Location: /cgiplus-bin/%s\r\n\
\r\n\
Must be CGIplus!\n",
                           cptr+9);
               }
               else
               {
                  /* not the "standard" path */
                  fprintf (stdout, "Status: 500\r\n\r\nMust be CGIplus!\n");
               }
               exit (SS$_NORMAL);
            }
         }

         fprintf (stdout,
"Content-Type: text/html\r\n\
\r\n\
<!DOCTYPE html>\n\
<html>\n\
<head>\n\
<title>&agrave;lamode</title>\n\
<meta http-equiv=\"X-UA-Compatible\" content=\"IE=edge\">\n\
<meta name=\"copyright\" content=\"%s - GPL licensed\">\n\
<script language=\"JavaScript\">\n\
var $ExeVersion = \'%s\';\n\
var $NoLoad = \'JavaScript not loading!\';\n\
var $ScriptName = \'%s\';\n\
var $WebSocket = %s;\n\
function $LoadError() { alert($NoLoad); \
document.getElementsByTagName(\'body\')[0].innerHTML += $NoLoad; }\n\
</script>\n\
<script type=\"text/javascript\" onerror=\"$LoadError()\" \
src=\"/alamode/-/acme.js\"></script>\n\
</head>\n\
<body>\n\
&agrave;lamode v%s <noscript>requires JavaScript!</noscript>\n\
</body>\n\
</html>\n",
                 SOFTWARECR, SOFTWAREVN, WsLibCgiVar("SCRIPT_NAME"),
                 WebSockAvailable() && IsCgiPlus ? "true" : "false",
                 SOFTWAREVN);
      }

      if (!IsCgiPlus) break;

      WsLibCgiPlusEof ();
   }

   exit (SS$_NORMAL);
}

/*****************************************************************************/
/*
Allocate a client structure and add it to the head of the list.  Establish the
WebSocket IPC and begin processing.
*/

void AddClient ()

{
   int  status;
   CLIENT_DATA  *clptr;

   /*********/
   /* begin */
   /*********/

   if (!(clptr = calloc (1, sizeof(CLIENT_DATA))))
      EXIT_FI_LI (vaxc$errno);

   /* create a WebSocket library structure for the client */
   if (!(clptr->WsLibPtr = WsLibCreate (clptr, RemoveClient)))
   {
      /* failed, commonly on some WebSocket protocol issue */
      return;
   }

   /* open the IPC to the WebSocket (mailboxes) */
   status = WsLibOpen (clptr->WsLibPtr);
   if (VMSnok (status)) EXIT_FI_LI (status);

   if (!strcmp (CgiRemoteUserPtr, "WORLD"))
      clptr->WorldAccess = true;

   WsLibWatchScript (clptr->WsLibPtr, FI_LI, "!AZ", SOFTWAREID);

   /* provide the static system data */
   WsLibWrite (clptr->WsLibPtr,
               SystemData.StaticBuffer,
               SystemData.StaticBufferLength,
               WSLIB_ASYNCH);

   /* queue an asynchronous read from the client */
   WsLibRead (clptr->WsLibPtr,
              clptr->InputBuffer,
              sizeof(clptr->InputBuffer)-1,
              ReadClient);

   ConnectedCount++;
}

/*****************************************************************************/
/*
Remove the client structure from the list and free the memory.
*/

void RemoveClient (struct WsLibStruct *wsptr)

{
   struct MonClient  *clptr;

   /*********/
   /* begin */
   /*********/

   clptr = WsLibGetUserData(wsptr);

   if (ConnectedCount) ConnectedCount--;
}

/*****************************************************************************/
/*
*/

void ReadClient (struct WsLibStruct *wsptr)

{
   CLIENT_DATA  *clptr;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "ReadClient() %%X%08.08X\n",  WsLibReadStatus (wsptr));

   if (VMSnok (WsLibReadStatus (wsptr)))
   {
      WsLibClose (wsptr, 0, NULL);
      return;
   }

   clptr = WsLibGetUserData (wsptr);

   /* ensure it's null-terminated */
   clptr->InputBuffer[WsLibReadCount(wsptr)] = '\0';

   if (strstr (clptr->InputBuffer, "detail=1"))
   {
      clptr->ProvideDetailData = true;
      JsonDetailData ();
      WsLibWrite (wsptr,
                  SystemData.DetailBuffer,
                  SystemData.DetailBufferLength,
                  WSLIB_ASYNCH);
   }
   else
   if (strstr (clptr->InputBuffer, "detail=0"))
      clptr->ProvideDetailData = false;

   if (strstr (clptr->InputBuffer, "instance=1"))
   {
      clptr->ProvideInstanceData = true;
      JsonInstanceData ();
      WsLibWrite (wsptr,
                  SystemData.InstanceBuffer,
                  SystemData.InstanceBufferLength,
                  WSLIB_ASYNCH);
   }
   else
   if (strstr (clptr->InputBuffer, "instance=0"))
      clptr->ProvideInstanceData = false;

   if (strstr (clptr->InputBuffer, "proxy=1"))
   {
      clptr->ProvideProxyData = true;
      JsonProxyData (true);
      WsLibWrite (wsptr,
                  SystemData.ProxyBuffer,
                  SystemData.ProxyBufferLength,
                  WSLIB_ASYNCH);
   }
   else
   if (strstr (clptr->InputBuffer, "proxy=0"))
      clptr->ProvideProxyData = false;

   if (strstr (clptr->InputBuffer, "request=1"))
   {
      clptr->ProvideRequestData = true;
      JsonRequestData (clptr, true);
      WsLibWrite (wsptr,
                  SystemData.RequestBuffer,
                  SystemData.RequestBufferLength,
                  WSLIB_ASYNCH);
   }
   else
   if (strstr (clptr->InputBuffer, "request=0"))
      clptr->ProvideRequestData = false;

   if (strstr (clptr->InputBuffer, "script=1"))
   {
      clptr->ProvideScriptData = true;
      JsonScriptData ();
      WsLibWrite (wsptr,
                  SystemData.ScriptBuffer,
                  SystemData.ScriptBufferLength,
                  WSLIB_ASYNCH);
   }
   else
   if (strstr (clptr->InputBuffer, "script=0"))
      clptr->ProvideScriptData = false;

   if (strstr (clptr->InputBuffer, "webdav=1"))
   {
      clptr->ProvideWebDavData = true;
      JsonWebDavData ();
      WsLibWrite (wsptr,
                  SystemData.WebDavBuffer,
                  SystemData.WebDavBufferLength,
                  WSLIB_ASYNCH);
   }
   else
   if (strstr (clptr->InputBuffer, "webdav=0"))
      clptr->ProvideWebDavData = false;

   /* read next */
   WsLibRead (wsptr,
              clptr->InputBuffer,
              sizeof(clptr->InputBuffer)-1,
              ReadClient);
}

/*****************************************************************************/
/*
For an XMLHttpRequest() HTTP request.  The response body will stream chunks of
JSON data to be parsed by the browser JavaScript receiving it.  Do the best we
can to disable proxy caching ensuring a timely flow of response data.
*/

void XhrResponse ()

{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "XhrResponse()\n");

   fprintf (stdout,
"Status: 200\n\
Content-Type: text/html\n\
Script-Control: X-stream-mode=1\n\
Script-Control: X-content-encoding-gzip=0\n\
Pragma: no-cache\n\
Cache-Control: no-cache, no-store, private\n\
\n");

   fprintf (stdout, "%s%s", SystemData.StaticBuffer, JSON_SENTINAL);
   fflush (stdout);
} 

/*****************************************************************************/
/*
Send a JSON data structure containing a message to be displayed as an alert.
*/

void JsonAlert
(
CLIENT_DATA *ClientPtr,
char *FormatString,
...
)
{
   static char  CurTime [24];
   static $DESCRIPTOR (CurTimeDsc, CurTime);
   static $DESCRIPTOR (CurTimeFaoDsc, "!20%D\0");

   int  len;
   char  jsonbuf [512],
         msgbuf [256];
   va_list  argptr;
   CLIENT_DATA  *clptr;
   struct WsLibStruct  *wsctx, *wsptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "JsonAlert()\n");

   va_start (argptr, FormatString);
#ifndef __VAX
   len = vsnprintf (msgbuf, sizeof(msgbuf), FormatString, argptr);
#else
   len = vsprintf (msgbuf, FormatString, argptr);
#endif
   if (len >= sizeof(msgbuf)) EXIT_FI_LI (SS$_RESULTOVF);

   sys$fao (&CurTimeFaoDsc, 0, &CurTimeDsc, &CurrentBinTime);
   if (CurTime[0] == ' ') CurTime[0] = '0';

#ifndef __VAX
   len = snprintf (jsonbuf, sizeof(jsonbuf),
                   "{\"$data\":\"alert\",\"time\":\"%s\",\"message\":\"%s\"}",
                   CurTime, msgbuf);
#else
   len = sprintf (jsonbuf,
                  "{\"$data\":\"alert\",\"time\":\"%s\",\"message\":\"%s\"}",
                  CurTime, msgbuf);
#endif
   if (len >= sizeof(jsonbuf)) EXIT_FI_LI (SS$_RESULTOVF);

   if (DataViaCGI)
   {
#if CgiClientData
      if (!ClientPtr || ClientPtr == &CgiClientData)
#endif
      {
         fputs (jsonbuf, stdout);
         fputs (JSON_SENTINAL, stdout);
         fflush (stdout);
      }
   }

   wsctx = NULL;
   while (wsptr = WsLibNext(&wsctx))
   {
      clptr = WsLibGetUserData (wsptr);
      if (ClientPtr && clptr != ClientPtr) continue;
      /* set and forget; buffer each message internally */
      WsLibSetBuffer (clptr->WsLibPtr);
      WsLibWrite (wsptr, jsonbuf, len, WSLIB_ASYNCH);
      /* revert to requiring an external, persistent buffer */
      WsLibSetNoBuffer (clptr->WsLibPtr);
   }
} 

/*****************************************************************************/
/*
Indepdendent, timer AST driven update of the system performance data.
This data is then written to connected clients.
*/

void ProvideData ()

{
   static const long cvtf_mode = LIB$K_DELTA_SECONDS_F;
   static ulong  PrevBinTime [2],
                 SecondsDelta [2];

   int  status;
   ulong  fltf;  /* generic 32 bits for the VAX float */
   ulong  DiffBinTime [2];
   char  *cptr, *sptr, *zptr;
   struct WsLibStruct  *wsctx, *wsptr;
   CLIENT_DATA  *clptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ProvideData()\n");

   GetTimFunction (&CurrentBinTime);
   sys$numtim (&CurrentNumTime, &CurrentBinTime);
   TimeStamp = decc$fix_time (&CurrentBinTime);

   if (ExitTimeStamp && TimeStamp > ExitTimeStamp) exit (SS$_NORMAL); 

   if (QUAD_ZERO(&PrevBinTime))
   {
      PUT_QUAD_QUAD (&CurrentBinTime, &PrevBinTime);
      if (Debug) fprintf (stdout, "%d %s\n", SystemData.StaticBufferLength,
                                             SystemData.StaticBuffer);
   }

   /* calculate a floating point delta time */
   status = lib$sub_times (&CurrentBinTime,
                           &PrevBinTime,
                           &DiffBinTime);
   if (VMSnok (status)) EXIT_FI_LI (status);

   PUT_QUAD_QUAD (&CurrentBinTime, &PrevBinTime);

#ifdef __ia64
   lib$cvts_from_internal_time (&cvtf_mode, &DeltaSeconds, &DiffBinTime);
#endif
#ifdef __ALPHA
   /* Alpha 7.3-1 does not have lib$cvts_..() */
   lib$cvtf_from_internal_time (&cvtf_mode, &fltf, &DiffBinTime);
   /* convert VAX to IEEE float */
   status = cvt$convert_float (&fltf, CVT$K_VAX_F,
                               &DeltaSeconds, CVT$K_IEEE_S,
                               CVT$M_ROUND_TO_NEAREST);
   if (VMSnok (status)) EXIT_FI_LI (status);
#endif
#ifdef __VAX
   lib$cvtf_from_internal_time (&cvtf_mode, &DeltaSeconds, &DiffBinTime);
#endif
   if (Debug) fprintf (stdout, "DeltaSeconds: %f\n", DeltaSeconds);

   /***********/
   /* collect */
   /***********/

   CollectSPI ();
   CollectNetIntData ();
   CollectInstanceData ();

   if (DataViaCGI)
   {
      /*******/
      /* CGI */
      /*******/

      /* collect it all */
      CollectProxyData ();  /* data dependency pre CollectAccounting() */
      CollectScriptData ();
      CollectWebDavData ();

      /* format it all */
      JsonDetailData ();
      JsonInstanceData ();
      JsonProxyData (false);
      JsonRequestData (NULL, false);
      JsonScriptData ();
      JsonWebDavData ();
   }
   else
   {
      /*************/
      /* WebSocket */
      /*************/

      /* only collect these data if a client requires them */
      for (wsctx = NULL; wsptr = WsLibNext(&wsctx);)
         if (((CLIENT_DATA*)WsLibGetUserData(wsptr))->ProvideDetailData)
         {
            JsonDetailData ();
            break;
         }

      for (wsctx = NULL; wsptr = WsLibNext(&wsctx);)
         if (((CLIENT_DATA*)WsLibGetUserData(wsptr))->ProvideInstanceData)
         {
            JsonInstanceData ();
            break;
         }

      for (wsctx = NULL; wsptr = WsLibNext(&wsctx);)
         if (((CLIENT_DATA*)WsLibGetUserData(wsptr))->ProvideProxyData)
         {
            CollectProxyData ();  /* data dependency pre CollectAccounting() */
            JsonProxyData (false);
            break;
         }

      for (wsctx = NULL; wsptr = WsLibNext(&wsctx);)
         if ((clptr = (CLIENT_DATA*)WsLibGetUserData(wsptr))->ProvideRequestData)
         {
            JsonRequestData (clptr, false);
            break;
         }

      for (wsctx = NULL; wsptr = WsLibNext(&wsctx);)
         if (((CLIENT_DATA*)WsLibGetUserData(wsptr))->ProvideScriptData)
         {
            CollectScriptData ();
            JsonScriptData ();
            break;
         }

      for (wsctx = NULL; wsptr = WsLibNext(&wsctx);)
         if (((CLIENT_DATA*)WsLibGetUserData(wsptr))->ProvideWebDavData)
         {
            CollectWebDavData ();
            JsonWebDavData ();
            break;
         }
   }

   CollectAccounting ();

   JsonSummaryData ();

   /**********/
   /* alerts */
   /**********/

   if (SystemData.InstanceStartupCount > SystemData.PrevInstanceStartupCount)
      JsonAlert (NULL, "!Server instance startup from %d to %d",
                 SystemData.PrevInstanceStartupCount,
                 SystemData.InstanceStartupCount);
   SystemData.PrevInstanceStartupCount = SystemData.InstanceStartupCount;

   if (SystemData.TotalDclHardLimit > SystemData.PrevTotalDclHardLimit)
      JsonAlert (NULL, "!DCL scripting hard-limit hit from %d to %d",
                 SystemData.PrevTotalDclHardLimit,
                 SystemData.TotalDclHardLimit);
   SystemData.PrevTotalDclHardLimit = SystemData.TotalDclHardLimit;

   if (SystemData.StatusCode403 >
       SystemData.PrevStatusCode403 + AlertStatusCode403)
   {
      JsonAlert (NULL, "!HTTP status 403 exceeded %d/minute",
                 AlertStatusCode403);
      SystemData.PrevStatusCode403 = SystemData.StatusCode403;
   }
   else
   if (CurrentNumTime[4] != SystemData.PrevMinute)
      SystemData.PrevStatusCode403 = SystemData.StatusCode403;

   if (SystemData.StatusCodeGroup[5] >
       SystemData.PrevStatusCodeGroup5 + AlertStatusCodeGroup5)
   {
      JsonAlert (NULL, "!HTTP status 5nn exceeded %d/minute",
                 AlertStatusCodeGroup5);
      SystemData.PrevStatusCodeGroup5 = SystemData.StatusCodeGroup[5];
   }
   else
   if (CurrentNumTime[4] != SystemData.PrevMinute)
      SystemData.PrevStatusCodeGroup5 = SystemData.StatusCodeGroup[5];

   GetTimFunction (&CollectBinTime);

   /**********/
   /* supply */
   /**********/

   if (DataViaCGI)
   {
      /*******/
      /* CGI */
      /*******/

      /* puts three backslashes into the stream - impossible JSON data */
      if (SystemData.SummaryBufferLength)
      {
         fputs (SystemData.SummaryBuffer, stdout);
         fputs (JSON_SENTINAL, stdout);
      }

      if (SystemData.DetailBufferLength)
      {
         fputs (SystemData.DetailBuffer, stdout);
         fputs (JSON_SENTINAL, stdout);
      }

      if (SystemData.InstanceBufferLength)
      {
         fputs (SystemData.InstanceBuffer, stdout);
         fputs (JSON_SENTINAL, stdout);
      }

      if (SystemData.ProxyBufferLength)
      {
         fputs (SystemData.ProxyBuffer, stdout);
         fputs (JSON_SENTINAL, stdout);
      }

      if (SystemData.RequestBufferLength)
      {
         fputs (SystemData.RequestBuffer, stdout);
         fputs (JSON_SENTINAL, stdout);
      }

      if (SystemData.ScriptBufferLength)
      {
         fputs (SystemData.ScriptBuffer, stdout);
         fputs (JSON_SENTINAL, stdout);
      }

      if (SystemData.WebDavBufferLength)
      {
         fputs (SystemData.WebDavBuffer, stdout);
         fputs (JSON_SENTINAL, stdout);
      }

      if (SystemData.RuntimeBufferLength)
      {
         fprintf (stdout, "%s,\"CGIplus\":%s}",
                  SystemData.RuntimeBuffer, IsCgiPlus ? "true" : "false");
         fputs (JSON_SENTINAL, stdout);
      }

      fflush (stdout);
   }

   /* only supply (some of) these data if the client requires them */
   for (wsctx = NULL; wsptr = WsLibNext(&wsctx);)
   {
      /*************/
      /* WebSocket */
      /*************/

      clptr = WsLibGetUserData (wsptr);

      zptr = (sptr = clptr->OutputBuffer) + sizeof(clptr->OutputBuffer)-1;

      for (cptr = SystemData.SummaryBuffer;
           *cptr && sptr < zptr;
           *sptr++ = *cptr++);
      for (cptr = JSON_SENTINAL; *cptr && sptr < zptr; *sptr++ = *cptr++);

      if (clptr->ProvideDetailData &&
          SystemData.DetailBufferLength)
      {
         for (cptr = SystemData.DetailBuffer;
              *cptr && sptr < zptr;
              *sptr++ = *cptr++);
         for (cptr = JSON_SENTINAL; *cptr && sptr < zptr; *sptr++ = *cptr++);
      }

      if (clptr->ProvideInstanceData &&
          SystemData.InstanceBufferLength)
      {
         for (cptr = SystemData.InstanceBuffer;
              *cptr && sptr < zptr;
              *sptr++ = *cptr++);
         for (cptr = JSON_SENTINAL; *cptr && sptr < zptr; *sptr++ = *cptr++);
      }

      if (clptr->ProvideProxyData &&
          SystemData.ProxyBufferLength)
      {
         for (cptr = SystemData.ProxyBuffer;
              *cptr && sptr < zptr;
              *sptr++ = *cptr++);
         for (cptr = JSON_SENTINAL; *cptr && sptr < zptr; *sptr++ = *cptr++);
      }

      if (clptr->ProvideRequestData &&
          SystemData.RequestBufferLength)
      {
         for (cptr = SystemData.RequestBuffer;
              *cptr && sptr < zptr;
              *sptr++ = *cptr++);
         for (cptr = JSON_SENTINAL; *cptr && sptr < zptr; *sptr++ = *cptr++);
      }

      if (clptr->ProvideScriptData &&
          SystemData.ScriptBufferLength)
      {
         for (cptr = SystemData.ScriptBuffer;
              *cptr && sptr < zptr;
              *sptr++ = *cptr++);
         for (cptr = JSON_SENTINAL; *cptr && sptr < zptr; *sptr++ = *cptr++);
      }

      if (clptr->ProvideWebDavData &&
          SystemData.WebDavBufferLength)
      {
         for (cptr = SystemData.WebDavBuffer;
              *cptr && sptr < zptr;
              *sptr++ = *cptr++);
         for (cptr = JSON_SENTINAL; *cptr && sptr < zptr; *sptr++ = *cptr++);
      }

      if (SystemData.RuntimeBufferLength)
      {
         for (cptr = SystemData.RuntimeBuffer;
              *cptr && sptr < zptr;
              *sptr++ = *cptr++);
         for (cptr = ",\"WebSocket\":true}";
              *cptr && sptr < zptr;
              *sptr++ = *cptr++);
         for (cptr = JSON_SENTINAL; *cptr && sptr < zptr; *sptr++ = *cptr++);
      }

      if (sptr > zptr) EXIT_FI_LI (SS$_RESULTOVF);

      WsLibWrite (wsptr, clptr->OutputBuffer,
                  sptr - clptr->OutputBuffer, WSLIB_ASYNCH);
   }

   if (Debug) DisplayDebug();

   /*************/
   /* and again */
   /*************/

   SystemData.PrevMinute = CurrentNumTime[4];

   /* one to sixty seconds, just multiply by interval */
   if (QUAD_ZERO (SecondsDelta))
   {
      PUT_LONG_QUAD (RefreshInterval, SecondsDelta);
      MUL_LONG_QUAD (-10000000, SecondsDelta);
   }

   status = sys$setimr (0, &SecondsDelta, ProvideData, NULL, 0);

   if (VMSnok (status)) EXIT_FI_LI (status);

   ProvideDataSeconds += RefreshInterval;

   GetTimFunction (&ProvideBinTime);
   JsonRuntime ();
}

/*****************************************************************************/
/*
Provide some runtime data.  Average duration of each data collection and total
monitor cycle (to differentiate the time required to write the data to the
client(s)), average CPU consumed, average user and non-user CPU modes.  All in
seconds.  With VAX clock granularity, durations may fail to register if
sub-10mS.  This data is not available to the end-user, is for developer and 
technical insight purposes, and can be seen from the data stream.
*/

void JsonRuntime ()

{
   static ulong  CollectTotalTime [2],
                 ProvideTotalTime [2];

   static long  CpuSekInit,
                CpuTimInit,
                CpuUsrInit,
                JpiCpuTim,
                JpiExecTim,
                JpiKrnlTim,
                JpiSuprTim,
                JpiUserTim,
                StatCount;

   static struct
   {
      short  BufferLength;
      short  ItemCode;
      void  *BufferPtr;
      void  *LengthPtr;
   } ItemListJpi[] =
   {
       { sizeof(JpiCpuTim), JPI$_CPUTIM, &JpiCpuTim, 0 },
#ifndef __VAX 
       { sizeof(JpiExecTim),   JPI$_EXECTIM,   &JpiExecTim, 0 },
       { sizeof(JpiKrnlTim),   JPI$_KRNLTIM,   &JpiKrnlTim, 0 },
       { sizeof(JpiSuprTim),   JPI$_SUPRTIM,   &JpiSuprTim, 0 },
       { sizeof(JpiUserTim),   JPI$_USERTIM,   &JpiUserTim, 0 },
#endif
       {0,0,0,0}
   };

   int  status;
   ulong  CpuUSR,
          CpuSEK,
          CpuTIM;
   ulong  RunBinTime [2],
          DiffBinTime [2];
   char  *sptr, *zptr;
   float  fcount;
   IOSBLK  IOsb;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "\nJsonRuntime()\n");

   status = sys$getjpiw (
            SpiEfn,  /* efn */
            0,  /* pidaddr */
            0,  /* prcnam */
            &ItemListJpi,  /* item list */
            &IOsb,  /* iosb */
            0,      /* astaddr */
            0 );    /* astprm */

   if (VMSok (status)) status = IOsb.iosb$w_status;
   if (VMSnok (status)) EXIT_FI_LI (status);

   if ((fcount = (float)(StatCount++)) == 0.0)
   {
      /* track and offset initial setup overheads */
      CpuTimInit = JpiCpuTim;
      CpuUsrInit = JpiUserTim;
      CpuSekInit = JpiSuprTim + JpiExecTim + JpiKrnlTim;
      return;
   }

   SUB_QUAD_QUAD (CurrentBinTime, CollectBinTime);
   ADD_QUAD_QUAD (CollectBinTime, CollectTotalTime);
   SUB_QUAD_QUAD (CurrentBinTime, ProvideBinTime);
   ADD_QUAD_QUAD (ProvideBinTime, ProvideTotalTime);
   MINUS_QUAD_QUAD (CurrentBinTime, StartBinTime, RunBinTime);

   CpuTIM = JpiCpuTim - CpuTimInit;
   CpuUSR = JpiUserTim - CpuUsrInit;
   CpuSEK = JpiSuprTim + JpiExecTim + JpiKrnlTim - CpuSekInit;

   zptr = (sptr = SystemData.RuntimeBuffer) + sizeof(SystemData.RuntimeBuffer);

   sptr += snprintf (sptr, zptr-sptr,
"{\"$data\":\"runtime\",\
\"cpu\":%f,\
\"cpuTIM\":%f,",
                     (float)JpiCpuTim / 100.0,
                     (float)CpuTIM / 100.0 / fcount);

#ifndef __VAX
   sptr += snprintf (sptr, zptr-sptr,
"\"cpuUSR\":%f,\
\"cpuSEK\":%f,",
                     (float)CpuUSR / 100.0 / fcount,
                     (float)CpuSEK / 100.0 / fcount);
#endif

   sptr += snprintf (sptr, zptr-sptr,
"\"run\":\"%s\",\
\"collect\":%f,\
\"provide\":%f,\
\"usage\":%d,\
\"connected\":%d",
                     MungeTime(RunBinTime),
                     FLOAT_FROM_QUAD(CollectTotalTime) / 10000000.0 / fcount,
                     FLOAT_FROM_QUAD(ProvideTotalTime) / 10000000.0 / fcount,
                     UsageCount,
                     ConnectedCount);

   if (sptr >= zptr) EXIT_FI_LI (SS$_RESULTOVF);

   SystemData.RuntimeBufferLength = sptr - SystemData.RuntimeBuffer;
}

/*****************************************************************************/
/*
For the command-line debug mode.
*/

void DisplayDebug ()

{
   JsonDetailData ();
   JsonInstanceData ();
   JsonProxyData ();
   JsonRequestData (NULL, true);
   JsonScriptData ();
   JsonWebDavData ();
   fprintf (stdout, "*****%d %s\n", SystemData.SummaryBufferLength,
                                    SystemData.SummaryBuffer);
   fprintf (stdout, "*****%d %s\n", SystemData.DetailBufferLength,
                                    SystemData.DetailBuffer);
   fprintf (stdout, "*****%d %s\n", SystemData.InstanceBufferLength,
                                    SystemData.InstanceBuffer);
   fprintf (stdout, "*****%d %s\n", SystemData.ProxyBufferLength,
                                    SystemData.ProxyBuffer);
   fprintf (stdout, "*****%d %s\n", SystemData.RequestBufferLength,
                                    SystemData.RequestBuffer);
   fprintf (stdout, "*****%d %s\n", SystemData.ScriptBufferLength,
                                    SystemData.ScriptBuffer);
   fprintf (stdout, "*****%d %s\n", SystemData.WebDavBufferLength,
                                    SystemData.WebDavBuffer);
}

/*****************************************************************************/
/*
Create a JSON data structure in a buffer representing the per-boot
characteristics of the system.
*/

void JsonStaticData ()

{
   static char  BootTime [24];
   static $DESCRIPTOR (BootTimeFaoDsc, "!17%D\0");
   static $DESCRIPTOR (BootTimeDsc, BootTime);

   ulong  Size64 [2];
   char  *cptr, *sptr, *zptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "JsonStaticData()\n");

   zptr = (sptr = SystemData.HttpdVersion) + sizeof(SystemData.HttpdVersion)-1;
   for (cptr = gblptr->HttpdVersion;
        *cptr && sptr < zptr;
        *sptr++ = *cptr++);
   *sptr = '\0';

   PUT_LONG_QUAD (SystemData.MemSize, Size64);
   MUL_LONG_QUAD (SystemData.PageSize, Size64);

   zptr = (sptr = SystemData.StaticBuffer) + sizeof(SystemData.StaticBuffer);

   sptr += snprintf (sptr, zptr-sptr, "{\"$data\":\"static\"");

   sptr += snprintf (sptr, zptr-sptr, ",\"WASD\":\"%s\"",
                     SystemData.HttpdVersion);
   sptr += snprintf (sptr, zptr-sptr, ",\"alamode\":\"%s\"", SOFTWAREVN);
   sptr += snprintf (sptr, zptr-sptr, ",\"cpuAvail\":%d", SystemData.AvailCpuCount);
   sptr += snprintf (sptr, zptr-sptr, ",\"archName\":\"%s\"",
                     SystemData.ArchName);
   sptr += snprintf (sptr, zptr-sptr, ",\"hwName\":\"%s\"", SystemData.HwName);
   sptr += snprintf (sptr, zptr-sptr, ",\"interval\":%d", RefreshInterval);
   sptr += snprintf (sptr, zptr-sptr, ",\"jpiPID\":\"%08.08X\"",
                     SystemData.JpiPid);
   sptr += snprintf (sptr, zptr-sptr, ",\"nodeName\":\"%s\"",
                     SystemData.NodeName);
   sptr += snprintf (sptr, zptr-sptr, ",\"vmsVersion\":\"%s\"",
                     SystemData.VmsVersion);
   sptr += snprintf (sptr, zptr-sptr, ",\"memSize\":%s", JsonNumber(64,Size64));
   sys$fao (&BootTimeFaoDsc, 0, &BootTimeDsc, &SystemData.BootBinTime);
   sptr += snprintf (sptr, zptr-sptr, ",\"bootTime\":\"%s\"", BootTime);
   sptr += snprintf (sptr, zptr-sptr, ",\"proxyEnabled\":%s",
                    SystemData.ProxyEnabled ? "true" : "false");
   sptr += snprintf (sptr, zptr-sptr, ",\"webdavEnabled\":%s",
                    SystemData.WebDavEnabled ? "true" : "false");

   sptr += snprintf (sptr, zptr-sptr, "}");

   if (sptr >= zptr) EXIT_FI_LI (SS$_RESULTOVF);

   SystemData.StaticBufferLength = sptr - SystemData.StaticBuffer;
}

/*****************************************************************************/
/*
Create a JSON data structure in a buffer representing the current state of the
system and server.  The data provided is "sparse" in the sense that any datum
whose value is zero is not explicitly supplied.  The JavaScript receiving the
data assumes that any datum it accesses as 'undefined' is zero.
*/

void JsonSummaryData ()

{
   static char  CurTime [24],
                UpTime [16];
   static $DESCRIPTOR (CurTimeFaoDsc, "!20%D\0");
   static $DESCRIPTOR (UpTimeFaoDsc, "!10%D\0");
   static $DESCRIPTOR (CurTimeDsc, CurTime);
   static $DESCRIPTOR (UpTimeDsc, UpTime);

   ulong  InUse64 [2];
   char  *cptr, *sptr, *zptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "JsonSummaryData()\n");

   PUT_LONG_QUAD (SystemData.MemInUse, InUse64);
   MUL_LONG_QUAD (SystemData.PageSize, InUse64);

   SystemData.SummaryBufferLength = 0;
   zptr = (sptr = SystemData.SummaryBuffer) + sizeof(SystemData.SummaryBuffer);

   sptr += snprintf (sptr, zptr-sptr, "{\"$data\":\"summary\"");

   /***************/
   /* system data */
   /***************/

   sptr += snprintf (sptr, zptr-sptr, ",\"timestamp\":%d", TimeStamp);
   sptr += snprintf (sptr, zptr-sptr, ",\"usage\":%d", UsageCount);
   sptr += snprintf (sptr, zptr-sptr, ",\"count\":%d", ConnectedCount);
   sptr += snprintf (sptr, zptr-sptr, ",\"cpuActive\":%d",
                     SystemData.ActiveCpuCount);
   if (SystemData.CpuBusyPercent)
      sptr += snprintf (sptr, zptr-sptr, ",\"cpuBusy\":%d",
                        SystemData.CpuBusyPercent);
   if (SystemData.CpuBusyPercent - SystemData.CpuUserPercent)
      sptr += snprintf (sptr, zptr-sptr, ",\"cpuSEK\":%d",
                       SystemData.CpuBusyPercent - SystemData.CpuUserPercent);
   if (SystemData.BufferedIO)
      sptr += snprintf (sptr, zptr-sptr, ",\"bio\":%u",
                        SystemData.BufferedIO);
   if (SystemData.FcpRead + SystemData.FcpWrite)
      sptr += snprintf (sptr, zptr-sptr, ",\"fcpRW\":%u",
                       SystemData.FcpRead + SystemData.FcpWrite);
   if (SystemData.FcpWrite)
      sptr += snprintf (sptr, zptr-sptr, ",\"fcpWrite\":%u",
                        SystemData.FcpWrite);
   sptr += snprintf (sptr, zptr-sptr, ",\"memInUse\":%s",
                     JsonNumber(64,InUse64));
   sptr += snprintf (sptr, zptr-sptr, ",\"sysProcs\":%d",
                     SystemData.ProcessCount);
   sptr += snprintf (sptr, zptr-sptr, ",\"webProcs\":%d",
                     SystemData.InstanceNodeCount +
                        SystemData.DclProcessCurrent,
                     SystemData.CurrentDECnetTasks);

   sptr += snprintf (sptr, zptr-sptr, ",\"netPerSec\":%d",
                     SystemData.NetBytesPerSec);
   sptr += snprintf (sptr, zptr-sptr, ",\"webPerSec\":%d",
                     SystemData.BytesRawPerSec);
   sptr += snprintf (sptr, zptr-sptr, ",\"netDgramRx\":%d",
                     SystemData.NetDatagramRx);
   sptr += snprintf (sptr, zptr-sptr, ",\"netDgramTx\":%d",
                     SystemData.NetDatagramTx);
   sptr += snprintf (sptr, zptr-sptr, ",\"webBlockRead\":%d",
                     SystemData.BlockReadSize);
   sptr += snprintf (sptr, zptr-sptr, ",\"webBlockWrite\":%d",
                     SystemData.BlockWriteSize);

   sys$fao (&UpTimeFaoDsc, 0, &UpTimeDsc, &SystemData.UpBinTime);
   for (cptr = UpTime; *cptr && *cptr == ' '; cptr++);
   sptr += snprintf (sptr, zptr-sptr, ",\"upTime\":\"%s\"", cptr);

   sys$fao (&CurTimeFaoDsc, 0, &CurTimeDsc, &CurrentBinTime);
   if (CurTime[0] == ' ') CurTime[0] = '0';
   sptr += snprintf (sptr, zptr-sptr, ",\"vmsTime\":\"%s\"", CurTime);

   /***************/
   /* server data */
   /***************/

   sptr += snprintf (sptr, zptr-sptr, ",\"instanceNode\":%d",
                     SystemData.InstanceNodeCount);
   sptr += snprintf (sptr, zptr-sptr, ",\"instanceCluster\":%d",
                     SystemData.InstanceClusterCount);
   sptr += snprintf (sptr, zptr-sptr, ",\"instanceStartup\":%d",
                     SystemData.InstanceStartupCount);

   if (SystemData.TotalFile)
      sptr += snprintf (sptr, zptr-sptr, ",\"totalFile\":%d",
                        SystemData.TotalFile);
   sptr += snprintf (sptr, zptr-sptr, ",\"totalRequest\":%d",
                     SystemData.TotalRequest);
   if (SystemData.TotalAllScript)
      sptr += snprintf (sptr, zptr-sptr, ",\"totalScript\":%d",
                        SystemData.TotalAllScript);
   if (SystemData.SslCount)
      sptr += snprintf (sptr, zptr-sptr, ",\"totalSSL\":%d",
                        SystemData.SslCount);
   if (SystemData.TotalAllOther)
      sptr += snprintf (sptr, zptr-sptr, ",\"totalOther\":%d",
                        SystemData.TotalAllOther);
   if (SystemData.TotalProxy)
      sptr += snprintf (sptr, zptr-sptr, ",\"totalProxy\":%d",
                        SystemData.TotalProxy);
   if (SystemData.TotalWebDav)
      sptr += snprintf (sptr, zptr-sptr, ",\"totalWebDAV\":%d",
                        SystemData.TotalWebDav);

   if (SystemData.TotalThrottled)
      sptr += snprintf (sptr, zptr-sptr, ",\"totalThrottled\":%d",
                        SystemData.TotalThrottled);
   if (SystemData.TotalThrottleQueued)
      sptr += snprintf (sptr, zptr-sptr, ",\"totalThrottleQueued\":%d",
                        SystemData.TotalThrottleQueued);
   if (SystemData.TotalThrottleBusy)
      sptr += snprintf (sptr, zptr-sptr, ",\"totalThrottleBusy\":%d",
                        SystemData.TotalThrottleBusy);

   sptr += snprintf (sptr, zptr-sptr, ",\"bytesRx\":%s",
                     JsonNumber(64,&SystemData.BytesRawRx));
   sptr += snprintf (sptr, zptr-sptr, ",\"bytesTx\":%s",
                     JsonNumber(64,&SystemData.BytesRawTx));

   sptr += snprintf (sptr, zptr-sptr, ",\"statusCodes\":[%d,%d,%d,%d,%d,%d]",
                     SystemData.StatusCodeGroup[0],
                     SystemData.StatusCodeGroup[1],
                     SystemData.StatusCodeGroup[2],
                     SystemData.StatusCodeGroup[3],
                     SystemData.StatusCodeGroup[4],
                     SystemData.StatusCodeGroup[5]);
   if (SystemData.StatusCode403)
      sptr += snprintf (sptr, zptr-sptr, ",\"status403\":%d",
                        SystemData.StatusCode403);

   /****************/
   /* current data */
   /****************/

   sptr += snprintf (sptr, zptr-sptr, ",\"nowConnect\":%d",
                     SystemData.CurrentConnected);
   sptr += snprintf (sptr, zptr-sptr, ",\"peakConnect\":%d",
                     SystemData.ConnectPeak);
   if (SystemData.CurrentProcessing)
      sptr += snprintf (sptr, zptr-sptr, ",\"nowProcess\":%d",
                        SystemData.CurrentProcessing);
   sptr += snprintf (sptr, zptr-sptr, ",\"peakProcess\":%d",
                     SystemData.ProcessingPeak);

   if (SystemData.DeltaFile)
      sptr += snprintf (sptr, zptr-sptr, ",\"deltaFile\":%d",
                        SystemData.DeltaFile);
   if (SystemData.DeltaFile304)
      sptr += snprintf (sptr, zptr-sptr, ",\"deltaFile304\":%d",
                        SystemData.DeltaFile304);
   if (SystemData.DeltaOther)
      sptr += snprintf (sptr, zptr-sptr, ",\"deltaOther\":%d",
                        SystemData.DeltaOther);

   if (SystemData.DeltaRequest)
      sptr += snprintf (sptr, zptr-sptr, ",\"deltaRequest\":%d",
                        SystemData.DeltaRequest);
   if (SystemData.DeltaRequestFail)
      sptr += snprintf (sptr, zptr-sptr, ",\"deltaRequestFail\":%d",
                        SystemData.DeltaRequestFail);

   if (SystemData.DeltaThrottled)
      sptr += snprintf (sptr, zptr-sptr, ",\"deltaThrottled\":%d",
                        SystemData.DeltaThrottled);
   if (SystemData.DeltaThrottleQueued)
      sptr += snprintf (sptr, zptr-sptr, ",\"deltaThrottleQueued\":%d",
                        SystemData.DeltaThrottleQueued);
   if (SystemData.DeltaThrottleBusy)
      sptr += snprintf (sptr, zptr-sptr, ",\"deltaThrottleBusy\":%d",
                        SystemData.DeltaThrottleBusy);

   if (SystemData.CurrentThrottled)
      sptr += snprintf (sptr, zptr-sptr, ",\"nowThrottled\":%d",
                        SystemData.CurrentThrottled);
   if (SystemData.CurrentThrottleProcessing)
      sptr += snprintf (sptr, zptr-sptr, ",\"nowThrottledProcessing\":%d",
                        SystemData.CurrentThrottleProcessing);
   if (SystemData.CurrentThrottleQueued)
      sptr += snprintf (sptr, zptr-sptr, ",\"nowThrottledQueued\":%d",
                        SystemData.CurrentThrottleQueued);
   if (SystemData.CurrentThrottleBusyMetric)
      sptr += snprintf (sptr, zptr-sptr, ",\"nowThrottledBusy\":%d",
                        SystemData.CurrentThrottleBusyMetric);

    if (SystemData.DeltaWebDavTotal)
      sptr += snprintf (sptr, zptr-sptr, ",\"deltaWebDAVtotal\":%d",
                        SystemData.DeltaWebDavTotal);
    if (SystemData.DeltaWebDavRead)
      sptr += snprintf (sptr, zptr-sptr, ",\"deltaWebDAVread\":%d",
                        SystemData.DeltaWebDavRead);
    if (SystemData.DeltaWebDavWrite)
      sptr += snprintf (sptr, zptr-sptr, ",\"deltaWebDAVwrite\":%d",
                        SystemData.DeltaWebDavWrite);
 
   if (SystemData.TotalWebSocket)
      sptr += snprintf (sptr, zptr-sptr, ",\"totalWebSocket\":%d",
                        SystemData.TotalWebSocket);
   if (SystemData.CurrentWebSockets)
      sptr += snprintf (sptr, zptr-sptr, ",\"nowWebSocket\":%d",
                        SystemData.CurrentWebSockets);

   /****************/
   /* delta script */
   /****************/

   if (SystemData.DeltaAllScript)
      sptr += snprintf (sptr, zptr-sptr, ",\"deltaScript\":%d",
                        SystemData.DeltaAllScript);
   if (SystemData.DeltaScriptCgi)
      sptr += snprintf (sptr, zptr-sptr, ",\"deltaScriptCGI\":%d",
                        SystemData.DeltaScriptCgi);

   if (SystemData.DeltaDECnetCgi + SystemData.DeltaDECnetOsu)
      sptr += snprintf (sptr, zptr-sptr, ",\"deltaDECnet\":%d",
                        SystemData.DeltaDECnetCgi + SystemData.DeltaDECnetOsu);
   if (SystemData.DeltaDECnetOsu)
      sptr += snprintf (sptr, zptr-sptr, ",\"deltaDECnetOSU\":%d",
                        SystemData.DeltaDECnetOsu);

   if (SystemData.DeltaCgiPlus + SystemData.DeltaRte)
      sptr += snprintf (sptr, zptr-sptr, ",\"deltaCGIplusRTE\":%d",
                        SystemData.DeltaCgiPlus + SystemData.DeltaRte);
   if (SystemData.DeltaCgiPlusReused + SystemData.DeltaRteReused)
      sptr += snprintf (sptr, zptr-sptr, ",\"deltaCGIplusRTEreused\":%d",
                        SystemData.DeltaCgiPlusReused +
                        SystemData.DeltaRteReused);

   if (SystemData.DeltaDclCrePrc)
      sptr += snprintf (sptr, zptr-sptr, ",\"deltaCREPRC\":%d",
                        SystemData.DeltaDclCrePrc);
   if (SystemData.DeltaDclDelPrc)
      sptr += snprintf (sptr, zptr-sptr, ",\"deltaDELPRC\":%d",
                        SystemData.DeltaDclDelPrc);

   /**************/
   /* proxy data */
   /**************/

   if (SystemData.ProxyEnabled)
   {
      if (SystemData.ProxyDeltaTotal)
         sptr += snprintf (sptr, zptr-sptr, ",\"deltaProxy\":%d",
                           SystemData.ProxyDeltaTotal);
      if (SystemData.ProxyDeltaCacheRead)
         sptr += snprintf (sptr, zptr-sptr, ",\"deltaProxyCache\":%d",
                           SystemData.ProxyDeltaCacheRead);
      if (SystemData.ProxyDeltaCacheWrite)
         sptr += snprintf (sptr, zptr-sptr, ",\"deltaProxyWrite\":%d",
                           SystemData.ProxyDeltaCacheWrite);
   }

   sptr += snprintf (sptr, zptr-sptr, "}");

   if (sptr >= zptr) EXIT_FI_LI (SS$_RESULTOVF);

   SystemData.SummaryBufferLength = sptr - SystemData.SummaryBuffer;
}

/*****************************************************************************/
/*
Create a JSON data structure in a buffer containing the detailed server data.
*/

void JsonDetailData ()

{
   char  *cptr, *sptr, *zptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "JsonDetailData()\n");

   zptr = (sptr = SystemData.DetailBuffer) + sizeof(SystemData.DetailBuffer);

   sptr += snprintf (sptr, zptr-sptr, "{\"$data\":\"detail\"");

   if (SystemData.TooBusy)
      sptr += snprintf (sptr, zptr-sptr, ",\"tooBusy\":%d", SystemData.TooBusy);

   if (SystemData.MethodConnect)
      sptr += snprintf (sptr, zptr-sptr, ",\"methConnect\":%d",
                        SystemData.MethodConnect);
   if (SystemData.MethodCopy)
      sptr += snprintf (sptr, zptr-sptr, ",\"methCopy\":%d",
                        SystemData.MethodCopy);
   if (SystemData.MethodDelete)
      sptr += snprintf (sptr, zptr-sptr, ",\"methDelete\":%d",
                        SystemData.MethodDelete);
   if (SystemData.MethodGet)
      sptr += snprintf (sptr, zptr-sptr, ",\"methGet\":%d",
                        SystemData.MethodGet);
   if (SystemData.MethodHead)
      sptr += snprintf (sptr, zptr-sptr, ",\"methHead\":%d",
                        SystemData.MethodHead);
   if (SystemData.MethodLock)
      sptr += snprintf (sptr, zptr-sptr, ",\"methLock\":%d",
                        SystemData.MethodLock);
   if (SystemData.MethodMkCol)
      sptr += snprintf (sptr, zptr-sptr, ",\"methMkCol\":%d",
                        SystemData.MethodMkCol);
   if (SystemData.MethodMove)
      sptr += snprintf (sptr, zptr-sptr, ",\"methMove\":%d",
                        SystemData.MethodMove);
   if (SystemData.MethodOptions)
      sptr += snprintf (sptr, zptr-sptr, ",\"methOptions\":%d",
                        SystemData.MethodOptions);
   if (SystemData.MethodPost)
      sptr += snprintf (sptr, zptr-sptr, ",\"methPost\":%d",
                        SystemData.MethodPost);
   if (SystemData.MethodPropFind)
      sptr += snprintf (sptr, zptr-sptr, ",\"methPropFind\":%d",
                       SystemData.MethodPropFind);
   if (SystemData.MethodPropPatch)
      sptr += snprintf (sptr, zptr-sptr, ",\"methPropPatch\":%d",
                       SystemData.MethodPropPatch);
   if (SystemData.MethodPut)
      sptr += snprintf (sptr, zptr-sptr, ",\"methPut\":%d",
                        SystemData.MethodPut);
   if (SystemData.MethodSsh)
      sptr += snprintf (sptr, zptr-sptr, ",\"methSsh\":%d",
                        SystemData.MethodSsh);
   if (SystemData.MethodTrace)
      sptr += snprintf (sptr, zptr-sptr, ",\"methTrace\":%d",
                        SystemData.MethodTrace);
   if (SystemData.MethodUnLock)
      sptr += snprintf (sptr, zptr-sptr, ",\"methUnLock\":%d",
                        SystemData.MethodUnLock);

   if (SystemData.TotalAdmin)
      sptr += snprintf (sptr, zptr-sptr, ",\"doAdmin\":%d",
                        SystemData.TotalAdmin);
   if (SystemData.TotalDclScript)
      sptr += snprintf (sptr, zptr-sptr, ",\"doDclScript\":%d",
                        SystemData.TotalDclScript);
   if (SystemData.TotalDECnet)
      sptr += snprintf (sptr, zptr-sptr, ",\"doDECnet\":%d",
                        SystemData.TotalDECnet);
   if (SystemData.TotalDir)
      sptr += snprintf (sptr, zptr-sptr, ",\"doDir\":%d",
                        SystemData.TotalDir);
   if (SystemData.TotalFile)
      sptr += snprintf (sptr, zptr-sptr, ",\"doFile\":%d",
                        SystemData.TotalFile);
   if (SystemData.TotalFile304)
      sptr += snprintf (sptr, zptr-sptr, ",\"doFile304\":%d",
                        SystemData.TotalFile304);
   if (SystemData.TotalIsMap)
      sptr += snprintf (sptr, zptr-sptr, ",\"doIsMap\":%d",
                        SystemData.TotalIsMap);
   if (SystemData.TotalMenu)
      sptr += snprintf (sptr, zptr-sptr, ",\"doMenu\":%d",
                        SystemData.TotalMenu);
   if (SystemData.TotalAllOther)
      sptr += snprintf (sptr, zptr-sptr, ",\"doOther\":%d",
                        SystemData.TotalAllOther);
   if (SystemData.TotalNoModule)
      sptr += snprintf (sptr, zptr-sptr, ",\"doNoModule\":%d",
                        SystemData.TotalNoModule);
   if (SystemData.TotalProxy)
      sptr += snprintf (sptr, zptr-sptr, ",\"doProxy\":%d",
                        SystemData.TotalProxy);
   if (SystemData.TotalPut)
      sptr += snprintf (sptr, zptr-sptr, ",\"doPut\":%d", SystemData.TotalPut);
   if (SystemData.TotalSsi)
      sptr += snprintf (sptr, zptr-sptr, ",\"doSSI\":%d", SystemData.TotalSsi);
   if (SystemData.TotalUpdate)
      sptr += snprintf (sptr, zptr-sptr, ",\"doUpdate\":%d",
                        SystemData.TotalUpdate);
   if (SystemData.TotalWebDav)
      sptr += snprintf (sptr, zptr-sptr, ",\"doWebDAV\":%d",
                        SystemData.TotalWebDav);

   if (SystemData.NetReadErr)
      sptr += snprintf (sptr, zptr-sptr, ",\"errRx\":%d",
                        SystemData.NetReadErr);
   if (SystemData.NetWriteErr)
      sptr += snprintf (sptr, zptr-sptr, ",\"errTx\":%d",
                        SystemData.NetWriteErr);

   sptr += snprintf (sptr, zptr-sptr, ",\"bpsAve\":%d",
                     SystemData.BytesPerSecAve);
   sptr += snprintf (sptr, zptr-sptr, ",\"bpsMax\":%d",
                     SystemData.BytesPerSecMax);
   sptr += snprintf (sptr, zptr-sptr, ",\"bpsMin\":%d",
                     SystemData.BytesPerSecMin);

   sptr += snprintf (sptr, zptr-sptr, "}");

   if (sptr >= zptr) EXIT_FI_LI (SS$_RESULTOVF);

   SystemData.DetailBufferLength = sptr - SystemData.DetailBuffer;
}

/*****************************************************************************/
/*
Create a JSON data structure in a buffer representing the proxy data.
*/

void JsonProxyData ()

{
   char  *cptr, *sptr, *zptr;
   ulong  quad [2];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "JsonProxyData()\n");

   SystemData.ProxyBuffer[SystemData.ProxyBufferLength = 0] = '\0';
   zptr = (sptr = SystemData.ProxyBuffer) + sizeof(SystemData.ProxyBuffer);

   if (!SystemData.ProxyEnabled) return;

   sptr += snprintf (sptr, zptr-sptr, "{\"$data\":\"proxy\"");

   if (SystemData.ProxyTotal)
      sptr += snprintf (sptr, zptr-sptr, ",\"total\":%d",
                        SystemData.ProxyTotal);

   if (QUAD_NOT_ZERO(SystemData.ProxyBytesRawRx))
      sptr += snprintf (sptr, zptr-sptr, ",\"networkRx\":%s",
                        JsonNumber(64,SystemData.ProxyBytesRawRx));

   if (QUAD_NOT_ZERO(SystemData.ProxyBytesRawTx))
      sptr += snprintf (sptr, zptr-sptr, ",\"networkTx\":%s",
                        JsonNumber(64,SystemData.ProxyBytesRawTx));

   if (QUAD_NOT_ZERO(SystemData.ProxyBytesCacheRx))
      sptr += snprintf (sptr, zptr-sptr, ",\"cacheRx\":%s",
                        JsonNumber(64,SystemData.ProxyBytesCacheRx));

   if (QUAD_NOT_ZERO(SystemData.ProxyBytesCacheTx))
      sptr += snprintf (sptr, zptr-sptr, ",\"cacheTx\":%s",
                        JsonNumber(64,SystemData.ProxyBytesCacheTx));

   if (SystemData.ProxyCacheEnabled)
      sptr += snprintf (sptr, zptr-sptr, ",\"cacheEnabled\":true");

   if (SystemData.ProxyRequestNoCache)
      sptr += snprintf (sptr, zptr-sptr, ",\"requestNoCache\":%d",
                        SystemData.ProxyRequestNoCache);

   if (SystemData.ProxyResponseNoCache)
      sptr += snprintf (sptr, zptr-sptr, ",\"responseNoCache\":%d",
                        SystemData.ProxyResponseNoCache);

   if (SystemData.ProxyLookupCache)
      sptr += snprintf (sptr, zptr-sptr, ",\"lookupCache\":%d",
                        SystemData.ProxyLookupCache);
   if (SystemData.ProxyLookupDNS)
      sptr += snprintf (sptr, zptr-sptr, ",\"lookupDNS\":%d",
                        SystemData.ProxyLookupDNS);
   if (SystemData.ProxyLookupLiteral)
      sptr += snprintf (sptr, zptr-sptr, ",\"lookupLiteral\":%d",
                        SystemData.ProxyLookupLiteral);
   if (SystemData.ProxyLookupError)
      sptr += snprintf (sptr, zptr-sptr, ",\"lookupError\":%d",
                        SystemData.ProxyLookupError);

   if (SystemData.ProxyNetwork)
      sptr += snprintf (sptr, zptr-sptr, ",\"network\":%d",
                        SystemData.ProxyNetwork);

   if (SystemData.ProxyTunnel)
      sptr += snprintf (sptr, zptr-sptr, ",\"tunnel\":%d",
                        SystemData.ProxyTunnel);

   if (SystemData.ProxyTunnelCurrent)
      sptr += snprintf (sptr, zptr-sptr, ",\"tunnelNow\":%d",
                        SystemData.ProxyTunnelCurrent);

   if (SystemData.ProxyCacheRead)
      sptr += snprintf (sptr, zptr-sptr, ",\"cacheRead\":%d",
                        SystemData.ProxyCacheRead);
   if (SystemData.ProxyCacheRead304)
      sptr += snprintf (sptr, zptr-sptr, ",\"cacheRead304\":%d",
                        SystemData.ProxyCacheRead304);
   if (SystemData.ProxyCacheWrite)
      sptr += snprintf (sptr, zptr-sptr, ",\"cacheWrite\":%d",
                       SystemData.ProxyCacheWrite);

   if (SystemData.ProxyCacheDeviceNamePtr)
      sptr += snprintf (sptr, zptr-sptr, ",\"cacheDevice\":\"%s\"",
                        SystemData.ProxyCacheDeviceNamePtr);

   if (SystemData.ProxyCacheTotalBlocks)
   {
      PUT_LONG_QUAD (SystemData.ProxyCacheTotalBlocks, quad);
      MUL_LONG_QUAD (512, quad);
      sptr += snprintf (sptr, zptr-sptr, ",\"cacheSize\":%s",
                        JsonNumber(64,quad));
   }

   if (SystemData.ProxyCacheUsedBlocks)
   {
      PUT_LONG_QUAD (SystemData.ProxyCacheUsedBlocks, quad);
      MUL_LONG_QUAD (512, quad);
      sptr += snprintf (sptr, zptr-sptr, ",\"cacheUsed\":%s",
                        JsonNumber(64,quad));
   }

   if (SystemData.ProxyCacheFreeBlocks)
   {
      PUT_LONG_QUAD (SystemData.ProxyCacheFreeBlocks, quad);
      MUL_LONG_QUAD (512, quad);
      sptr += snprintf (sptr, zptr-sptr, ",\"cacheFree\":%s",
                        JsonNumber(64,quad));
   }

   if (SystemData.ProxyCacheDeviceNamePtr &&
       SystemData.ProxyCacheSpace)
      sptr += snprintf (sptr, zptr-sptr, ",\"cacheSpace\":true");

   sptr += snprintf (sptr, zptr-sptr, "}");

   if (sptr >= zptr) EXIT_FI_LI (SS$_RESULTOVF);

   SystemData.ProxyBufferLength = sptr - SystemData.ProxyBuffer;
}

/*****************************************************************************/
/*
Create a JSON data structure in a buffer representing the WebDAV data.
*/

void JsonWebDavData ()

{
   char  *cptr, *sptr, *zptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "JsonWebDavData()\n");

   SystemData.WebDavBuffer[SystemData.WebDavBufferLength = 0] = '\0';
   zptr = (sptr = SystemData.WebDavBuffer) + sizeof(SystemData.WebDavBuffer);

   if (!SystemData.WebDavEnabled) return;

   sptr += snprintf (sptr, zptr-sptr, "{\"$data\":\"webdav\"");

   if (SystemData.TotalWebDav)
      sptr += snprintf (sptr, zptr-sptr, ",\"total\":%u",
                        SystemData.TotalWebDav);

   if (SystemData.TotalWebDavRead)
      sptr += snprintf (sptr, zptr-sptr, ",\"totalRead\":%u",
                        SystemData.TotalWebDavRead);

   if (SystemData.TotalWebDavWrite)
      sptr += snprintf (sptr, zptr-sptr, ",\"totalWrite\":%u",
                        SystemData.TotalWebDavWrite);

   if (SystemData.TotalWebDavLock)
      sptr += snprintf (sptr, zptr-sptr, ",\"totalLock\":%u",
                        SystemData.TotalWebDavLock);

   if (SystemData.TotalWebDavOther)
      sptr += snprintf (sptr, zptr-sptr, ",\"totalOther\":%u",
                        SystemData.TotalWebDavOther);

   if (sptr > zptr) EXIT_FI_LI (SS$_RESULTOVF);

   if (SystemData.TotalWebDav)
      sptr += snprintf (sptr, zptr-sptr, ",\"module\":%d",
                        SystemData.TotalWebDav);

   if (SystemData.WebDavRequestCount)
      sptr += snprintf (sptr, zptr-sptr, ",\"request\":%u",
                        SystemData.WebDavRequestCount);

   if (SystemData.WebDavMetaReadAttemptCount)
      sptr += snprintf (sptr, zptr-sptr, ",\"metaReadAttempt\":%u",
                        SystemData.WebDavMetaReadAttemptCount);

   if (SystemData.WebDavMetaReadCount)
      sptr += snprintf (sptr, zptr-sptr, ",\"metaRead\":%u",
                        SystemData.WebDavMetaReadCount);

   if (SystemData.WebDavMetaWriteAttemptCount)
      sptr += snprintf (sptr, zptr-sptr, ",\"metaWriteAttempt\":%u",
                        SystemData.WebDavMetaWriteAttemptCount);

   if (SystemData.WebDavMetaWriteCount)
      sptr += snprintf (sptr, zptr-sptr, ",\"metaWrite\":%u",
                        SystemData.WebDavMetaWriteCount);

   if (SystemData.WebDavXmlParseCount)
      sptr += snprintf (sptr, zptr-sptr, ",\"parseXML\":%u",
                        SystemData.WebDavXmlParseCount);

   if (sptr > zptr) EXIT_FI_LI (SS$_RESULTOVF);

   if (SystemData.WebDavAgentAppleCount)
      sptr += snprintf (sptr, zptr-sptr, ",\"agentApple\":%u",
                        SystemData.WebDavAgentAppleCount);

   if (SystemData.WebDavAgentBsdCount)
      sptr += snprintf (sptr, zptr-sptr, ",\"agentBsd\":%u",
                        SystemData.WebDavAgentBsdCount);

   if (SystemData.WebDavAgentLinuxCount)
      sptr += snprintf (sptr, zptr-sptr, ",\"agentLinux\":%u",
                        SystemData.WebDavAgentLinuxCount);

   if (SystemData.WebDavAgentMicrosoftCount)
      sptr += snprintf (sptr, zptr-sptr, ",\"agentMicrosoft\":%u",
                        SystemData.WebDavAgentMicrosoftCount);

   if (SystemData.WebDavAgentOtherCount)
      sptr += snprintf (sptr, zptr-sptr, ",\"agentOther\":%u",
                        SystemData.WebDavAgentOtherCount);

   if (SystemData.WebDavAgentSunCount)
      sptr += snprintf (sptr, zptr-sptr, ",\"agentSun\":%u",
                        SystemData.WebDavAgentSunCount);

   sptr += snprintf (sptr, zptr-sptr, ",\"agentTotal\":%u",
                     SystemData.WebDavAgentAppleCount +
                     SystemData.WebDavAgentBsdCount +
                     SystemData.WebDavAgentLinuxCount +
                     SystemData.WebDavAgentMicrosoftCount +
                     SystemData.WebDavAgentOtherCount +
                     SystemData.WebDavAgentSunCount);

   sptr += snprintf (sptr, zptr-sptr, ",\"bytesRawRx\":%s",
                     JsonNumber(64,&SystemData.WebDavBytesRawRx));
   sptr += snprintf (sptr, zptr-sptr, ",\"bytesRawTx\":%s",
                     JsonNumber(64,&SystemData.WebDavBytesRawTx));

   sptr += snprintf (sptr, zptr-sptr, "}");

   if (sptr >= zptr) EXIT_FI_LI (SS$_RESULTOVF);

   SystemData.WebDavBufferLength = sptr - SystemData.WebDavBuffer;
}

/*****************************************************************************/
/*
Create a JSON data structure in a buffer representing the instance (server)
process data.  This is an array 0..n of instance processes.
*/

void JsonInstanceData ()

{
   static char  CpuTime [32],
                UpTime [32];
   static $DESCRIPTOR (CpuTimeFaoDsc, "!%D\0");
   static $DESCRIPTOR (CpuTimeDsc, CpuTime);
   static $DESCRIPTOR (UpTimeFaoDsc, "!10%D\0");
   static $DESCRIPTOR (UpTimeDsc, UpTime);
   static PROCESS_DATA  ProcessData [INSTANCE_MAX];

   int  cnt, status;
   char  *cptr, *sptr, *zptr;
   PROCESS_DATA  *pdptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "JsonInstanceData()\n");

   zptr = (sptr = SystemData.InstanceBuffer) +
          sizeof(SystemData.InstanceBuffer);

   sptr += snprintf (sptr, zptr-sptr, "{\"$data\":\"instance\",\"instance\":[");

   for (cnt = 0; cnt < SystemData.InstanceNodeCount; cnt++)
   {
      pdptr = &ProcessData[cnt];

      for (cptr = (cnt ? ",{" : "{"); *cptr; *sptr++ = *cptr++);

      sptr += snprintf (sptr, zptr-sptr, "\"pid\":\"%08.08X\"",
                        SystemData.InstanceNodePid[cnt]);

      status = CollectProcessData (SystemData.InstanceNodePid[cnt], pdptr);

      if (VMSnok (status))
      {
         sptr += snprintf (sptr, zptr-sptr, ",\"status\":%d}", status);
         continue;
      }
                                                               
      sptr += snprintf (sptr, zptr-sptr, ",\"prcnam\":\"%s\"",
                        pdptr->JpiPrcNam);

      if (sptr > zptr) EXIT_FI_LI (SS$_RESULTOVF);

      sptr += snprintf (sptr, zptr-sptr, ",\"astcnt\":%d", pdptr->JpiAstCnt);
      sptr += snprintf (sptr, zptr-sptr, ",\"astlm\":%d", pdptr->JpiAstLm);
      sptr += snprintf (sptr, zptr-sptr, ",\"biocnt\":%d", pdptr->JpiBioCnt);
      sptr += snprintf (sptr, zptr-sptr, ",\"biolm\":%d", pdptr->JpiBioLm);
      sptr += snprintf (sptr, zptr-sptr, ",\"bytcnt\":%d", pdptr->JpiBytCnt);
      sptr += snprintf (sptr, zptr-sptr, ",\"bufio\":%d", pdptr->JpiBufIo);
      sptr += snprintf (sptr, zptr-sptr, ",\"bytlm\":%d", pdptr->JpiBytLm);
      if (pdptr->JpiCpuId >= 0)
         sptr += snprintf (sptr, zptr-sptr, ",\"cpuid\":%d", pdptr->JpiCpuId);
      sptr += snprintf (sptr, zptr-sptr, ",\"diocnt\":%d", pdptr->JpiDioCnt);
      sptr += snprintf (sptr, zptr-sptr, ",\"dirio\":%d", pdptr->JpiDirIo);
      sptr += snprintf (sptr, zptr-sptr, ",\"diolm\":%d", pdptr->JpiDioLm);
      sptr += snprintf (sptr, zptr-sptr, ",\"enqcnt\":%d", pdptr->JpiEnqCnt);
      sptr += snprintf (sptr, zptr-sptr, ",\"enqlm\":%d", pdptr->JpiEnqLm);
      sptr += snprintf (sptr, zptr-sptr, ",\"filcnt\":%d", pdptr->JpiFilCnt);
      sptr += snprintf (sptr, zptr-sptr, ",\"fillm\":%d", pdptr->JpiFilLm);

      if (sptr > zptr) EXIT_FI_LI (SS$_RESULTOVF);

      sptr += snprintf (sptr, zptr-sptr, ",\"gpgcnt\":%d", pdptr->JpiGpgCnt);
      sptr += snprintf (sptr, zptr-sptr, ",\"ktcount\":%d", pdptr->JpiKtCount);
      if (pdptr->JpiMultiThread)
         sptr += snprintf (sptr, zptr-sptr, ",\"multithread\":%d",
                           pdptr->JpiMultiThread);
      sptr += snprintf (sptr, zptr-sptr, ",\"pagfilcnt\":%d",
                        pdptr->JpiPagFilCnt);
      sptr += snprintf (sptr, zptr-sptr, ",\"pageflts\":%d",
                        pdptr->JpiPageFlts);
      sptr += snprintf (sptr, zptr-sptr, ",\"pgflquota\":%d",
                        pdptr->JpiPgFlQuota);
      sptr += snprintf (sptr, zptr-sptr, ",\"ppgcnt\":%d",
                        pdptr->JpiPpgCnt);
      sptr += snprintf (sptr, zptr-sptr, ",\"prccnt\":%d",
                        pdptr->JpiPrcCnt);
      sptr += snprintf (sptr, zptr-sptr, ",\"prclm\":%d",
                        pdptr->JpiPrcLm);

      if (sptr > zptr) EXIT_FI_LI (SS$_RESULTOVF);

      sptr += snprintf (sptr, zptr-sptr, ",\"astFlags\":\"%s\"",
                        pdptr->AstFlags);
      sptr += snprintf (sptr, zptr-sptr, ",\"pri\":%d", pdptr->JpiPri);
      sptr += snprintf (sptr, zptr-sptr, ",\"prib\":%d", pdptr->JpiPrib);
      sptr += snprintf (sptr, zptr-sptr, ",\"state\":\"%s\"",
                        pdptr->ProcessState);
      sptr += snprintf (sptr, zptr-sptr, ",\"tqlm\":%d", pdptr->JpiTqLm);
      sptr += snprintf (sptr, zptr-sptr, ",\"tqcnt\":%d", pdptr->JpiTqCnt);
      sptr += snprintf (sptr, zptr-sptr, ",\"wsextent\":%d",
                        pdptr->JpiWsExtent);
      sptr += snprintf (sptr, zptr-sptr, ",\"wspeak\":%d", pdptr->JpiWsPeak);
      sptr += snprintf (sptr, zptr-sptr, ",\"wsquota\":%d", pdptr->JpiWsQuota);
      sptr += snprintf (sptr, zptr-sptr, ",\"virtpeak\":%d",
                        pdptr->JpiVirtPeak);
      if (pdptr->JpiVolumes)
         sptr += snprintf (sptr, zptr-sptr, ",\"volumes\":%d",
                           pdptr->JpiVolumes);

      if (pdptr->PageFaults)
         sptr += snprintf (sptr, zptr-sptr, ",\"pageFaults\":%d",
                           pdptr->PageFaults);

      sys$fao (&CpuTimeFaoDsc, 0, &CpuTimeDsc, &pdptr->CpuBinTime);
      for (cptr = CpuTime;
           *cptr && (*cptr == ' ' || *cptr == ':' || *cptr == '0')
                 && (*cptr+1) != '.';
           cptr++);
      sptr += snprintf (sptr, zptr-sptr, ",\"cpuTime\":\"%s\"", cptr);

      sys$fao (&UpTimeFaoDsc, 0, &UpTimeDsc, &pdptr->ConBinTime);
      for (cptr = UpTime; *cptr && *cptr == ' '; cptr++);
      sptr += snprintf (sptr, zptr-sptr, ",\"conTime\":\"%s\"", cptr);

      sptr += snprintf (sptr, zptr-sptr, "}");
   }

   sptr += snprintf (sptr, zptr-sptr, "]}");

   if (sptr >= zptr) EXIT_FI_LI (SS$_RESULTOVF);

   SystemData.InstanceBufferLength = sptr - SystemData.InstanceBuffer;
}

/*****************************************************************************/
/*
Create a JSON data structure in a buffer representing the last request data.
*/

void JsonRequestData
(
CLIENT_DATA *clptr,
BOOL ForceUpdate
)
{
   static BOOL  LookItUp = true;
   static char  LastDuration[sizeof(gblptr->Request.Time)],
                LastTime[sizeof(gblptr->Request.Time)];

   int  cnt, status;
   char  *cptr, *sptr, *zptr;
   PROCESS_DATA  ProcessData;
   PROCESS_DATA  *pdptr = &ProcessData;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "JsonRequestData()\n");

   if (!ForceUpdate &&
       !LookItUp &&
       !strcmp (LastTime, gblptr->Request.Time) &&
       !strcmp (LastDuration, gblptr->Request.Duration))
   {
      /* only provide request data when it has changed */
      SystemData.RequestBuffer[SystemData.RequestBufferLength = 0] = '\0';
      return;
   }
   strcpy (LastTime, gblptr->Request.Time);
   strcpy (LastDuration, gblptr->Request.Duration);
   LookItUp = false;

   if (isdigit(gblptr->Request.ClientHostName[0]))
   {
      cptr = LookupHostName(gblptr->Request.ClientHostName);
      if (*cptr == '?')
      {
         LookItUp = true;
         cptr = NULL;
      }
      else
      if (*cptr == '!')
         cptr = NULL;
   }
   else
      cptr = NULL;

   SystemData.RequestBuffer[SystemData.RequestBufferLength = 0] = '\0';
   zptr = (sptr = SystemData.RequestBuffer) + sizeof(SystemData.RequestBuffer);

   sptr += snprintf (sptr, zptr-sptr, "{\"$data\":\"request\"");
   sptr += snprintf (sptr, zptr-sptr, ",\"alert\":%s",
                     gblptr->Request.Alert ? "true" : "false");
   sptr += snprintf (sptr, zptr-sptr, ",\"time\":\"%s\"",
                     gblptr->Request.Time);
   sptr += snprintf (sptr, zptr-sptr, ",\"client\":\"%s\"",
                     gblptr->Request.ClientHostName);

   if (cptr)
   {
      sptr += snprintf (sptr, zptr-sptr, ",\"client\":\"%s\"", cptr);
      sptr += snprintf (sptr, zptr-sptr, ",\"clientIP\":\"%s\"",
                        gblptr->Request.ClientHostName);
   }
   else
      sptr += snprintf (sptr, zptr-sptr, ",\"client\":\"%s\"",
                        gblptr->Request.ClientHostName);

   sptr += snprintf (sptr, zptr-sptr, ",\"service\":\"%s\"",
                     gblptr->Request.Service);
   sptr += snprintf (sptr, zptr-sptr, ",\"prcnam\":\"%s\"",
                     gblptr->Request.PrcNam);
   sptr += snprintf (sptr, zptr-sptr, ",\"method\":\"%s\"",
                     gblptr->Request.MethodName);

   /* URI has potential to have unescaped reserved characters */
   for (cptr = ",\"uri\":\""; *cptr; *sptr++ = *cptr++);
   if (clptr && clptr->WorldAccess && gblptr->Request.AuthUser[0])
      for (cptr = gblptr->Request.Uri; *cptr && sptr < zptr; cptr++)
         *sptr++ = '*';
   else
      for (cptr = gblptr->Request.Uri; *cptr && sptr < zptr; *sptr++ = *cptr++)
         if (*cptr == '\"' || *cptr == '\\') *sptr++ = '\\';
   if (cptr - gblptr->Request.Uri == sizeof(gblptr->Request.Uri)-1)
      for (cptr = "..."; *cptr; *sptr++ = *cptr++);
   *sptr++ = '\"';

   sptr += snprintf (sptr, zptr-sptr, ",\"status\":%d",
                     gblptr->Request.HttpStatus);
   sptr += snprintf (sptr, zptr-sptr, ",\"rx\":%s",
                     JsonNumber(64,&gblptr->Request.BytesRawRx));
   sptr += snprintf (sptr, zptr-sptr, ",\"tx\":%s",
                     JsonNumber(64,&gblptr->Request.BytesRawTx));
   sptr += snprintf (sptr, zptr-sptr, ",\"bps\":%d",
                     gblptr->Request.BytesPerSecond);
   sptr += snprintf (sptr, zptr-sptr, ",\"gzipTx\":%d",
                     gblptr->Request.BytesTxGzipPercent);
   sptr += snprintf (sptr, zptr-sptr, ",\"duration\":\"%s\"",
                     gblptr->Request.Duration);

   if (gblptr->Request.AuthUser[0])
      if (clptr && clptr->WorldAccess)
         sptr += snprintf (sptr, zptr-sptr, ",\"user\":\"********\"");
      else
         sptr += snprintf (sptr, zptr-sptr, ",\"user\":\"%s\"",
                           gblptr->Request.AuthUser);

   sptr += snprintf (sptr, zptr-sptr, "}");

   if (sptr >= zptr) EXIT_FI_LI (SS$_RESULTOVF);

   SystemData.RequestBufferLength = sptr - SystemData.RequestBuffer;
}

/*****************************************************************************/
/*
Create a JSON data structure in a buffer representing the scripting data.
*/

void JsonScriptData ()

{
   char  *cptr, *sptr, *zptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "JsonScriptData()\n");

   SystemData.ScriptBuffer[SystemData.ScriptBufferLength = 0] = '\0';
   zptr = (sptr = SystemData.ScriptBuffer) + sizeof(SystemData.ScriptBuffer);

   sptr += snprintf (sptr, zptr-sptr, "{\"$data\":\"script\"");

   if (SystemData.CurrentScript)
      sptr += snprintf (sptr, zptr-sptr, ",\"scriptNow\":%d",
                        SystemData.CurrentScript);
   if (SystemData.CurrentScriptCGIplus)
      sptr += snprintf (sptr, zptr-sptr, ",\"scriptCGIplusNow\":%d",
                        SystemData.CurrentScriptCGIplus);
   if (SystemData.CurrentScriptRTE)
      sptr += snprintf (sptr, zptr-sptr, ",\"scriptRTENow\":%d",
                        SystemData.CurrentScriptRTE);
   if (SystemData.TotalDclHardLimit)
      sptr += snprintf (sptr, zptr-sptr, ",\"scriptHardLimit\":%d",
                        SystemData.TotalDclHardLimit);

   if (SystemData.TotalCgi)
      sptr += snprintf (sptr, zptr-sptr, ",\"totalCGI\":%d",
                        SystemData.TotalCgi);
   if (SystemData.TotalDclCommand)
      sptr += snprintf (sptr, zptr-sptr, ",\"totalCLI\":%d",
                        SystemData.TotalDclCommand);

   if (SystemData.TotalCgiPlus)
      sptr += snprintf (sptr, zptr-sptr, ",\"totalCGIplus\":%d",
                        SystemData.TotalCgiPlus);
   if (SystemData.TotalCgiPlus)
      sptr += snprintf (sptr, zptr-sptr, ",\"totalCGIplusReused\":%d",
                        SystemData.TotalCgiPlusReused);

   if (SystemData.TotalRte)
      sptr += snprintf (sptr, zptr-sptr, ",\"totalRTE\":%d",
                        SystemData.TotalRte);
   if (SystemData.TotalRteReused)
      sptr += snprintf (sptr, zptr-sptr, ",\"totalRTEreused\":%d",
                        SystemData.TotalRteReused);

   if (SystemData.CurrentDECnetTasks)
      sptr += snprintf (sptr, zptr-sptr, ",\"DECnetNow\":%d",
                        SystemData.CurrentDECnetTasks);
   if (SystemData.CurrentDECnetCGI)
      sptr += snprintf (sptr, zptr-sptr, ",\"DECnetCGINow\":%d",
                        SystemData.CurrentDECnetCGI);
   if (SystemData.CurrentDECnetOSU)
      sptr += snprintf (sptr, zptr-sptr, ",\"DECnetOSUNow\":%d",
                        SystemData.CurrentDECnetOSU);

   if (SystemData.TotalDECnetCgi)
      sptr += snprintf (sptr, zptr-sptr, ",\"totalDECnetCGI\":%d",
                        SystemData.TotalDECnetCgi);
   if (SystemData.TotalDECnetCgiReused)
      sptr += snprintf (sptr, zptr-sptr, ",\"totalDECnetCGIreused\":%d",
                        SystemData.TotalDECnetCgiReused);

   if (SystemData.TotalDECnetOsu)
      sptr += snprintf (sptr, zptr-sptr, ",\"totalDECnetOSU\":%d",
                        SystemData.TotalDECnetOsu);
   if (SystemData.TotalDECnetOsuReused)
      sptr += snprintf (sptr, zptr-sptr, ",\"totalDECnetOSUreused\":%d",
                        SystemData.TotalDECnetOsuReused);

   sptr += snprintf (sptr, zptr-sptr, "}");

   if (sptr >= zptr) EXIT_FI_LI (SS$_RESULTOVF);

   SystemData.ScriptBufferLength = sptr - SystemData.ScriptBuffer;
}

/*****************************************************************************/
/*
Get the IP name using asynchronous address-to-name lookup.  Despite the implied
IPv6 functionality TCP/IP Services lookup only supports IPv4 (sigh, one day
perhaps).  There will always be a delay in the lookup because this function is 
called during AST delivery, so buffer the IP addresses and host names.  Returns
a pointer to the host name, to '?' during lookup, or '!' if not resolvable.
*/

char* LookupHostName (ulong arg1)
{
#define CACHE_MAX 128  /* absolute max is 1024 */
#define EXPIRE_SECONDS 300
#define RETRY_ATTEMPTS   5
#define RETRY_SECONDS   60

#define INETACP$C_TRANS 2
#define INETACP_FUNC$C_GETHOSTBYNAME 1
#define INETACP_FUNC$C_GETHOSTBYADDR 2

   struct LookupCacheStruct
   {
       int  RetryCount;
       ushort  HostNameLength;
       ulong  ExpiresStamp;
       ulong  Ip4Address;
       ulong  Ip6Address [4];
       char  HostName [127+1],
             IpAddrStr [31+1];
       struct dsc$descriptor HostAddressDsc;
       struct dsc$descriptor HostNameDsc;
       IOSBLK  LookupIOsb;
   };
   static struct LookupCacheStruct  LookupCache [CACHE_MAX];

   static $DESCRIPTOR (TcpIpDeviceDsc, "UCX$DEVICE");
   static unsigned char ControlSubFunction [4] =
      { INETACP_FUNC$C_GETHOSTBYADDR, INETACP$C_TRANS, 0, 0 };
   static struct dsc$descriptor ControlSubFunctionDsc =
      { 4, DSC$K_DTYPE_T, DSC$K_CLASS_S, (char*)&ControlSubFunction };

   static int  CacheUsed;
   static unsigned short  LookupChannel;

   int  idx, status;
   ulong  ThisTimeStamp;
   ulong  ThisBinTime [2];
   char  *cptr, *sptr, *zptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "LookupHostName() %d\n", arg1);

   if (!LookupChannel)
   {
      /* assign a channel to the internet template device */
      status = sys$assign (&TcpIpDeviceDsc, &LookupChannel, 0, 0);
      if (VMSnok (status)) EXIT_FI_LI (status);
      /* initialise the descriptors */
      for (idx = 0; idx < CACHE_MAX; idx++)
      {
         LookupCache[idx].HostNameDsc.dsc$b_class = DSC$K_CLASS_S;
         LookupCache[idx].HostNameDsc.dsc$b_dtype = DSC$K_DTYPE_T;
         LookupCache[idx].HostNameDsc.dsc$w_length =
            sizeof(LookupCache[idx].HostName)-1;
         LookupCache[idx].HostNameDsc.dsc$a_pointer =
            LookupCache[idx].HostName;
         LookupCache[idx].HostAddressDsc.dsc$b_class = DSC$K_CLASS_S;
         LookupCache[idx].HostAddressDsc.dsc$b_dtype = DSC$K_DTYPE_T;
      }
   }

   if (arg1 > 1023)
   {
      /**************************/
      /* lookup this IP address */
      /**************************/

      sys$gettim (&ThisBinTime);
      ThisTimeStamp = decc$fix_time (&ThisBinTime);

      cptr = (char*)arg1;
      if (Debug) fprintf (stdout, "->|%s|\n", cptr);

      /* search the cache */
      for (idx = 0; idx < CacheUsed; idx++)
      {
         if (!MATCH8 (cptr, LookupCache[idx].IpAddrStr)) continue;
         if (strcmp (cptr, LookupCache[idx].IpAddrStr)) continue;
         if (Debug) fprintf (stdout, "|%s|->\n", LookupCache[idx].HostName);
         return (LookupCache[idx].HostName);
      }
      /* look for an unused or expired entry */
      for (idx = 0; idx < CACHE_MAX; idx++)
      {
         if (!LookupCache[idx].IpAddrStr[0]) break;
         if (ThisTimeStamp >= LookupCache[idx].ExpiresStamp) break;
      }
      /* otherwise look for an unresolveable entry */
      if (idx >= CACHE_MAX)
         for (idx = 0; idx < CACHE_MAX; idx++)
            if (LookupCache[idx].IpAddrStr[0] == '!') break;
      /* otherwise just cycle through the cache array entries */
      if (idx >= CACHE_MAX) idx = 0;
      if (idx > CacheUsed) CacheUsed = idx;
      if (Debug) fprintf (stdout, "idx %d %d\n", idx, CacheUsed);

      LookupCache[idx].IpAddrStr[0] = LookupCache[idx].HostName[1] = '\0';

      if (strchr (cptr, ':'))
      {
         if (inet_pton (AF_INET6, cptr, &LookupCache[idx].Ip6Address) != 1)
         {
            LookupCache[idx].HostName[0] = '!';
            LookupCache[idx].ExpiresStamp = ThisTimeStamp + RETRY_SECONDS;
            return (LookupCache[idx].HostName);
         }
         LookupCache[idx].Ip4Address = 0;
      }
      else
      if (inet_pton (AF_INET, cptr, &LookupCache[idx].Ip4Address) != 1)
      {
         LookupCache[idx].HostName[0] = '!';
         LookupCache[idx].ExpiresStamp = ThisTimeStamp + RETRY_SECONDS;
         return (LookupCache[idx].HostName);
      }

      LookupCache[idx].HostName[0] = '?';
      LookupCache[idx].ExpiresStamp = ThisTimeStamp + EXPIRE_SECONDS;

      zptr = (sptr = LookupCache[idx].IpAddrStr) +
              sizeof(LookupCache[idx].IpAddrStr)-1;
      while (*cptr && sptr < zptr) *sptr++ = *cptr++;
      *sptr = '\0';

      if (LookupCache[idx].Ip4Address)
      {
         LookupCache[idx].HostAddressDsc.dsc$w_length =
            sizeof(LookupCache[idx].Ip4Address);
         LookupCache[idx].HostAddressDsc.dsc$a_pointer =
            (char*)&LookupCache[idx].Ip4Address;
      }
      else
      {
         LookupCache[idx].HostAddressDsc.dsc$w_length =
            sizeof(LookupCache[idx].Ip6Address);
         LookupCache[idx].HostAddressDsc.dsc$a_pointer =
            (char*)&LookupCache[idx].Ip6Address;
      }
      LookupCache[idx].RetryCount = RETRY_ATTEMPTS;
   }
   else
   {
      /**************/
      /* lookup AST */
      /**************/

      idx = arg1;
      status = LookupCache[idx].LookupIOsb.iosb$w_status;

      if (Debug) fprintf (stdout, "sys$qio() %%X%08.08X\n", status);

      if (VMSok (status)) 
      {
         LookupCache[idx].HostName[LookupCache[idx].HostNameLength] = '\0';
         if (Debug) fprintf (stdout, "|%s|\n", LookupCache[idx].HostName);
         if (!LookupCache[idx].HostName[0])
         {
            LookupCache[idx].HostName[0] = '!';
            LookupCache[idx].HostName[1] = '\0';
         }
         return (LookupCache[idx].HostName);
      }

      if (status == SS$_ENDOFFILE ||
         !LookupCache[idx].RetryCount--)
      {
         sys$gettim (&ThisBinTime);
         ThisTimeStamp = decc$fix_time (&ThisBinTime);
         LookupCache[idx].ExpiresStamp = ThisTimeStamp + RETRY_SECONDS;
         LookupCache[idx].HostName[0] = '!';
         LookupCache[idx].HostName[1] = '\0';
         return (LookupCache[idx].HostName);
      }
      LookupCache[idx].RetryCount--;
   }

   status = sys$qio (0, LookupChannel, IO$_ACPCONTROL | IO$M_EXTEND,
                     &LookupCache[idx].LookupIOsb,
                     LookupHostName, idx,
                     &ControlSubFunctionDsc,
                     &LookupCache[idx].HostAddressDsc,
                     &LookupCache[idx].HostNameLength,
                     &LookupCache[idx].HostNameDsc, 0, 0);
   if (Debug) fprintf (stdout, "[%d] sys$qio() %%X%08.08X\n", idx, status);

   if (VMSnok(status)) LookupCache[idx].IpAddrStr[0] = '\0';

   return (LookupCache[idx].HostName);
}

/*****************************************************************************/
/*
Server process accounting data.
*/ 
void CollectAccounting ()

{
   static ulong  MemSize,
                 PageSize,
                 PrevFileCount,
                 PrevFile304Count,
                 PrevOtherCount,
                 PrevProxyCacheRead,
                 PrevProxyCacheWrite,
                 PrevProxyNetwork,
                 PrevProxyTotal,
                 PrevRequestAllCount,
                 PrevRequestFailCount,
                 PrevScriptCgiCount,
                 PrevThrottleBusy,
                 PrevThrottled,
                 PrevThrottleQueued,
                 PrevTotalScript,
                 PrevWebDavRead,
                 PrevWebDavTotal,
                 PrevWebDavWrite;

   static ulong  BytesRawRxTx [2],
                 PrevBlocksRawRx [2],
                 PrevBlocksRawTx [2],
                 PrevBytesRawRx [2],
                 PrevBytesRawTx [2],
                 PrevBytesRawRxTx [2];

   int  idx, status, total;
   ulong  TotalScript,
          RequestAllCount,
          RequestFailCount,
          ScriptCgiCount;
   ulong  quad [2],
          BlocksRawRx [2],
          BlocksRawTx [2],
          BytesRawRx [2],
          BytesRawTx [2];
   char  *cptr, *sptr, *zptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CollectAccounting()\n");

   SystemData.InstanceStartupCount = accptr->StartupCount;

   for (total = idx = 0; idx <= gblptr->InstanceNodeCurrent; idx++)
      total += accptr->CurrentConnected[idx];
   SystemData.CurrentConnected = total;
   SystemData.ConnectPeak = accptr->ConnectPeak;
   for (total = idx = 0; idx <= gblptr->InstanceNodeCurrent; idx++)
      total += accptr->CurrentProcessing[idx];
   SystemData.CurrentProcessing = total;

   SystemData.ProcessingPeak = accptr->CurrentProcessingPeak;
   SystemData.SslCount = accptr->ConnectSslCount;
   SystemData.TooBusy = accptr->ConnectTooBusyCount;
   for (total = idx = 0; idx <= gblptr->InstanceNodeCurrent; idx++)
      total += accptr->CurrentWebSockets[idx];
   SystemData.CurrentWebSockets = total;

   /****************/
   /* HTTP methods */
   /****************/

   SystemData.MethodConnect = accptr->MethodConnectCount;
   SystemData.MethodCopy = accptr->MethodWebDavCopyCount;
   SystemData.MethodDelete = accptr->MethodDeleteCount;
   SystemData.MethodGet = accptr->MethodGetCount;
   SystemData.MethodHead = accptr->MethodHeadCount;
   SystemData.MethodLock = accptr->MethodWebDavLockCount;
   SystemData.MethodMkCol = accptr->MethodWebDavMkColCount;
   SystemData.MethodMove = accptr->MethodWebDavMoveCount;
   SystemData.MethodPropFind = accptr->MethodWebDavPropFindCount;
   SystemData.MethodPropPatch = accptr->MethodWebDavPropPatchCount;
   SystemData.MethodOptions = accptr->MethodOptionsCount;
   SystemData.MethodPost = accptr->MethodPostCount;
   SystemData.MethodPut = accptr->MethodPutCount;
   SystemData.MethodSsh = accptr->MethodSshCount;
   SystemData.MethodTrace = accptr->MethodTraceCount;
   SystemData.MethodUnLock = accptr->MethodWebDavUnLockCount;

   /*******************/
   /* response status */
   /*******************/

   SystemData.StatusCodeGroup[0] = accptr->ResponseStatusCodeGroup[0];
   SystemData.StatusCodeGroup[1] = accptr->ResponseStatusCodeGroup[1];
   SystemData.StatusCodeGroup[2] = accptr->ResponseStatusCodeGroup[2];
   SystemData.StatusCodeGroup[3] = accptr->ResponseStatusCodeGroup[3];
   SystemData.StatusCodeGroup[4] = accptr->ResponseStatusCodeGroup[4];
   SystemData.StatusCodeGroup[5] = accptr->ResponseStatusCodeGroup[5];
   SystemData.StatusCode403 = accptr->RequestForbiddenCount;

   RequestAllCount = SystemData.StatusCodeGroup[0] +
                     SystemData.StatusCodeGroup[1] +
                     SystemData.StatusCodeGroup[2] +
                     SystemData.StatusCodeGroup[3] +
                     SystemData.StatusCodeGroup[4] +
                     SystemData.StatusCodeGroup[5];

   RequestFailCount = SystemData.StatusCodeGroup[0] +
                      SystemData.StatusCodeGroup[4] +
                      SystemData.StatusCodeGroup[5];

   SystemData.TotalRequest = RequestAllCount;

   if (PrevRequestAllCount && RequestAllCount > PrevRequestAllCount)
      SystemData.DeltaRequest = RequestAllCount - PrevRequestAllCount;
   else
      SystemData.DeltaRequest = 0;
   PrevRequestAllCount = RequestAllCount;

   if (PrevRequestFailCount && RequestFailCount > PrevRequestFailCount)
      SystemData.DeltaRequestFail = RequestFailCount - PrevRequestFailCount;
   else
      SystemData.DeltaRequestFail = 0;
   PrevRequestFailCount = RequestFailCount;

   /***************/
   /* bytes rx/tx */
   /***************/

   PUT_QUAD_QUAD (&accptr->BlocksRawRx, &SystemData.BlocksRawRx);
   PUT_QUAD_QUAD (&accptr->BlocksRawTx, &SystemData.BlocksRawTx);
   PUT_QUAD_QUAD (&accptr->BytesRawRx, &SystemData.BytesRawRx);
   PUT_QUAD_QUAD (&accptr->BytesRawTx, &SystemData.BytesRawTx);

   PUT_QUAD_QUAD (&SystemData.BytesRawRx, &BytesRawRxTx);
   ADD_QUAD_QUAD (&SystemData.BytesRawTx, &BytesRawRxTx);
   ADD_QUAD_QUAD (&SystemData.ProxyBytesRawRx, &BytesRawRxTx);
   ADD_QUAD_QUAD (&SystemData.ProxyBytesRawTx, &BytesRawRxTx);
   PUT_QUAD_QUAD (&BytesRawRxTx, &quad);
   SUB_QUAD_QUAD (&PrevBytesRawRxTx, &quad);
   PUT_QUAD_QUAD (&BytesRawRxTx, &PrevBytesRawRxTx);
   SystemData.BytesRawPerSec = (ulong)(FLOAT_FROM_QUAD(&quad) / DeltaSeconds);

   PUT_QUAD_QUAD (&SystemData.BytesRawRx, &BytesRawRx);
   SUB_QUAD_QUAD (&PrevBytesRawRx, &BytesRawRx);
   PUT_QUAD_QUAD (&SystemData.BytesRawRx, &PrevBytesRawRx);

   PUT_QUAD_QUAD (&SystemData.BlocksRawRx, &BlocksRawRx);
   SUB_QUAD_QUAD (&PrevBlocksRawRx, &BlocksRawRx);
   PUT_QUAD_QUAD (&SystemData.BlocksRawRx, &PrevBlocksRawRx);
   if (QUAD_NOT_ZERO(&BlocksRawRx))
      SystemData.BlockReadSize =
       (ulong)(FLOAT_FROM_QUAD(&BytesRawRx) /
               FLOAT_FROM_QUAD(&BlocksRawRx) /
               DeltaSeconds);
   else
      SystemData.BlockReadSize = 0;

   PUT_QUAD_QUAD (&SystemData.BytesRawTx, &BytesRawTx);
   SUB_QUAD_QUAD (&PrevBytesRawTx, &BytesRawTx);
   PUT_QUAD_QUAD (&SystemData.BytesRawTx, &PrevBytesRawTx);

   PUT_QUAD_QUAD (&SystemData.BlocksRawTx, &BlocksRawTx);
   SUB_QUAD_QUAD (&PrevBlocksRawTx, &BlocksRawTx);
   PUT_QUAD_QUAD (&SystemData.BlocksRawTx, &PrevBlocksRawTx);
   if (QUAD_NOT_ZERO(&BlocksRawTx))
      SystemData.BlockWriteSize =
       (ulong)(FLOAT_FROM_QUAD(&BytesRawTx) /
               FLOAT_FROM_QUAD(&BlocksRawTx) /
               DeltaSeconds);
   else
      SystemData.BlockWriteSize = 0;

   SystemData.NetReadErr = accptr->NetReadErrorCount;
   SystemData.NetWriteErr = accptr->NetWriteErrorCount;

   SystemData.BytesPerSecAve = accptr->BytesPerSecondAve;
   SystemData.BytesPerSecMax = accptr->BytesPerSecondMax;
   SystemData.BytesPerSecMin = accptr->BytesPerSecondMin;

   /****************/
   /* basic totals */
   /****************/

   SystemData.TotalAdmin = accptr->DoServerAdminCount;
   SystemData.TotalDECnet = accptr->DoDECnetCount;
   SystemData.TotalDir = accptr->DoDirectoryCount;
   SystemData.TotalFile = accptr->DoFileCount;
   SystemData.TotalFile304 = accptr->DoFileNotModifiedCount;
   SystemData.TotalIsMap = accptr->DoIsMapCount;
   SystemData.TotalMenu = accptr->DoMenuCount;
   SystemData.TotalProxy = accptr->DoProxyCount;
   SystemData.TotalPut = accptr->DoPutCount;
   SystemData.TotalSsi = accptr->DoSsiCount;
   SystemData.TotalUpdate = accptr->DoUpdateCount;
   SystemData.TotalWebSocket = accptr->DclWebSocketCount;

   /****************/
   /* basic deltas */
   /****************/

   SystemData.DeltaFile = SystemData.TotalFile;
   if (PrevFileCount && SystemData.TotalFile > PrevFileCount)
      SystemData.DeltaFile -= PrevFileCount;
   else
      SystemData.DeltaFile = 0;
   PrevFileCount = SystemData.TotalFile;

   SystemData.DeltaFile304 = SystemData.TotalFile304;
   if (PrevFile304Count && SystemData.TotalFile304 > PrevFile304Count)
      SystemData.DeltaFile304 -= PrevFile304Count;
   else
      SystemData.DeltaFile304 = 0;
   PrevFile304Count = SystemData.TotalFile304;

   /* accptr->DoDclCommandCount is a special "script" for "internal" purposes */
   TotalScript = accptr->DoDECnetCount +
                 accptr->DoCgiPlusScriptCount +
                 accptr->DoRteScriptCount +
                 accptr->DoScriptCount;
   SystemData.TotalAllScript = TotalScript;
   if (PrevTotalScript && TotalScript > PrevTotalScript)
      SystemData.DeltaAllScript = TotalScript - PrevTotalScript;
   else
      SystemData.DeltaAllScript = 0;
   PrevTotalScript = TotalScript;

   SystemData.TotalDclScript = TotalScript - accptr->DoDECnetCount;

   ScriptCgiCount = accptr->DoDclCommandCount +
                    accptr->DoDECnetCount +
                    accptr->DoScriptCount;
   SystemData.DeltaScriptCgi = ScriptCgiCount;
   if (PrevScriptCgiCount && ScriptCgiCount > PrevScriptCgiCount)
      SystemData.DeltaScriptCgi = ScriptCgiCount - PrevScriptCgiCount;
   else
      SystemData.DeltaScriptCgi = 0;
   PrevScriptCgiCount = ScriptCgiCount;

   /*********/
   /* proxy */
   /*********/

   if (SystemData.ProxyEnabled)
   {
      SystemData.ProxyNetwork = pacptr->ConnectIpv4Count +
                                pacptr->ConnectIpv6Count;
      SystemData.ProxyCacheRead = pacptr->CacheReadCount;
      SystemData.ProxyCacheRead304 = pacptr->CacheRead304Count;
      SystemData.ProxyCacheWrite = pacptr->CacheWriteCount;
      SystemData.ProxyTotal = SystemData.ProxyNetwork +
                              SystemData.ProxyCacheRead;

      if (PrevProxyTotal && SystemData.ProxyTotal > PrevProxyTotal)
         SystemData.ProxyDeltaTotal = SystemData.ProxyTotal - PrevProxyTotal; 
      else
         SystemData.ProxyDeltaTotal = 0;
      PrevProxyTotal = SystemData.ProxyTotal;

      if (PrevProxyCacheRead && SystemData.ProxyCacheRead > PrevProxyCacheRead)
         SystemData.ProxyDeltaCacheRead = SystemData.ProxyCacheRead -
                                          PrevProxyCacheRead; 
      else
         SystemData.ProxyDeltaCacheRead = 0;
      PrevProxyCacheRead = SystemData.ProxyCacheRead;

      if (PrevProxyCacheWrite &&
          SystemData.ProxyCacheWrite > PrevProxyCacheWrite)
         SystemData.ProxyDeltaCacheWrite = SystemData.ProxyCacheWrite -
                                           PrevProxyCacheWrite; 
      else
         SystemData.ProxyDeltaCacheWrite = 0;
      PrevProxyCacheWrite = SystemData.ProxyCacheWrite;

      if (PrevProxyNetwork && SystemData.ProxyNetwork > PrevProxyNetwork)
         SystemData.ProxyDeltaNetwork = SystemData.ProxyNetwork -
                                        PrevProxyNetwork; 
      else
         SystemData.ProxyDeltaNetwork = 0;
      PrevProxyNetwork = SystemData.ProxyNetwork;
   }

   /**********/
   /* WebDAV */
   /**********/

   if (SystemData.WebDavEnabled)
   {
      SystemData.TotalWebDavRead =  accptr->MethodWebDav_GetCount;

      SystemData.TotalWebDavWrite = accptr->MethodWebDavCopyCount +
                                    accptr->MethodWebDav_DeleteCount +
                                    accptr->MethodWebDavMkColCount +
                                    accptr->MethodWebDavMoveCount +
                                    accptr->MethodWebDav_PutCount;

      SystemData.TotalWebDavLock = accptr->MethodWebDavLockCount +
                                   accptr->MethodWebDavUnLockCount;

      SystemData.TotalWebDavOther = accptr->MethodWebDav_OptionsCount +
                                    accptr->MethodWebDavPropFindCount +
                                    accptr->MethodWebDavPropPatchCount;

      SystemData.TotalWebDav = SystemData.TotalWebDavLock +
                               SystemData.TotalWebDavOther +
                               SystemData.TotalWebDavRead +
                               SystemData.TotalWebDavWrite;

      if (PrevWebDavTotal && SystemData.TotalWebDav > PrevWebDavTotal)
         SystemData.DeltaWebDavTotal = SystemData.TotalWebDav - PrevWebDavTotal;
      else
         SystemData.DeltaWebDavTotal = 0;
      PrevWebDavTotal = SystemData.TotalWebDav;

      if (PrevWebDavRead && SystemData.TotalWebDavRead > PrevWebDavRead)
         SystemData.DeltaWebDavRead = SystemData.TotalWebDavRead -
                                      PrevWebDavRead;
      else
         SystemData.DeltaWebDavRead = 0;
      PrevWebDavRead = SystemData.TotalWebDavRead;

      if (PrevWebDavWrite && SystemData.TotalWebDavWrite > PrevWebDavWrite)
         SystemData.DeltaWebDavWrite = SystemData.TotalWebDavWrite -
                                       PrevWebDavWrite;
      else
         SystemData.DeltaWebDavWrite = 0;
      PrevWebDavWrite = SystemData.TotalWebDavWrite;
   }

   /************/
   /* throttle */
   /************/

   for (total = idx = 0; idx <= gblptr->InstanceNodeCurrent; idx++)
      total += accptr->CurrentThrottleProcessing[idx];
   SystemData.CurrentThrottleProcessing = total;

   for (total = idx = 0; idx <= gblptr->InstanceNodeCurrent; idx++)
      total += accptr->CurrentThrottleQueued [idx];
   SystemData.CurrentThrottleQueued = total;

   SystemData.CurrentThrottled = SystemData.CurrentThrottleProcessing + total;

   /* throttled is initially processing plus initially queued plus init busy */
   SystemData.TotalThrottled = accptr->ThrottleTotalProcessed +
                               accptr->ThrottleTotalQueued +
                               accptr->ThrottleTotalBusy;
   if (PrevThrottled && SystemData.TotalThrottled > PrevThrottled)
      SystemData.DeltaThrottled = SystemData.TotalThrottled - PrevThrottled;
   else
      SystemData.DeltaThrottled = 0;
   PrevThrottled = SystemData.TotalThrottled;

   SystemData.TotalThrottleQueued = accptr->ThrottleTotalQueued;
   if (PrevThrottleQueued &&
       SystemData.TotalThrottleQueued > PrevThrottleQueued)
      SystemData.DeltaThrottleQueued = SystemData.TotalThrottleQueued -
                                       PrevThrottleQueued;
   else
      SystemData.DeltaThrottleQueued = 0;
   PrevThrottleQueued = SystemData.TotalThrottleQueued;

   SystemData.CurrentThrottleBusyMetric = accptr->ThrottleBusyMetric;
   SystemData.TotalThrottleBusy = accptr->ThrottleTotalBusy;
   if (PrevThrottleBusy && SystemData.TotalThrottleBusy > PrevThrottleBusy)
      SystemData.DeltaThrottleBusy = SystemData.TotalThrottleBusy -
                                     PrevThrottleBusy;
   else
      SystemData.DeltaThrottleBusy = 0;
   PrevThrottleBusy = SystemData.TotalThrottleBusy;

   SystemData.DclProcessCurrent =
      SystemData.CurrentDECnetTasks =
      SystemData.CurrentDECnetCGI =
      SystemData.CurrentDECnetOSU = 0;
   for (idx = 0; idx <= gblptr->InstanceNodeCurrent; idx++)
   {
      SystemData.DclProcessCurrent += accptr->CurrentDclScriptProcess[idx];
      SystemData.CurrentDECnetTasks += accptr->CurrentDECnetTasks[idx];
      SystemData.CurrentDECnetCGI += accptr->CurrentDECnetCGI[idx];
      SystemData.CurrentDECnetOSU += accptr->CurrentDECnetOSU[idx];
   }

   SystemData.TotalNoModule = accptr->DoNoModuleCount;

   SystemData.TotalAllOther = SystemData.TotalRequest -
                              SystemData.TotalAllScript -
                              SystemData.TotalFile -
                              SystemData.TotalProxy;

   if (PrevOtherCount && SystemData.TotalAllOther > PrevOtherCount)
      SystemData.DeltaOther = SystemData.TotalAllOther - PrevOtherCount; 
   else
      SystemData.DeltaOther = 0;
   PrevOtherCount = SystemData.TotalAllOther;
} 

/*****************************************************************************/
/*
Collect proxy data.
*/ 

void CollectProxyData ()

{
   int  in, out;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CollectProxyData()\n");

   if (!SystemData.ProxyEnabled) return;

   SystemData.ProxyLookupCache = accptr->LookupCacheNameCount;
   SystemData.ProxyLookupDNS = accptr->LookupDnsNameCount;
   SystemData.ProxyLookupLiteral = accptr->LookupLiteralCount;
   SystemData.ProxyLookupError = accptr->LookupDnsNameErrorCount;

   SystemData.ProxyRequestNoCache = pacptr->RequestCannotCacheCount;
   SystemData.ProxyResponseNoCache = pacptr->ResponseCannotCacheCount;

   PUT_QUAD_QUAD (pacptr->BlocksRawRx, SystemData.ProxyBlocksRawRx);
   PUT_QUAD_QUAD (pacptr->BlocksRawTx, SystemData.ProxyBlocksRawTx);

   PUT_QUAD_QUAD (pacptr->BytesRawRx, SystemData.ProxyBytesRawRx);
   PUT_QUAD_QUAD (pacptr->BytesRawTx, SystemData.ProxyBytesRawTx);
   PUT_QUAD_QUAD (pacptr->BytesCacheRx, SystemData.ProxyBytesCacheRx);
   PUT_QUAD_QUAD (pacptr->BytesCacheTx, SystemData.ProxyBytesCacheTx);
   PUT_QUAD_QUAD (pacptr->BytesCannotCacheRx, SystemData.ProxyBytesNoCacheRx);
   PUT_QUAD_QUAD (pacptr->BytesCannotCacheTx, SystemData.ProxyBytesNoCacheTx);

   PLUS_QUAD_QUAD (SystemData.ProxyBytesRawRx,
                   SystemData.ProxyBytesRawTx,
                   SystemData.ProxyBytesRawRxTx);
   PLUS_QUAD_QUAD (SystemData.ProxyBytesCacheRx,
                   SystemData.ProxyBytesCacheTx,
                   SystemData.ProxyBytesCacheRxTx);
   PLUS_QUAD_QUAD (SystemData.ProxyBytesRawRxTx,
                   SystemData.ProxyBytesCacheRxTx,
                   SystemData.ProxyBytesRxTx);
   PLUS_QUAD_QUAD (SystemData.ProxyBytesNoCacheRx,
                   SystemData.ProxyBytesNoCacheTx,
                   SystemData.ProxyBytesNoCacheRxTx);

   SystemData.ProxyTunnel = 0;
   for (out = 0; out < TUNNEL_COUNT_MAX; out++)
   {
      if (out+1 == PROXY_TUNNEL_FIREWALL) continue;
      for (in = 0; in < TUNNEL_COUNT_MAX; in++)
      {
         if (in+1 == PROXY_TUNNEL_HTTP || in+1 == PROXY_TUNNEL_HTTPS) continue;
         SystemData.ProxyTunnel += pacptr->TunnelCount[in][out];
      }
   }
   SystemData.ProxyTunnelCurrent = pacptr->TunnelCurrent;

   SystemData.ProxyCacheEnabled = pacptr->CacheEnabled;
   SystemData.ProxyCacheSpace = pacptr->FreeSpaceAvailable;

   CollectProxyDevice ();
} 

/****************************************************************************/
/*
Return information about file system space on the proxy cache device.  Will
function correctly with volume sets of up to eight members.  Returns a VMS
status code that should be checked for success.
*/

int CollectProxyDevice ()

{
#define PROXY_MAINT_CACHE_DEVICE_MAX 8

   static int  DeviceCount,
               ErrCnt,
               FreeBlocks,
               MaxBlock,
               PrevMinute = -1,
               VolCount;
   static unsigned short  Length;
   static short  DevChannel [PROXY_MAINT_CACHE_DEVICE_MAX];
   static char  CacheDevName [65],
                DevName [65];
   static $DESCRIPTOR (DevNameDsc, "");
   static struct {
      short BufferLength;
      short ItemCode;
      void  *BufferPtr;
      void  *LengthPtr;
   }
   DevNamItemList [] = 
   {
      { sizeof(CacheDevName), DVI$_DEVNAM, &CacheDevName, &Length },
      { sizeof(VolCount), DVI$_VOLCOUNT, &VolCount, 0 },
      { 0, 0, 0, 0 }
   },
   NextDevNamItemList [] = 
   {
      { sizeof(DevName), DVI$_NEXTDEVNAM, &DevName, &Length },
      { 0, 0, 0, 0 }
   },
   BlocksItemList [] = 
   {
      { sizeof(MaxBlock), DVI$_MAXBLOCK, &MaxBlock, 0 },
      { sizeof(FreeBlocks), DVI$_FREEBLOCKS, &FreeBlocks, 0 },
      { sizeof(ErrCnt), DVI$_ERRCNT, &ErrCnt, 0 },
      { 0, 0, 0, 0 }
   };

   int  idx,
        status,
        ErrorCount,
        TotalBlocks,
        TotalFreeBlocks,
        TotalUsedBlocks;
   IOSBLK  IOsb;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CollectProxyDevice() |%s|\n", CacheDevName);

   if (!DeviceCount)
   {
      /**************/
      /* initialize */
      /**************/

      SystemData.ProxyCacheDeviceNamePtr = DevName;

      /* assign a channel to the cache device (or primary if a volume set) */
      DevNameDsc.dsc$a_pointer = "WASD_CACHE_ROOT";
      DevNameDsc.dsc$w_length = strlen(DevNameDsc.dsc$a_pointer);
      status = sys$assign (&DevNameDsc, &DevChannel[DeviceCount], 0, 0);
      if (VMSnok (status))
      {
         DevNameDsc.dsc$a_pointer = "HT_CACHE_ROOT";
         DevNameDsc.dsc$w_length = strlen(DevNameDsc.dsc$a_pointer);
         status = sys$assign (&DevNameDsc, &DevChannel[DeviceCount], 0, 0);
         if (VMSnok (status)) return (status);
      }

      DeviceCount++;

      status = sys$getdviw (0, DevChannel[0], 0,
                            &DevNamItemList, &IOsb, 0, 0, 0);
      if (VMSok (status)) status = IOsb.iosb$w_status;
      if (VMSnok (status)) return (status);

      CacheDevName[Length] = '\0';
      if (CacheDevName[0] == '_')
         memmove (CacheDevName, CacheDevName+1, Length);
      if (Debug) fprintf (stdout, "|%s| %d\n", CacheDevName, VolCount);

      /* loop assigning a channel to all devices in volume set (if it is!) */
      while (--VolCount)
      {
         if (DeviceCount >= PROXY_MAINT_CACHE_DEVICE_MAX)
            exit (SS$_BUGCHECK);

         status = sys$getdviw (0, DevChannel[0], 0,
                               &NextDevNamItemList, &IOsb, 0, 0, 0);
         if (VMSok (status)) status = IOsb.iosb$w_status;
         if (VMSnok (status)) return (status);

         DevName[Length] = '\0';
         if (Debug) fprintf (stdout, "|%s|\n", DevName);

         if (!Length) break;

         DevNameDsc.dsc$w_length = Length;
         DevNameDsc.dsc$a_pointer = DevName;

         if (VMSnok (status =
             sys$assign (&DevNameDsc, &DevChannel[DeviceCount++], 0, 0)))
            return (status);
      }

      if (Debug) fprintf (stdout, "DeviceCount: %d\n", DeviceCount);
   }

   /* once a minute is more than enough */
   if (PrevMinute == CurrentNumTime[4]) return (SS$_NORMAL);
   PrevMinute = CurrentNumTime[4];

   /***********/
   /* process */
   /***********/

   ErrorCount = TotalBlocks = TotalFreeBlocks = 0;

   for (idx = 0; idx < DeviceCount; idx++)
   {
      status = sys$getdviw (0, DevChannel[idx], 0,
                            &BlocksItemList, &IOsb, 0, 0, 0);
      if (VMSok (status)) status = IOsb.iosb$w_status;
      if (VMSnok (status)) return (status);

      if (Debug) fprintf (stdout, "%d %d\n", MaxBlock, FreeBlocks);

      TotalBlocks += MaxBlock;
      TotalFreeBlocks += FreeBlocks;
      ErrorCount += ErrCnt;
   }

   TotalUsedBlocks = TotalBlocks - TotalFreeBlocks;

   if (Debug)
      fprintf (stdout, "%d %d %d %dMB %dMB %dMB\n",
               TotalBlocks, TotalFreeBlocks, TotalUsedBlocks,
               TotalBlocks >> 11, TotalFreeBlocks >> 11, TotalUsedBlocks >> 11);

   SystemData.ProxyCacheTotalBlocks = TotalBlocks;
   SystemData.ProxyCacheUsedBlocks = TotalUsedBlocks;
   SystemData.ProxyCacheFreeBlocks = TotalFreeBlocks;
   SystemData.ProxyCacheErrorCount = ErrorCount;
   SystemData.ProxyCacheDeviceNamePtr = CacheDevName;

   return (SS$_NORMAL);
}

/*****************************************************************************/
/*
Collect scripting data.
*/ 

void CollectScriptData ()

{
   static ulong  PrevCgi,
                 PrevCgiPlus,
                 PrevCgiPlusReused,
                 PrevDclCrePrc,
                 PrevDclDelPrc,
                 PrevDclForceX,
                 PrevRte,
                 PrevRteReused,
                 PrevDclCommand,
                 PrevDECnet,
                 PrevDECnetReused,
                 PrevDECnetCgi,
                 PrevDECnetCgiReused,
                 PrevDECnetOsu,
                 PrevDECnetOsuReused;     

   int  idx;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CollectScriptData()\n");

   /* DCL process */

   SystemData.CurrentScript =
      SystemData.CurrentScriptCGIplus =
      SystemData.CurrentScriptRTE = 0;
   for (idx = 0; idx <= gblptr->InstanceNodeCurrent; idx++)
   {
      SystemData.CurrentScript += accptr->CurrentDclScriptProcess[idx];
      SystemData.CurrentScriptCGIplus += accptr->CurrentDclScriptCgiPlus[idx];
      SystemData.CurrentScriptRTE += accptr->CurrentDclScriptRTE[idx];
   }

   SystemData.TotalDclCrePrc = accptr->DclCrePrcCount;
   SystemData.TotalDclCrePrcDetach = accptr->DclCrePrcDetachCount;
   SystemData.TotalDclCrePrcPersona = accptr->DclCrePrcPersonaCount;
   SystemData.TotalDclCrePrcPersonaDefault =
      accptr->DclCrePrcPersonaDefaultCount;
   SystemData.TotalDclCrePrcPersonaInvUser =
      accptr->DclCrePrcPersonaInvUserCount;
   SystemData.TotalDclCrePrcPersonaPrvUser =
      accptr->DclCrePrcPersonaPrvUserCount;
   SystemData.TotalDclDelPrc = accptr->DclDelPrcCount;
   SystemData.TotalDclForceX = accptr->DclForceXCount;
   SystemData.TotalDclProctor = accptr->DclProctorCount;
   SystemData.TotalDclHardLimit == accptr->DclHitHardLimitCount;

   /* totals */

   SystemData.TotalCgi = accptr->DoScriptCount;
   SystemData.TotalCgiPlus = accptr->DoCgiPlusScriptCount;
   SystemData.TotalCgiPlusReused = accptr->DclCgiPlusReusedCount;
   SystemData.TotalRte = accptr->DoRteScriptCount;
   SystemData.TotalRteReused = accptr->DclRteReusedCount;
   SystemData.TotalAutoScript = accptr->DoAutoScriptCount;
   SystemData.TotalDclCommand = accptr->DoDclCommandCount;

   SystemData.TotalDECnet = accptr->DoDECnetCount;
   SystemData.TotalDECnetReused = accptr->DoDECnetReuseCount;
   SystemData.TotalDECnetCgi = accptr->DoDECnetCgiCount;
   SystemData.TotalDECnetCgiReused = accptr->DoDECnetCgiReuseCount;
   SystemData.TotalDECnetOsu = accptr->DoDECnetOsuCount;
   SystemData.TotalDECnetOsuReused = accptr->DoDECnetOsuReuseCount;

   /* deltas */

   if (PrevDclCrePrc && SystemData.TotalDclCrePrc > PrevDclCrePrc)
      SystemData.DeltaDclCrePrc = SystemData.TotalDclCrePrc - PrevDclCrePrc;
   else
      SystemData.DeltaDclCrePrc = 0;
   PrevDclCrePrc = SystemData.TotalDclCrePrc;

   if (PrevDclDelPrc && SystemData.TotalDclDelPrc > PrevDclDelPrc)
      SystemData.DeltaDclDelPrc = SystemData.TotalDclDelPrc - PrevDclDelPrc;
   else
      SystemData.DeltaDclDelPrc = 0;
   PrevDclDelPrc = SystemData.TotalDclDelPrc;
   /* roll those forced to exit into the delprc delta */
   if (PrevDclForceX && SystemData.TotalDclForceX > PrevDclForceX)
      SystemData.DeltaDclDelPrc += (SystemData.TotalDclForceX - PrevDclForceX);
   PrevDclForceX = SystemData.TotalDclForceX;

   if (PrevCgi && SystemData.TotalCgi > PrevCgi)
      SystemData.DeltaCgi = SystemData.TotalCgi - PrevCgi;
   else
      SystemData.DeltaCgi = 0;
   PrevCgi = SystemData.TotalCgi;

   if (PrevCgiPlus && SystemData.TotalCgiPlus > PrevCgiPlus)
      SystemData.DeltaCgiPlus = SystemData.TotalCgiPlus - PrevCgiPlus;
   else
      SystemData.DeltaCgiPlus = 0;
   PrevCgiPlus = SystemData.TotalCgiPlus;

   if (PrevCgiPlusReused && SystemData.TotalCgiPlusReused > PrevCgiPlusReused)
      SystemData.DeltaCgiPlusReused = SystemData.TotalCgiPlusReused -
                                      PrevCgiPlusReused;
   else
      SystemData.DeltaCgiPlusReused = 0;
   PrevCgiPlusReused = SystemData.TotalCgiPlusReused;

   if (PrevRte && SystemData.TotalRte > PrevRte)
      SystemData.DeltaRte = SystemData.TotalRte - PrevRte;
   else
      SystemData.DeltaRte = 0;
   PrevRte = SystemData.TotalRte;

   if (PrevRteReused && SystemData.TotalRteReused > PrevRteReused)
      SystemData.DeltaRteReused = SystemData.TotalRteReused - PrevRteReused;
   else
      SystemData.DeltaRteReused = 0;
   PrevRteReused = SystemData.TotalRteReused;

   if (PrevDECnet && SystemData.TotalDECnet > PrevDECnet)
      SystemData.DeltaDECnet = SystemData.TotalDECnet - PrevDECnet;
   else
      SystemData.DeltaDECnet = 0;
   PrevDECnet = SystemData.TotalDECnet;

   if (PrevDECnetReused && SystemData.TotalDECnetReused > PrevDECnetReused)
      SystemData.DeltaDECnetReused = SystemData.TotalDECnetReused -
                                     PrevDECnetReused;
   else
      SystemData.DeltaDECnetReused = 0;
   PrevDECnetReused = SystemData.TotalDECnetReused;

   if (PrevDECnetCgi && SystemData.TotalDECnetCgi > PrevDECnetCgi)
      SystemData.DeltaDECnetCgi = SystemData.TotalDECnetCgi - PrevDECnetCgi;
   else
      SystemData.DeltaDECnetCgi = 0;
   PrevDECnetCgi = SystemData.TotalDECnetCgi;

   if (PrevDECnetCgiReused && 
       SystemData.TotalDECnetCgiReused > PrevDECnetCgiReused)
      SystemData.DeltaDECnetCgiReused = SystemData.TotalDECnetCgiReused -
                                        PrevDECnetCgiReused;
   else
      SystemData.DeltaDECnetCgiReused = 0;
   PrevDECnetCgiReused = SystemData.TotalDECnetCgiReused;

   if (PrevDECnetOsu && SystemData.TotalDECnetOsu > PrevDECnetOsu)
      SystemData.DeltaDECnetOsu = SystemData.TotalDECnetOsu - PrevDECnetOsu;
   else
      SystemData.DeltaDECnetOsu = 0;
   PrevDECnetOsu = SystemData.TotalDECnetOsu;

   if (PrevDECnetOsuReused &&
       SystemData.TotalDECnetOsuReused > PrevDECnetOsuReused)
      SystemData.DeltaDECnetOsuReused = SystemData.TotalDECnetOsuReused -
                                        PrevDECnetOsuReused;
   else
      SystemData.DeltaDECnetOsuReused = 0;
   PrevDECnetOsuReused = SystemData.TotalDECnetOsuReused;
} 

/*****************************************************************************/
/*
Collect specific data when [x]WebDAV checkbox.
*/ 

void CollectWebDavData ()

{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CollectWebDavData()\n");

   if (!SystemData.WebDavEnabled) return;

   SystemData.WebDavAgentAppleCount = accptr->WebDavAgentAppleCount;
   SystemData.WebDavAgentBsdCount = accptr->WebDavAgentBsdCount;
   SystemData.WebDavAgentMicrosoftCount = accptr->WebDavAgentMicrosoftCount;
   SystemData.WebDavAgentLinuxCount = accptr->WebDavAgentLinuxCount;
   SystemData.WebDavAgentOtherCount = accptr->WebDavAgentOtherCount;
   SystemData.WebDavAgentSunCount = accptr->WebDavAgentSunCount;
   SystemData.WebDavMetaReadAttemptCount = accptr->WebDavMetaReadAttemptCount;
   SystemData.WebDavMetaReadCount = accptr->WebDavMetaReadCount;
   SystemData.WebDavMetaWriteAttemptCount = accptr->WebDavMetaWriteAttemptCount;
   SystemData.WebDavMetaWriteCount = accptr->WebDavMetaWriteCount;
   SystemData.WebDavRequestCount = accptr->WebDavRequestCount;
   SystemData.WebDavXmlParseCount = accptr->WebDavXmlParseCount;

   PUT_QUAD_QUAD (&accptr->WebDavBytesRawRx, &SystemData.WebDavBytesRawRx);
   PUT_QUAD_QUAD (&accptr->WebDavBytesRawTx, &SystemData.WebDavBytesRawTx);
}

/*****************************************************************************/
/*
Per-boot system data.
*/ 

int CollectStaticData ()

{
   static ulong  ActiveCpuCnt,
                 AvailCpuCount,
                 ClusterNodes,
                 JpiPid,
                 MemSize,
                 PageSize;
   static ulong  BootBinTime [2];
   static ushort  HwNameLength;
   static uchar  ClusterMember;
   static char  ArchName [15+1],
                HwName [1+60],
                NodeName [15+1],
                VmsVersion [8+1];

   /* initialize structure for SYI items */
   static struct {
      short  BufferLength;
      short  ItemCode;
      void  *BufferPtr;
      void  *LengthPtr;
   } SyiItemList[] = 
   {
       { sizeof(AvailCpuCount), SYI$_AVAILCPU_CNT, &AvailCpuCount, 0 },
       { sizeof(ActiveCpuCnt), SYI$_ACTIVECPU_CNT, &ActiveCpuCnt, 0 },
       { sizeof(ArchName), SYI$_ARCH_NAME, &ArchName, 0 },
       { sizeof(BootBinTime), SYI$_BOOTTIME, &BootBinTime, 0 },
       { sizeof(ClusterMember), SYI$_CLUSTER_MEMBER, &ClusterMember, 0 },
       { sizeof(ClusterNodes), SYI$_CLUSTER_NODES, &ClusterNodes, 0 },
       { sizeof(HwName), SYI$_HW_NAME, &HwName, &HwNameLength },
       { sizeof(MemSize), SYI$_MEMSIZE, &MemSize, 0 },
       { sizeof(NodeName), SYI$_SCSNODE, &NodeName, 0 },
       { sizeof(PageSize), SYI$_PAGE_SIZE, &PageSize, 0 },
       { sizeof(VmsVersion), SYI$_VERSION, &VmsVersion, 0 },
       {0,0,0,0}
   }, JpiItemList[] = 
   {
       { sizeof(JpiPid), JPI$_PID, &JpiPid, 0 },
       {0,0,0,0}
   };

   int  status;
   char  *cptr, *sptr, *zptr;
   IOSBLK  IOsb;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CollectStaticData()\n");

   status = sys$getsyiw (
            0,  /* efn */
            0,  /* csiaddr */
            0,  /* nodename */
            &SyiItemList,  /* item list */
            &IOsb,  /* iosb */
            0,      /* astaddr */
            0 );    /* astprm */

   if (Debug)
      fprintf (stdout, "sys$getsyi() %%X%08.08X  IOsb: %%X%08.08X\n",
               status, IOsb.iosb$w_status);

   if (VMSnok (status)) return (status);

   zptr = (sptr = SystemData.ArchName) + sizeof(SystemData.ArchName)-1;
   for (cptr = ArchName;
        *cptr && *cptr != ' ' && sptr < zptr;
        *sptr++ = *cptr++); 
   *sptr = '\0';

   HwName[HwNameLength] = '\0';
   zptr = (sptr = SystemData.HwName) + sizeof(SystemData.HwName)-1;
   for (cptr = HwName; *cptr && sptr < zptr; *sptr++ = *cptr++); 
   *sptr = '\0';

   zptr = (sptr = SystemData.NodeName) + sizeof(SystemData.NodeName)-1;
   for (cptr = NodeName;
        *cptr && *cptr != ' ' && sptr < zptr;
        *sptr++ = *cptr++); 
   *sptr = '\0';

   zptr = (sptr = SystemData.VmsVersion) + sizeof(SystemData.VmsVersion)-1;
   for (cptr = VmsVersion;
        *cptr && *cptr != ' ' && sptr < zptr;
        *sptr++ = *cptr++); 
   *sptr = '\0';

   SystemData.MemSize = MemSize;
   SystemData.PageSize = PageSize;

   PUT_QUAD_QUAD (&BootBinTime, &SystemData.BootBinTime);

   SystemData.AvailCpuCount = AvailCpuCount;
   SystemData.ActiveCpuCount = ActiveCpuCnt;
   SystemData.ClusterMember = ClusterMember;
   SystemData.ClusterNodes = ClusterNodes;

   if (Debug)
      fprintf (stdout, "|%s|%s|%s|%s| %d %d\n",
               SystemData.ArchName, SystemData.HwName,
               SystemData.NodeName, SystemData.VmsVersion,
               SystemData.AvailCpuCount, ClusterMember);

   status = sys$getjpiw (
            0,  /* efn */
            0,  /* csiaddr */
            0,  /* nodename */
            &JpiItemList,  /* item list */
            &IOsb,  /* iosb */
            0,      /* astaddr */
            0 );    /* astprm */

   if (Debug)
      fprintf (stdout, "sys$getjpi() %%X%08.08X  IOsb: %%X%08.08X\n",
               status, IOsb.iosb$w_status);

   if (VMSnok (status)) return (status);

   SystemData.JpiPid = JpiPid;

   SystemData.ProxyEnabled = pacptr->ServingEnabled;

   SystemData.WebDavEnabled = accptr->WebDavEnabled;

   return (status);
} 

/*****************************************************************************/
/*
Collect System Performance Information.
*/ 

int CollectSPI ()

{
#define INTSTK_MODE  0
#define MPSYNCH_MODE 1
#define KERNEL_MODE  2
#define EXEC_MODE    3
#define SUPER_MODE   4
#define USER_MODE    5
/* 6 is reserved */
#define IDLE_MODE    7

#define MODE_COUNT IDLE_MODE+1
#define CPU_MAX             32

   /* structure for a single CPU's mode ticks */
   static struct SingleCPU {
       uchar  CPUid;
       ulong  Count [MODE_COUNT];
   };

   /* a structure all CPU's mode ticks for a system with up to 32 CPUs */
   static struct {
       ulong  NumberOfCPUs;
       struct SingleCPU  CPU [CPU_MAX];
   } ModeTicks;

   static ulong   BufferedIO,
                  CputTicksBusy,
                  CpuUser,
                  DirectIO,
                  FcpRead,
                  FcpWrite,
                  PrevBufferedIO,
                  PrevCpuTicksBusy,
                  PrevCpuTicksUser,
                  PrevDirectIO,
                  PrevFcpRead,
                  PrevFcpWrite,
                  ProcessCount;

#ifdef __ia64
   static unsigned __int64  FreeList,
                            ModList;
#else
   static ulong  FreeList,
                 ModList;
#endif

   /* initialize structure for RMI items */
   static struct {
      short  BufferLength;
      short  ItemCode;
      void  *BufferPtr;
      void  *LengthPtr;
   } ItemList[] =
   {
       { sizeof(BufferedIO), SPI$_BUFIO, &BufferedIO, 0 },
       { sizeof(ModeTicks), SPI$_MODES, &ModeTicks, 0 },
       { sizeof(DirectIO), SPI$_DIRIO, &DirectIO, 0 },
       { sizeof(FcpRead), SPI$_FCPREAD, &FcpRead, 0 }, 
       { sizeof(FcpWrite), SPI$_FCPWRITE, &FcpWrite, 0 }, 
       { sizeof(FreeList), SPI$_FRLIST, &FreeList, 0 }, 
       { sizeof(ModList), SPI$_MODLIST, &ModList, 0 },
       { sizeof(ProcessCount), SPI$_PROCS, &ProcessCount, 0 },
       {0,0,0,0}
   };

   int  cidx, midx, status, tlong;
   ulong  datsize,
          CpuTicksBusy;
   IOSBLK  IOsb,
           SynchIOsb;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CollectSPI()\n");

   status = lib$sub_times (&CurrentBinTime,
                           &SystemData.BootBinTime,
                           &SystemData.UpBinTime);
   if (VMSnok (status)) EXIT_FI_LI (status);

   /* collect System Performance Information */
   status = exe$getspi (
            SpiEfn,  /* efn */
            0,  /* csiaddr */
            0,  /* nodename */
            &ItemList,  /* item list */
            &IOsb,  /* iosb */
            0,      /* astaddr */
            0 );    /* astprm */

   if (Debug)
      fprintf (stdout, "exe$getspi() %%X%08.08X  IOsb: %%X%08.08X\n",
               status, IOsb.iosb$w_status);

   if (VMSok (status)) status = IOsb.iosb$w_status;
   if (VMSnok (status)) return (status);

   /*******/
   /* CPU */
   /*******/

   for (cidx = 1; cidx < ModeTicks.NumberOfCPUs; cidx++)
      for (midx = INTSTK_MODE; midx <= IDLE_MODE; midx++)
         ModeTicks.CPU[0].Count[midx] += ModeTicks.CPU[cidx].Count[midx];

   CpuTicksBusy = CpuUser = 0;
   CpuTicksBusy = ModeTicks.CPU[0].Count[INTSTK_MODE];
   /* interrupt item is actually the sum of interrupt and idle! */
   CpuTicksBusy -= ModeTicks.CPU[0].Count[IDLE_MODE];
   CpuTicksBusy += ModeTicks.CPU[0].Count[KERNEL_MODE] +
                   ModeTicks.CPU[0].Count[MPSYNCH_MODE] +
                   ModeTicks.CPU[0].Count[EXEC_MODE] +
                   ModeTicks.CPU[0].Count[SUPER_MODE] +
                   ModeTicks.CPU[0].Count[USER_MODE];
   CpuUser = ModeTicks.CPU[0].Count[USER_MODE];

   if (PrevCpuTicksBusy && CpuTicksBusy > PrevCpuTicksBusy)
   {
      SystemData.CpuTicksBusy = CpuTicksBusy - PrevCpuTicksBusy;
      SystemData.CpuBusyPercent =
         (ulong)((float)SystemData.CpuTicksBusy / DeltaSeconds);
      if (SystemData.CpuBusyPercent > 100 * SystemData.ActiveCpuCount)
         SystemData.CpuBusyPercent = 100 * SystemData.ActiveCpuCount;
   }
   else
      SystemData.CpuBusyPercent = 0;

   PrevCpuTicksBusy = CpuTicksBusy;

   if (PrevCpuTicksUser)
   {
      SystemData.CpuTicksUser = CpuUser - PrevCpuTicksUser;
      SystemData.CpuUserPercent =
         (ulong)((float)SystemData.CpuTicksUser / DeltaSeconds);
      if (SystemData.CpuUserPercent > 100 * SystemData.ActiveCpuCount)
         SystemData.CpuUserPercent = 100 * SystemData.ActiveCpuCount;
   }
   else
      SystemData.CpuUserPercent = 0;

   PrevCpuTicksUser = CpuUser;

   if (Debug)
      fprintf (stdout, "CPU: %d%% USER: %d%%\n",
               SystemData.CpuBusyPercent, SystemData.CpuUserPercent);

   /*********/
   /* other */
   /*********/

   if (PrevBufferedIO && BufferedIO > PrevBufferedIO)
      SystemData.BufferedIO = BufferedIO - PrevBufferedIO;
   else
      SystemData.BufferedIO = 0;
   PrevBufferedIO = BufferedIO;
   SystemData.BufferedIO = (ulong)((float)SystemData.BufferedIO / DeltaSeconds);

   if (PrevDirectIO && DirectIO > PrevDirectIO)
      SystemData.DirectIO = DirectIO - PrevDirectIO;
   else
      SystemData.DirectIO = 0;
   PrevDirectIO = DirectIO;
   SystemData.DirectIO = (ulong)((float)SystemData.DirectIO / DeltaSeconds);

   if (PrevFcpRead && FcpRead > PrevFcpRead)
      SystemData.FcpRead = FcpRead - PrevFcpRead;
   else
      SystemData.FcpRead = 0;
   PrevFcpRead = FcpRead;
   SystemData.FcpRead = (ulong)((float)SystemData.FcpRead / DeltaSeconds);

   if (PrevFcpWrite && FcpWrite > PrevFcpWrite)
      SystemData.FcpWrite = FcpWrite - PrevFcpWrite;
   else
      SystemData.FcpWrite = 0;
   PrevFcpWrite = FcpWrite;
   SystemData.FcpWrite = (ulong)((float)SystemData.FcpWrite / DeltaSeconds);

   SystemData.MemInUse = SystemData.MemSize - FreeList - ModList;

   if (SystemData.MemInUse > SystemData.MemInUseMax)
      SystemData.MemInUseMax = SystemData.MemInUse;

   SystemData.ProcessCount = ProcessCount;

   return (SS$_NORMAL);
}

/*****************************************************************************/
/*
This function uses undocumented sense-mode functionality to load counters from
the network interface (commonly ethernet device).  Multiple devices are
supported using a multi-valued logical name ALAMODE_NI and the frame-rate
and byte-rate and soft/hard error-rate statistics are accumlated to represent
total system data.

Essential concepts have come from "EMON: moniteur ethernet (V2.1-2) Gerard
Guillaume" and found in DECUSLIB freeware, and from some peeks at the T4 source
code.  The implementation is mine.
*/

void CollectNetIntData ()

{
   static int  NetIntCount,
               IoClrCount;
   static ushort  NetIntChan [NI_DEVICE_MAX];
   static ulong  DcSCOM = 32,  /* DC$_SCOM */
                 DtNI = 13;    /* DT$_NI */

   static ulong  PrevBlocksRx [2],
                 PrevBlocksTx [2],
                 PrevBytesRx [2],
                 PrevBytesRxTx [2],
                 PrevBytesTx [2],
                 PrevErrorsHard [2],
                 PrevErrorsSoft [2];
   static long  DeviceCount = -1,
                DviUnit,
                LnmIndex = -1,
                LnmAttributes;
   static ushort  LogValueLen;
   static char  LogValue [32];
   static struct {
      short  BufferLength;
      short  ItemCode;
      void  *BufferPtr;
      void  *LengthPtr;
   }
   LnmItems [] =
   {
      { sizeof(LnmIndex), LNM$_INDEX, &LnmIndex, 0 },
      { sizeof(LnmAttributes), LNM$_ATTRIBUTES, &LnmAttributes, 0 },
      { sizeof(LogValue)-1, LNM$_STRING, &LogValue, &LogValueLen },
      { 0,0,0,0 }
   },
   DevScanItems [] =
   {
      { sizeof(DcSCOM), DVS$_DEVCLASS, &DcSCOM, 0 },
      { 0,0,0,0 }
   },
   GetDviItems [] =
   {
      { sizeof(DviUnit), DVI$_UNIT, &DviUnit, 0 },
      { 0,0,0,0 }
   };
   static $DESCRIPTOR (NetIntDsc, "");
   static $DESCRIPTOR (LogNameDsc, "ALAMODE_NI");
   static $DESCRIPTOR (LnmFileDevDsc, "LNM$FILE_DEV");

   int  ecnt, status;
   ushort  pword,
           retlen;
   ulong  LnmCount;
   ulong  value [2],
          DevScanContext [2],
          NetIntBlocksRx[2],
          NetIntBlocksTx[2],
          NetIntBytesRx[2],
          NetIntBytesRxTx[2],
          NetIntBytesTx[2],
          NetIntErrorsHard[2],
          NetIntErrorsSoft[2];
   char  *vptr;
   char  DevScanBuffer [64],
         CountBuffer [512];
   struct STRUCT_IOSBLK  IOsb;
   $DESCRIPTOR (CountDsc, CountBuffer);
   $DESCRIPTOR (DevScanDsc, DevScanBuffer);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CollectNetIntData() %s\n", NetIntDev);

   if (LnmIndex < 0)
   {
      /* first call, check for logical name defined network devices */
      for (LnmIndex = 0; LnmIndex < NI_DEVICE_MAX; LnmIndex++)
      {
         status = sys$trnlnm (0, &LnmFileDevDsc, &LogNameDsc, 0, &LnmItems);
         if (VMSnok (status) || !(LnmAttributes & LNM$M_EXISTS)) break;
         LogValue[LogValueLen] = '\0';

         NetIntDsc.dsc$a_pointer = LogValue;
         NetIntDsc.dsc$w_length = LogValueLen;

         status = sys$assign (&NetIntDsc, &NetIntChan[LnmIndex], 0, 0);
         if (VMSnok (status)) EXIT_FI_LI (status);
         if (NetIntDev[0]) strcat (NetIntDev, ",");
         strcat (NetIntDev, LogValue);
         NetIntCount++;
      }
   }

   if (DeviceCount < 0)
   {
      /* first call */
      DeviceCount = 0;

      /* if logical name defined and found devices */
      if (NetIntCount) return;

      /* scan system for network devices */
      DevScanContext[0] = DevScanContext[1] = 0;
      while (NetIntCount < NI_DEVICE_MAX)
      {
         status = sys$device_scan (&DevScanDsc, &retlen, 0,
                                   &DevScanItems, &DevScanContext);
         if (VMSnok (status))
         {
            if (status == SS$_NOMOREDEV || status == SS$_NOSUCHDEV) break;
            if (VMSnok (status)) EXIT_FI_LI (status);
         }

         status = sys$getdviw (0, 0, &DevScanDsc, &GetDviItems, &IOsb, 
                               0, 0, 0, 0);
         if (VMSok (status)) status = IOsb.iosb$w_status;
         if (VMSnok (status)) EXIT_FI_LI (status);

         /* if not a template device then continue */
         if (DviUnit != 0) continue;

         DevScanBuffer[retlen] = '\0';

         /* RMDRIVER gives %SYSTEM-F-SYSVERDIF */
         if (!strcmp (DevScanBuffer, "_RMA0:")) continue;

         status = sys$assign (&DevScanDsc, &NetIntChan[DeviceCount++], 0, 0);
         if (VMSnok (status)) EXIT_FI_LI (status);

         if (NetIntDev[0]) strcat (NetIntDev, ",");
         strcat (NetIntDev, DevScanBuffer+1);
         NetIntCount++;
      }

      /* first call, just return having assigned channels (perhaps) */
      return;
   }

   /****************/
   /* collect data */
   /****************/

   /* initialise structure and local storage */
   PUT_ZERO_QUAD (&NetIntBlocksRx);
   PUT_ZERO_QUAD (&NetIntBlocksTx);
   PUT_ZERO_QUAD (&NetIntBytesRx);
   PUT_ZERO_QUAD (&NetIntBytesTx);
   PUT_ZERO_QUAD (&NetIntBytesRxTx);
   PUT_ZERO_QUAD (&NetIntErrorsHard);
   PUT_ZERO_QUAD (&NetIntErrorsSoft);

   /* if no devices then just return from here */
   if (!NetIntCount) return;

   /* accumulate counts from each device */
   for (ecnt = 0; ecnt < NetIntCount; ecnt++)
   {
      memset (CountBuffer, 0, sizeof(CountBuffer));

      status = sys$qiow (0, NetIntChan[ecnt],
                         IO$_SENSEMODE | IO$M_RD_COUNT | IoClrCount |
#ifndef __VAX
#define IO$M_RD_64COUNT 0x4000
                         IO$M_RD_64COUNT | 
#endif
                         IO$M_CTRL,
                         &IOsb, 0, 0,
                         0, &CountDsc, 0, 0, 0, 0);

      if (VMSok (status)) status = IOsb.iosb$w_status;
      if (VMSnok (status)) EXIT_FI_LI (status);

      vptr = CountBuffer;
      while (vptr < CountBuffer + IOsb.iosb$w_bcnt)
      {
         pword = *(ushort*)vptr;
         vptr += 2;

         /* end of list */
         if (!pword) break;

         /* MSB should be set! */
         if (!(pword & 0x8000)) break;

         PUT_ZERO_QUAD (&value);
         if ((pword & 0x0fff) < 200)
         {
            /* quadword value (item number less than 200) */
            PUT_QUAD_QUAD (vptr, &value);
            vptr += 8;
         }
         else
         if ((pword & 0x6000) == 0x6000)
         {
            /* longword value */
            value[0] = *(ulong*)vptr;
            vptr += 4;
         } 
         else
         if ((pword & 0x6000) == 0x4000)
         {
            /* word value */
            value[0] = *(ushort*)vptr;
            vptr += 2;
         } 
         else
         if ((pword & 0x6000) == 0x2000)
         {
            /* byte value */
            value[0] = *(uchar*)vptr;
            vptr++;
         } 

         if (pword & 0x1000)
         {
            /* hw map (?) */
            value[0] = *(ulong*)vptr;
            vptr += 2;
            continue;
         } 

         switch (pword & 0x0fff)
         {
            /* these look suspiciously like "LANCP SHOW DEVICE /COUNT"! */

            case 1 :
                 /* seconds since last zeroed */
                 break;

            case 2 :
            case 6 :
                 ADD_QUAD_QUAD (&value, &NetIntBytesRx);
                 break;

            case 3 :
            case 7 :
                 ADD_QUAD_QUAD (&value, &NetIntBytesTx);
                 break;

            case 4 :
            case 8 :
                 ADD_QUAD_QUAD (&value, &NetIntBlocksRx);
                 break;

            case 5 :
            case 9 :
                 ADD_QUAD_QUAD (&value, &NetIntBlocksTx);
                 break;

            case 12 :
            case 13 :
            case 14 :
            case 15 :
            case 16 :
            case 17 :
            case 18 :
            case 19 :
            case 30 :
            case 21 :
            case 22 :
            case 23 :
            case 24 :
            case 27 :
            case 28 :
            case 29 :
                 ADD_QUAD_QUAD (&value, &NetIntErrorsSoft);
                 break;

            case 25 :
            case 26 :
            case 33 :
            case 34 :
                 ADD_QUAD_QUAD (&value, &NetIntErrorsHard);
                 break;

            /* more "traditional" counters */

            case 1000 /* NMA$C_CTLIN_BRC bytes received */ :
            case 1002 /* NMA$C_CTLIN_MBY multicast bytes received */ :
                 ADD_QUAD_QUAD (&value, &NetIntBytesRx);
                 break;

            case 1001 /* NMA$C_CTLIN_BSN bytes sent */ :
                 ADD_QUAD_QUAD (&value, &NetIntBytesTx);
                 break;

            case 1010 /* NMA$C_CTLIN_DBR data blocks received */ :
            case 1012 /* NMA$C_CTLIN_MBL multicast blocks received */ :
                 ADD_QUAD_QUAD (&value, &NetIntBlocksRx);
                 break;

            case 1011 /* NMA$C_CTLIN_DBS data blocks sent */ :
                 ADD_QUAD_QUAD (&value, &NetIntBlocksTx);
                 break;

            case 1013 /* NMA$C_CTLIN_BID */ :
            case 1014 /* NMA$C_CTLIN_BS1 */ :
            case 1015 /* NMA$C_CTLIN_BSM */ :
            case 1040 /* NMA$C_CTLIN_RBE */ :
            case 1041 /* NMA$C_CTLIN_LBE */ :
            case 1064 /* NMA$C_CTLIN_OVR */ :
            case 1066 /* NMA$C_CTLIN_UBU */ :

                 ADD_QUAD_QUAD (&value, &NetIntErrorsSoft);
                 break;

            case 1060 /* NMA$C_CTLIN_SFL */ :
            case 1061 /* NMA$C_CTLIN_CDC */ :
            case 1063 /* NMA$C_CTLIN_UFD */ :
            case 1062 /* NMA$C_CTLIN_RFL */ :
            case 1065 /* NMA$C_CTLIN_SBU */ :

                 ADD_QUAD_QUAD (&value, &NetIntErrorsHard);
                 break;

            default :
         }
      }
   }

#ifdef __VAX
   /* 32 bit accumulators tend to fill rather rapidly */
   if (NetIntBytesRx[0] == 0xffffffff ||
       NetIntBytesTx[0] == 0xffffffff)
      IoClrCount = IO$M_CLR_COUNT;
   else
      IoClrCount = 0;
#endif

   if (QUAD_NOT_ZERO (&PrevBytesRx) &&
       QUAD_MORE_THAN_QUAD (&NetIntBytesRx, &PrevBytesRx))
      MINUS_QUAD_QUAD (&PrevBytesRx, &NetIntBytesRx,
                       &SystemData.NetIntBytesRx);
   else
      PUT_ZERO_QUAD (&SystemData.NetIntBytesRx);
   PUT_QUAD_QUAD (&NetIntBytesRx, &PrevBytesRx);

   if (QUAD_NOT_ZERO (&PrevBytesTx) &&
       QUAD_MORE_THAN_QUAD (&NetIntBytesTx, &PrevBytesTx))
      MINUS_QUAD_QUAD (&PrevBytesTx, &NetIntBytesTx,
                       &SystemData.NetIntBytesTx);
   else
      PUT_ZERO_QUAD (&SystemData.NetIntBytesTx);
   PUT_QUAD_QUAD (&NetIntBytesTx, &PrevBytesTx);

   PLUS_QUAD_QUAD (&NetIntBytesRx, &NetIntBytesTx, &NetIntBytesRxTx);
   if (QUAD_NOT_ZERO (&PrevBytesRxTx) &&
       QUAD_MORE_THAN_QUAD (&NetIntBytesRxTx, &PrevBytesRxTx))
      MINUS_QUAD_QUAD (&PrevBytesRxTx, &NetIntBytesRxTx,
                       &SystemData.NetIntBytesRxTx);
   else
      PUT_ZERO_QUAD (&SystemData.NetIntBytesRxTx);
   PUT_QUAD_QUAD (&NetIntBytesRxTx, &PrevBytesRxTx);

   SystemData.NetBytesRxPerSec =
      (ulong)(FLOAT_FROM_QUAD (&SystemData.NetIntBytesRx) / DeltaSeconds);

   SystemData.NetBytesTxPerSec =
      (ulong)(FLOAT_FROM_QUAD (&SystemData.NetIntBytesTx) / DeltaSeconds);

   SystemData.NetBytesPerSec =
      (ulong)(FLOAT_FROM_QUAD (&SystemData.NetIntBytesRxTx) / DeltaSeconds);

   if (QUAD_NOT_ZERO (&PrevBlocksRx) &&
       QUAD_MORE_THAN_QUAD (&NetIntBlocksRx, &PrevBlocksRx))
      MINUS_QUAD_QUAD (&PrevBlocksRx, &NetIntBlocksRx,
                       &SystemData.NetIntBlocksRx);
   else
      PUT_ZERO_QUAD (&SystemData.NetIntBlocksRx);
   PUT_QUAD_QUAD (&NetIntBlocksRx, &PrevBlocksRx);

   if (QUAD_NOT_ZERO (&PrevBlocksTx) &&
       QUAD_MORE_THAN_QUAD (&NetIntBlocksTx, &PrevBlocksTx))
      MINUS_QUAD_QUAD (&PrevBlocksTx, &NetIntBlocksTx,
                       &SystemData.NetIntBlocksTx);
   else
     PUT_ZERO_QUAD (&SystemData.NetIntBlocksTx);
   PUT_QUAD_QUAD (&NetIntBlocksTx, &PrevBlocksTx);

   if (QUAD_ZERO (&SystemData.NetIntBlocksRx))
      SystemData.NetDatagramRx = 0;
   else
      SystemData.NetDatagramRx =
         (ulong)(FLOAT_FROM_QUAD (&SystemData.NetIntBytesRx) /
                 FLOAT_FROM_QUAD (&SystemData.NetIntBlocksRx) /
                 DeltaSeconds);

   if (QUAD_ZERO (&SystemData.NetIntBlocksTx))
      SystemData.NetDatagramTx = 0;
   else
      SystemData.NetDatagramTx =
         (ulong)(FLOAT_FROM_QUAD (&SystemData.NetIntBytesTx) /
                 FLOAT_FROM_QUAD (&SystemData.NetIntBlocksTx) /
                 DeltaSeconds);

   if (QUAD_NOT_ZERO (&PrevErrorsSoft) &&
       QUAD_MORE_THAN_QUAD (&NetIntErrorsSoft, &PrevErrorsSoft))
      MINUS_QUAD_QUAD (&PrevErrorsSoft, &NetIntErrorsSoft,
                       &SystemData.NetIntErrorsSoft);
   PUT_QUAD_QUAD (&NetIntErrorsSoft, &PrevErrorsSoft);

   if (QUAD_NOT_ZERO (&PrevErrorsHard) &&
       QUAD_MORE_THAN_QUAD (&NetIntErrorsHard, &PrevErrorsHard))
      MINUS_QUAD_QUAD (&PrevErrorsHard, &NetIntErrorsHard,
                       &SystemData.NetIntErrorsHard);
   PUT_QUAD_QUAD (&NetIntErrorsHard, &PrevErrorsHard);
}

/*****************************************************************************/
/*
See [SRC.HTTPD]CONTROL.C for other information.  Uses the VMS Distributed Lock
Manager to keep track of how many servers are currently executing on the
node/cluster.  NL locks indicate interest (used by this utility), CR locks
indicate a server waiting for a group directive.  This function enqueues a
single NL lock, then periodically get all the locks associated with that
resource and counts up the number of CR locks - giving the number of servers!
*/

void CollectInstanceData ()

{
   static int  InstanceCount,
               PrevClusterCount = -1,
               PrevNodeCount = -1;
   static ulong  InstancePid [INSTANCE_MAX];
   static char  ClusterLockName [32],
                NodeLockName [32];
   static $DESCRIPTOR (ClusterLockNameDsc, ClusterLockName);
   static $DESCRIPTOR (NodeLockNameDsc, NodeLockName);
   static LKIDEF  LkiLocks [32];
   static struct lksb  ClusterLockLksb,
                       NodeLockLksb;

   static struct
   {
      ushort  TotalLength,  /* bits 0..15 */
              LockLength;  /* bits 16..30 */
   } ReturnLength;

   static struct
   {
      ushort  buf_len;
      ushort  item;
      uchar  *buf_addr;
      void  *ret_len;
   }
   LkiItems [] =
   {
      { sizeof(LkiLocks), LKI$_LOCKS, &LkiLocks, &ReturnLength },
      {0,0,0,0}
   };

   int  cnt, status,
        ClusterCount,
        LockCount,
        LockNameMagic,
        NodeCount,
        NonNlLockCount;
   char  *cptr, *sptr, *zptr;
   IOSBLK  IOsb;
   LKIDEF  *lkiptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CollectInstanceData()\n");

   if (!ClusterLockName[0])
   {
      /**************/
      /* initialise */
      /**************/

      if (InstanceEnvNumber > INSTANCE_ENV_NUMBER_MAX)
         EXIT_FI_LI (STS$K_ERROR | STS$M_INHIB_MSG);

      /* a byte comprising two 4 bit fields, version and server group number */
      LockNameMagic = ((HTTPD_LOCK_VERSION & 0xf) << 4) |
                      (InstanceEnvNumber & 0xf);

      /* build the (binary) resource name for the cluster lock */
      zptr = (sptr = ClusterLockName) + sizeof(ClusterLockName)-1;
      for (cptr = HTTPD_NAME; *cptr && sptr < zptr; *sptr++ = *cptr++);
      if (sptr < zptr) *sptr++ = (char)LockNameMagic;
      if (sptr < zptr) *sptr++ = (char)INSTANCE_CLUSTER;
      ClusterLockNameDsc.dsc$w_length = sptr - ClusterLockName;

      /* enqueue a just-interested NL lock */
      status = sys$enqw (0, LCK$K_NLMODE, &ClusterLockLksb,
                         LCK$M_EXPEDITE | LCK$M_SYSTEM,
                         &ClusterLockNameDsc, 0, 0, 0, 0, 0, 2, 0);
      if (VMSok (status)) status = ClusterLockLksb.lksb$w_status;
      if (VMSnok (status)) EXIT_FI_LI (status);

      /* build the (binary) resource name for the node lock */
      zptr = (sptr = NodeLockName) + sizeof(NodeLockName)-1;
      for (cptr = HTTPD_NAME; *cptr && sptr < zptr; *sptr++ = *cptr++);
      if (sptr < zptr) *sptr++ = (char)LockNameMagic;
      for (cptr = SystemData.NodeName; *cptr && sptr < zptr; *sptr++ = *cptr++);
      if (sptr < zptr) *sptr++ = (char)INSTANCE_NODE;
      NodeLockNameDsc.dsc$w_length = sptr - NodeLockName;

      /* enqueue a just-interested NL lock */
      status = sys$enqw (0, LCK$K_NLMODE, &NodeLockLksb,
                         LCK$M_EXPEDITE | LCK$M_SYSTEM,
                         &NodeLockNameDsc, 0, 0, 0, 0, 0, 2, 0);
      if (VMSok (status)) status = NodeLockLksb.lksb$w_status;
      if (VMSnok (status)) EXIT_FI_LI (status);

   }

   /*******************/
   /* count instances */
   /*******************/

   /* get and count the cluster instance locks */

   status = sys$getlkiw (0, &ClusterLockLksb.lksb$l_lkid, &LkiItems, &IOsb,
                         0, 0, 0);
   if (VMSok (status)) status = ClusterLockLksb.lksb$w_status;
   if (VMSnok (status)) EXIT_FI_LI (status);

   if (ReturnLength.LockLength)
   {
      /* if insufficient buffer space */
      if (ReturnLength.LockLength & 0x8000) EXIT_FI_LI (SS$_BADPARAM);
      LockCount = ReturnLength.TotalLength / ReturnLength.LockLength;
   }
   else
      LockCount = 0;
   if (Debug) fprintf (stdout, "%d\n", LockCount);

   ClusterCount = 0;
   lkiptr = &LkiLocks;
   for (cnt = LockCount; cnt; cnt--)
   {
      if (Debug) fprintf (stdout, "lki$b_grmode: %d\n", lkiptr->lki$b_grmode);
      if (lkiptr->lki$b_grmode != LCK$K_NLMODE) ClusterCount++;
      lkiptr++;
   }

   /* get and count the node instance locks */

   status = sys$getlkiw (0, &NodeLockLksb.lksb$l_lkid, &LkiItems, &IOsb,
                         0, 0, 0);
   if (VMSok (status)) status = NodeLockLksb.lksb$w_status;
   if (VMSnok (status)) EXIT_FI_LI (status);

   if (ReturnLength.LockLength)
   {
      /* if insufficient buffer space */
      if (ReturnLength.LockLength & 0x8000) EXIT_FI_LI (SS$_BADPARAM);
      LockCount = ReturnLength.TotalLength / ReturnLength.LockLength;
   }
   else
      LockCount = 0;
   if (Debug) fprintf (stdout, "%d\n", LockCount);

   NodeCount = 0;
   for (cnt = 0; cnt < INSTANCE_MAX; cnt++) InstancePid[cnt] = 0;
   lkiptr = &LkiLocks;
   for (cnt = LockCount; cnt; cnt--)
   {
      if (Debug)
         fprintf (stdout, "lki$b_grmode: %d %08.08X\n",
                  lkiptr->lki$b_grmode, lkiptr->lki$l_pid);
      if (lkiptr->lki$b_grmode != LCK$K_NLMODE)
         InstancePid[NodeCount++] = lkiptr->lki$l_pid;
      lkiptr++;
   }

   if (++InstanceCount > NodeCount) InstanceCount = 1;

   if (PrevClusterCount < 0)
   {
      /* initialize */
      PrevClusterCount = ClusterCount;
      PrevNodeCount = NodeCount;
   }

   SystemData.InstanceNodeCount = NodeCount;
   SystemData.InstanceClusterCount = ClusterCount;
   for (cnt = 0; cnt < INSTANCE_MAX; cnt++)
      SystemData.InstanceNodePid[cnt] = InstancePid[cnt];
}

/*****************************************************************************/
/*
See [SRC.HTTPD]CONTROL.C for other information.  Uses the VMS Distributed Lock
Manager.  Called directly with a NULL parameter to initialise locks associated
with distributed node and cluster commands (/DO= and the Server Admin
equivalents).  Once initialised is called with the relevant lock status block
pointed to report an issued control command.  Retrieve The EX (exclusive) lock
enqueued by the controlling process has been dequeued allowing this AST to be
delivered.  The control command is then in the lock value block.
*/

void ReportInstanceControl (struct lksb *lksbptr)

{
   static char  ClusterDoLockName [32],
                NodeDoLockName [32];
   static $DESCRIPTOR (ClusterDoLockNameDsc, ClusterDoLockName);
   static $DESCRIPTOR (NodeDoLockNameDsc, NodeDoLockName);
   static struct lksb  ClusterDoLockLksb,
                       NodeDoLockLksb;

   int  status,
        LockNameMagic;
   char  *cptr, *sptr, *zptr;

   /*********/
   /* begin */
   /*********/

   if (lksbptr == NULL)
   {
      /**************/
      /* initialise */
      /**************/

      /* a byte comprising two 4 bit fields, version and server group number */
      LockNameMagic = ((HTTPD_LOCK_VERSION & 0xf) << 4) |
                      (InstanceEnvNumber & 0xf);

      /* build the (binary) resource name for the cluster DO lock */
      zptr = (sptr = ClusterDoLockName) + sizeof(ClusterDoLockName)-1;
      for (cptr = HTTPD_NAME; *cptr && sptr < zptr; *sptr++ = *cptr++);
      if (sptr < zptr) *sptr++ = (char)LockNameMagic;
      if (sptr < zptr) *sptr++ = (char)INSTANCE_CLUSTER_DO;
      ClusterDoLockNameDsc.dsc$w_length = sptr - ClusterDoLockName;

      /* enqueue a just-interested NL lock */
      status = sys$enqw (0, LCK$K_NLMODE, &ClusterDoLockLksb,
                         LCK$M_EXPEDITE | LCK$M_SYSTEM,
                         &ClusterDoLockNameDsc, 0, 0, 0, 0, 0, 2, 0);
      if (VMSok (status)) status = ClusterDoLockLksb.lksb$w_status;
      if (VMSnok (status)) EXIT_FI_LI (status);

      /* convert (wait) to CR to block the queue against EX mode */
      status = sys$enqw (0, LCK$K_CRMODE, &ClusterDoLockLksb,
                         LCK$M_VALBLK | LCK$M_QUECVT |
                         LCK$M_CONVERT | LCK$M_SYSTEM,
                         0, 0, 0, &ClusterDoLockLksb, &ReportInstanceControl, 0, 2, 0);
      if (VMSok (status)) status = ClusterDoLockLksb.lksb$w_status;
      if (VMSnok (status)) EXIT_FI_LI (status);

      /* build the (binary) resource name for the node DO lock */
      zptr = (sptr = NodeDoLockName) + sizeof(NodeDoLockName)-1;
      for (cptr = HTTPD_NAME; *cptr && sptr < zptr; *sptr++ = *cptr++);
      if (sptr < zptr) *sptr++ = (char)LockNameMagic;
      for (cptr = SystemData.NodeName; *cptr && sptr < zptr; *sptr++ = *cptr++);
      if (sptr < zptr) *sptr++ = (char)INSTANCE_NODE_DO;
      NodeDoLockNameDsc.dsc$w_length = sptr - NodeDoLockName;

      /* enqueue a just-interested NL lock */
      status = sys$enqw (0, LCK$K_NLMODE, &NodeDoLockLksb,
                         LCK$M_EXPEDITE | LCK$M_SYSTEM,
                         &NodeDoLockNameDsc, 0, 0, 0, 0, 0, 2, 0);
      if (VMSok (status)) status = NodeDoLockLksb.lksb$w_status;
      if (VMSnok (status)) EXIT_FI_LI (status);

      /* convert (wait) to CR to block the queue against EX mode */
      status = sys$enqw (0, LCK$K_CRMODE, &NodeDoLockLksb,
                         LCK$M_VALBLK | LCK$M_QUECVT |
                         LCK$M_CONVERT | LCK$M_SYSTEM,
                         0, 0, 0, &NodeDoLockLksb, &ReportInstanceControl, 0, 2, 0);
      if (VMSok (status)) status = NodeDoLockLksb.lksb$w_status;
      if (VMSnok (status)) EXIT_FI_LI (status);
   }
   else
   {
      /********/
      /* /DO= */
      /********/

      /* convert (wait) current CR mode lock to NL unblocking the queue */
      status = sys$enqw (0, LCK$K_NLMODE, lksbptr,
                         LCK$M_CONVERT | LCK$M_SYSTEM,
                         0, 0, 0, 0, 0, 0, 2, 0);
      if (VMSok (status)) status = lksbptr->lksb$w_status;
      if (VMSnok (status)) EXIT_FI_LI (status);

      /* convert (wait) back to CR to block queue against next EX mode */
      status = sys$enqw (0, LCK$K_CRMODE, lksbptr,
                         LCK$M_VALBLK | LCK$M_QUECVT |
                         LCK$M_CONVERT | LCK$M_SYSTEM,
                         0, 0, 0, lksbptr, &ReportInstanceControl, 0, 2, 0);
      if (VMSok (status)) status = lksbptr->lksb$w_status;
      if (VMSnok (status)) EXIT_FI_LI (status);

      /* the lock value block written by the dequeued EX mode lock */
      cptr = (char*)lksbptr->lksb$b_valblk;
      if (*cptr)
         JsonAlert (NULL, "#/DO=%s%s",
                    cptr, lksbptr == &ClusterDoLockLksb ? "/ALL" : "");
   }
}

/*****************************************************************************/
/*
Collect (server) per-process data.
*/ 

int CollectProcessData
(
ulong ProcessId,
PROCESS_DATA *DataPtr
)
{
/* e.g. U-S-E-K- U+S-E-K- U^S-E-K- */
#define ASTFLG(bit,idx) { if (JpiAstEn & bit) if (JpiAstAct & bit) \
                          pdptr->AstFlags[idx] = '^'; \
                          else pdptr->AstFlags[idx] = '+'; \
                          else pdptr->AstFlags[idx] = '-'; }

   static long  JpiAstAct,
                JpiAstCnt,
                JpiAstEn,
                JpiAstLm,
                JpiBioCnt,
                JpiBioLm,
                JpiBufIo,
                JpiBytCnt,
                JpiCpuId,
                JpiCpuTim,
                JpiDioCnt,
                JpiDioLm,
                JpiDirIo,
                JpiEnqCnt,
                JpiEnqLm,
                JpiExecTim,
                JpiFilCnt,
                JpiFilLm,
                JpiControlFlags = JPI$M_IGNORE_TARGET_STATUS,
                JpiGpgCnt,
                JpiJobPrcCnt,
                JpiJobType,
                JpiKtCount,
                JpiKrnlTim,
                JpiMode,
                JpiMultiThread,
                JpiOwner,
                JpiPageFlts,
                JpiPagFilCnt,
                JpiPgFlQuota,
                JpiPid,
                JpiPpgCnt,
                JpiPrcCnt,
                JpiPrcLm,
                JpiPri,
                JpiPrib,
                JpiState,
                JpiSts,
                JpiSts2,
                JpiSuprTim,
                JpiTqCnt,
                JpiTqLm,
                JpiUserTim,
                JpiVolumes,
                JpiWsAuth,
                JpiWsAuthExt,
                JpiWsExtent,
                JpiWsPeak,
                JpiWsQuota,
                JpiWsSize,
                UaiBytLm;

   static ulong  JpiFrePteCnt [2],
                 JpiLoginTim [2],
                 JpiVirtPeak [2];

   static char  JpiPrcNam [JPI_PRCNAM_MAX+1],
                JpiUserName [12+1];

   static $DESCRIPTOR (UserNameDsc, JpiUserName);

   /* initialize structure for SPI items */
   static struct
   {
      short  BufferLength;
      short  ItemCode;
      void  *BufferPtr;
      void  *LengthPtr;
   } ItemListJpi[] =
   {
       { sizeof(JpiControlFlags), JPI$_GETJPI_CONTROL_FLAGS,
         &JpiControlFlags, 0 },
       { sizeof(JpiAstAct), JPI$_ASTACT, &JpiAstAct, 0 },
       { sizeof(JpiAstCnt), JPI$_ASTCNT, &JpiAstCnt, 0 },
       { sizeof(JpiAstEn),  JPI$_ASTEN,  &JpiAstEn, 0 },
       { sizeof(JpiAstLm),  JPI$_ASTLM,  &JpiAstLm, 0 },
       { sizeof(JpiBioCnt), JPI$_BIOCNT, &JpiBioCnt, 0 },
       { sizeof(JpiBioLm),  JPI$_BIOLM,  &JpiBioLm, 0 },
       { sizeof(JpiBufIo),  JPI$_BUFIO,  &JpiBufIo, 0 },
       { sizeof(JpiBytCnt), JPI$_BYTCNT, &JpiBytCnt, 0 },
       { sizeof(JpiCpuId),  JPI$_CPU_ID,  &JpiCpuId, 0 },
       { sizeof(JpiCpuTim), JPI$_CPUTIM, &JpiCpuTim, 0 },
       { sizeof(JpiDioCnt), JPI$_DIOCNT,  &JpiDioCnt, 0 },
       { sizeof(JpiDioLm),  JPI$_DIOLM,  &JpiDioLm, 0 },
       { sizeof(JpiDirIo),  JPI$_DIRIO,  &JpiDirIo, 0 },
       { sizeof(JpiEnqCnt), JPI$_ENQCNT, &JpiEnqCnt, 0 },
       { sizeof(JpiEnqLm),  JPI$_ENQLM,  &JpiEnqLm, 0 },
       { sizeof(JpiFilCnt), JPI$_FILCNT, &JpiFilCnt, 0 },
       { sizeof(JpiFilLm),  JPI$_FILLM,  &JpiFilLm, 0 },
       { sizeof(JpiFrePteCnt), JPI$_FREPTECNT, &JpiFrePteCnt, 0 },
       { sizeof(JpiGpgCnt),    JPI$_GPGCNT,    &JpiGpgCnt, 0 },
       { sizeof(JpiJobPrcCnt), JPI$_JOBPRCCNT, &JpiJobPrcCnt, 0 },
       { sizeof(JpiJobType),   JPI$_JOBTYPE,   &JpiJobType, 0 },
       { sizeof(JpiKtCount),   JPI$_KT_COUNT,  &JpiKtCount, 0 },
       { sizeof(JpiMultiThread), JPI$_MULTITHREAD, &JpiMultiThread, 0 },
       { sizeof(JpiOwner),     JPI$_OWNER,     &JpiOwner, 0 },
       { sizeof(JpiPageFlts),  JPI$_PAGEFLTS,  &JpiPageFlts, 0 },
       { sizeof(JpiPagFilCnt), JPI$_PAGFILCNT, &JpiPagFilCnt, 0 },
       { sizeof(JpiPgFlQuota), JPI$_PGFLQUOTA, &JpiPgFlQuota, 0 },
       { sizeof(JpiPid),    JPI$_PID,    &JpiPid, 0 },
       { sizeof(JpiPpgCnt), JPI$_PPGCNT, &JpiPpgCnt, 0 },
       { sizeof(JpiPrcCnt), JPI$_PRCCNT, &JpiPrcCnt, 0 },
       { sizeof(JpiPrcLm),  JPI$_PRCLM,  &JpiPrcLm, 0 },
       { sizeof(JpiPri),    JPI$_PRI,    &JpiPri, 0 },
       { sizeof(JpiPrib),   JPI$_PRIB,   &JpiPrib, 0 },
       { sizeof(JpiState),  JPI$_STATE,  &JpiState, 0 },
       { sizeof(JpiTqCnt),  JPI$_TQCNT,  &JpiTqCnt, 0 },
       { sizeof(JpiTqLm),   JPI$_TQLM,   &JpiTqLm, 0 },
       { sizeof(JpiVirtPeak),  JPI$_VIRTPEAK,  &JpiVirtPeak, 0 },
       { sizeof(JpiVolumes),   JPI$_VOLUMES,   &JpiVolumes, 0 },
       { sizeof(JpiWsAuth),    JPI$_WSAUTH,    &JpiWsAuth, 0 },
       { sizeof(JpiWsAuthExt), JPI$_WSAUTHEXT, &JpiWsAuthExt, 0 },
       { sizeof(JpiWsExtent),  JPI$_WSEXTENT,  &JpiWsExtent, 0 },
       { sizeof(JpiWsPeak),    JPI$_WSPEAK,    &JpiWsPeak, 0 },
       { sizeof(JpiWsQuota),   JPI$_WSQUOTA,   &JpiWsQuota, 0 },
       { sizeof(JpiWsSize),    JPI$_WSSIZE,    &JpiWsSize, 0 },
       { sizeof(JpiLoginTim),  JPI$_LOGINTIM,  &JpiLoginTim, 0 },
       { sizeof(JpiPrcNam),    JPI$_PRCNAM,    &JpiPrcNam, 0 },
#ifndef __VAX
       { sizeof(JpiExecTim),   JPI$_EXECTIM,   &JpiExecTim, 0 },
       { sizeof(JpiKrnlTim),   JPI$_KRNLTIM,   &JpiKrnlTim, 0 },
       { sizeof(JpiSuprTim),   JPI$_SUPRTIM,   &JpiSuprTim, 0 },
       { sizeof(JpiUserTim),   JPI$_USERTIM,   &JpiUserTim, 0 },
#endif
       { sizeof(JpiUserName),  JPI$_USERNAME,  &JpiUserName, 0 },
       {0,0,0,0}
   },
   ItemListUai [] =
   {
      { sizeof(UaiBytLm), UAI$_BYTLM, &UaiBytLm, NULL },
      {0,0,0,0}
   };

   int  status;
   char  *cptr, *sptr, *zptr;
   ulong  pid,
          BufIo,
          CpuTime,
          CpuSekTime,
          CpuPercent,
          CpuSekPercent,
          DirIo,
          PageFlts;
   ulong  CpuBinTime[2],
          CpuExecTime[2],
          CpuKrnlTime[2],
          CpuSuprTime[2],
          CpuUserTime[2];
   unsigned short  slen;
   IOSBLK  IOsb;
   PROCESS_DATA  *pdptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CollectProcData() %08.08X\n", ProcessId);

   memset (pdptr = DataPtr, 0, sizeof(PROCESS_DATA));

   status = sys$getjpiw (
            SpiEfn,  /* efn */
            &ProcessId,  /* pidaddr */
            0,  /* prcnam */
            &ItemListJpi,  /* item list */
            &IOsb,  /* iosb */
            0,      /* astaddr */
            0 );    /* astprm */

   if (Debug)
      fprintf (stdout, "sys$getjpi() %%X%08.08X  IOsb: %%X%08.08X\n",
               status, IOsb.iosb$w_status);

   if (VMSok (status)) status = IOsb.iosb$w_status;
   if (VMSnok (status)) return (status);

   if (!UaiBytLm)
   {
      /*
         This is a workaround for establishing initial BYTLM.
         Apparently there is a long-standing issue getting representative
         BYTLM values from $GETJPI and F$GETJPI (and confirmed by alamode). 
         Search the 'net for "F$GETJPI using BYTLM item returns BYTCNT".
      */
      for (cptr = JpiUserName; (cptr-JpiUserName<=12) && *cptr != ' '; cptr++);
      UserNameDsc.dsc$w_length = cptr - JpiUserName;

      status = sys$getuai (
               0,   /* reserved */
               0,   /* context */
               &UserNameDsc,   /* username */
               &ItemListUai,   /* item list */
               &IOsb,  /* iosb */
               0,      /* astaddr */
               0 );    /* astprm */

      if (Debug)
         fprintf (stdout, "sys$getuai() %%X%08.08X  IOsb: %%X%08.08X\n",
                  status, IOsb.iosb$w_status);

      if (VMSok (status)) status = IOsb.iosb$w_status;
      if (VMSnok (status)) return (status);

      /* bit q&d but ... no longer need the JPI$_USERNAME item */
      memset (&ItemListJpi[(sizeof(ItemListJpi)/12)-2],0,12);
   }

   /* terminate and trim trailing spaces */
   JpiPrcNam[sizeof(JpiPrcNam)-1] = '\0';
   for (sptr = JpiPrcNam; *sptr; sptr++);
   *sptr-- = '\0';
   while (sptr > JpiPrcNam && *sptr == ' ') sptr--;
   *++sptr = '\0';

   strcpy (pdptr->JpiPrcNam, JpiPrcNam);

   MINUS_QUAD_QUAD (&CurrentBinTime, &JpiLoginTim, &pdptr->ConBinTime);

   /* make the 10mS tick a delta time */
   PUT_LONG_QUAD (JpiCpuTim, &CpuBinTime);
   MUL_LONG_QUAD (-100000, &CpuBinTime);
   PUT_QUAD_QUAD (&CpuBinTime, &pdptr->CpuBinTime);

#ifndef __VAX
   /* make the 10mS ticks into delta times */
   PUT_QUAD_QUAD (&JpiExecTim, &CpuExecTime);
   MUL_LONG_QUAD (-100000, &CpuExecTime);
   PUT_QUAD_QUAD (&JpiKrnlTim, &CpuKrnlTime);
   MUL_LONG_QUAD (-100000, &CpuKrnlTime);
   PUT_QUAD_QUAD (&JpiSuprTim, &CpuSuprTime);
   MUL_LONG_QUAD (-100000, &CpuSuprTime);
   PUT_QUAD_QUAD (&JpiUserTim, &CpuUserTime);
   MUL_LONG_QUAD (-100000, &CpuUserTime);
#endif

   pdptr->JpiAstCnt = JpiAstCnt;
   pdptr->JpiAstLm = JpiAstLm;
   pdptr->JpiBioCnt = JpiBioCnt;
   pdptr->JpiBioLm = JpiBioLm;
   pdptr->JpiBytCnt = JpiBytCnt;
   pdptr->JpiBytLm = UaiBytLm;  /* NOTE this workaround */
   pdptr->JpiBufIo = JpiBufIo;
   pdptr->JpiCpuId = JpiCpuId;
   pdptr->JpiDioCnt = JpiDioCnt;
   pdptr->JpiDirIo = JpiDirIo;
   pdptr->JpiDioLm = JpiDioLm;
   pdptr->JpiEnqCnt = JpiEnqCnt;
   pdptr->JpiEnqLm = JpiEnqLm;
   pdptr->JpiFilCnt = JpiFilCnt;
   pdptr->JpiFilLm = JpiFilLm;
   pdptr->JpiGpgCnt = JpiGpgCnt;
   pdptr->JpiKtCount = JpiKtCount;
   pdptr->JpiMultiThread = JpiMultiThread;
   pdptr->JpiPagFilCnt = JpiPagFilCnt;
   pdptr->JpiPageFlts = JpiPageFlts;
   pdptr->JpiPgFlQuota = JpiPgFlQuota;
   pdptr->JpiPpgCnt = JpiPpgCnt;
   pdptr->JpiPrcCnt = JpiPrcCnt;
   pdptr->JpiPrcLm = JpiPrcLm;
   pdptr->JpiPri = JpiPri;
   pdptr->JpiPrib = JpiPrib;
   pdptr->JpiTqLm = JpiTqLm;
   pdptr->JpiTqCnt = JpiTqCnt;
   pdptr->JpiWsExtent = JpiWsExtent;
   pdptr->JpiWsPeak = JpiWsPeak;
   pdptr->JpiWsQuota = JpiWsQuota;
   pdptr->JpiVirtPeak = JpiVirtPeak;
   pdptr->JpiVolumes = JpiVolumes;

   /* derive the number of page faults during the period */
   if (pdptr->ProcessId != ProcessId)
   {
      pdptr->ProcessId = ProcessId;
      pdptr->PrevPageFaults = 0;
   }
   if (pdptr->PrevPageFaults)
      pdptr->PageFaults = JpiPageFlts - pdptr->PrevPageFaults;
   pdptr->PrevPageFaults = JpiPageFlts;

   switch (JpiState)
   {
      case  1 : pdptr->ProcessState = "COLPG"; break;
      case  2 : pdptr->ProcessState = "MWAIT"; break;
      case  3 : pdptr->ProcessState = "CEF"; break;
      case  4 : pdptr->ProcessState = "PFW"; break;
      case  5 : pdptr->ProcessState = "LEF"; break;
      case  6 : pdptr->ProcessState = "LEFO"; break;
      case  7 : pdptr->ProcessState = "HIB"; break;
      case  8 : pdptr->ProcessState = "HIBO"; break;
      case  9 : pdptr->ProcessState = "SUSP"; break;
      case 10 : pdptr->ProcessState = "SUSPO"; break;
      case 11 : pdptr->ProcessState = "FPG"; break;
      case 12 : pdptr->ProcessState = "COM"; break;
      case 13 : pdptr->ProcessState = "COMO"; break;
      case 14 : pdptr->ProcessState = "CUR"; break;
      default : pdptr->ProcessState = "?";
   }

   /* AST flags */
   pdptr->AstFlags[0] = 'U'; ASTFLG(1,1)
   pdptr->AstFlags[2] = 'S'; ASTFLG(2,3)
   pdptr->AstFlags[4] = 'E'; ASTFLG(4,5)
   pdptr->AstFlags[6] = 'K'; ASTFLG(8,7)
   pdptr->AstFlags[8] = '\0';

   return (SS$_NORMAL);
}

/*****************************************************************************/
/*
Make a delta binary time into an null-terminated ASCII string, absorbing
non-significant leading characters.  Up to four concurrent times are stored in
static storage.
*/

char* MungeTime (ulong *btptr)

{
#undef BUFFER_MAX
#define BUFFER_MAX 4
   static int  BufferIdx;
   static char  Buffer [BUFFER_MAX][32];
   static $DESCRIPTOR (BufferDsc, Buffer[0]);
   static $DESCRIPTOR (TimeFaoDsc, "!%D\0");

   char  *cptr, *sptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "MungeTime()\n");

   BufferDsc.dsc$a_pointer = Buffer[BufferIdx++ % BUFFER_MAX];

   if (*btptr == 0)
      strcpy (BufferDsc.dsc$a_pointer, "0.00");
   else
   {
      sys$fao (&TimeFaoDsc, 0, &BufferDsc, btptr);
      for (cptr = sptr = BufferDsc.dsc$a_pointer;
           (*cptr == ' ' || *cptr == '0' || *cptr == ':') && cptr - sptr <= 11;
           cptr++);
      while (*cptr && *cptr != '.') *sptr++ = *cptr++;
      *sptr = '\0';
   }

   return (BufferDsc.dsc$a_pointer);
}

/*****************************************************************************/
/*
Convert the 32 bit integer, or 64 bit integer pointed to, into an ASCII number
string.  Uses static storage so only one at a time thank you!
*/

char* JsonNumber
(
int Bits,
ulong Value
)
{
   static char NumBuffer [48];
   static $DESCRIPTOR (NumBufferDsc, NumBuffer);
   static $DESCRIPTOR (Value32FaoDsc, "!UL\0");
   static $DESCRIPTOR (Value64FaoDsc, "!@SQ\0");

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Bits > 32)
#ifndef __VAX 
      status = sys$fao (&Value64FaoDsc, 0, &NumBufferDsc, Value);
#else
   {
      /* a bit of a kludge but apparently ok all the way up to 53 bits */
      double  dValue;
      dValue = (double)((ulong*)Value)[0];
      dValue += (double)((ulong*)Value)[1] * 4294967296.0;
      sprintf (NumBuffer, "%.0f", dValue);
      status = SS$_NORMAL;
   }
#endif
   else
      status = sys$fao (&Value32FaoDsc, 0, &NumBufferDsc, Value);
   if (VMSok (status)) return (NumBuffer);
   return ("NaN");
}

/*****************************************************************************/
/*
Get the HTTPd server data from its global section.
Data is written by UpdateGlobalSection() in [SRC.HTTPD]SUPPORT.C module.
*/

int MapGlobalSection ()

{
   /* system global section, map into first available virtual address */
   static int MapFlags = SEC$M_SYSGBL | SEC$M_EXPREG;
   /* it is recommended to map into any virtual address in the region (P0) */
   static ulong  InAddr [2] = { 0x200, 0x200 };

   static char  HttpdGblSecName [32];
   static $DESCRIPTOR (HttpdGblSecNameDsc, HttpdGblSecName);
   static $DESCRIPTOR (HttpdGblSecNameFaoDsc, GBLSEC_NAME_FAO);

   int  status,
        ByteSize,
        PageSize;
   ushort  ShortLength;
   ulong  RetAddr [2];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "MapGlobalSection()\n");

   if (!HttpdGblSecPtr)
   {
      /* only need to map it the one time */
      sys$fao (&HttpdGblSecNameFaoDsc, &ShortLength, &HttpdGblSecNameDsc,
               HTTPD_NAME, HTTPD_GBLSEC_VERSION_NUMBER,
               InstanceEnvNumber, "HTTPD");
      HttpdGblSecNameDsc.dsc$w_length = ShortLength;
      if (Debug) fprintf (stdout, "|%s|\n", HttpdGblSecName);

      /* map the specified global section */
      status = sys$mgblsc (&InAddr, &RetAddr, 0, MapFlags, &HttpdGblSecNameDsc,
                           0, 0);
      if (Debug)
         fprintf (stdout, "sys$mgblsc() %%X%08.08X begin:%d end:%d\n",
                  status, RetAddr[0], RetAddr[1]);

      if (VMSnok (status)) EXIT_FI_LI(status);

      ByteSize = (RetAddr[1]+1) - RetAddr[0];
      PageSize = (RetAddr[1]+1) - RetAddr[0] >> 9;
      HttpdGblSecPtr = (HTTPD_GBLSEC*)RetAddr[0];
      HttpdGblSecLength = ByteSize;
   }

   /* but we do need to check it's the right one before each data read */
   if (HttpdGblSecPtr->GblSecVersion != HTTPD_GBLSEC_VERSION_NUMBER ||
       HttpdGblSecPtr->GblSecLength != sizeof(HTTPD_GBLSEC))
   {
      fprintf (stdout, "Status: 500\r\n\r\n\
Global section mismatch! Rebuild ALAMODE?\n");
      exit (SS$_NORMAL);
   }

   return (SS$_NORMAL);
}

/*****************************************************************************/
/*
*/

int HavePriv (ulong *PrivMask)

{
   static ulong  JpiImagPriv [2],
                 JpiProcPriv [2];

   static struct {
      short  BufferLength;
      short  ItemCode;
      void  *BufferPtr;
      void  *LengthPtr;
   } JpiItemList[] = 
   {
       { sizeof(JpiImagPriv), JPI$_IMAGPRIV, &JpiImagPriv, 0 },
       { sizeof(JpiProcPriv), JPI$_PROCPRIV, &JpiProcPriv, 0 },
       {0,0,0,0}
   };

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "HavePriv() %08.08X%08.08X\n",
               PrivMask[0], PrivMask[1]);

   status = sys$getjpiw (0, 0, 0, &JpiItemList, 0, 0, 0);

   if (VMSnok (status)) EXIT_FI_LI(status);

   return ((((JpiImagPriv[0] & PrivMask[0]) == PrivMask[0]) &&
            ((JpiImagPriv[1] & PrivMask[1]) == PrivMask[1])) ||
           (((JpiProcPriv[0] & PrivMask[0]) == PrivMask[0]) &&
            ((JpiProcPriv[1] & PrivMask[1]) == PrivMask[1])));
}

/*****************************************************************************/
/*
Translate a logical name using LNM$FILE_DEV.  Returns a pointer to the value
string, or NULL if the name does not exist.  If 'LogValue' is supplied the
logical name is translated into that (assumed to be large enough), otherwise
it's translated into an internal static buffer.  'IndexValue' should be zero
for a 'flat' logical name, or 0..127 for interative translations.
*/

char* SysTrnLnm2
(
char *LogName,
char *LogValue,
int IndexValue
)
{
   static ushort  ValueLength;
   static ulong  LnmAttributes,
                 LnmIndex;
   static char  StaticLogValue [256];
   static $DESCRIPTOR (LogNameDsc, "");
   static $DESCRIPTOR (LnmTableDsc, "LNM$FILE_DEV");
   static struct {
      short int  buf_len;
      short int  item;
      void  *buf_addr;
      ushort  *ret_len;
   } LnmItems [] =
   {
      { sizeof(LnmIndex), LNM$_INDEX, &LnmIndex, 0 },
      { sizeof(LnmAttributes), LNM$_ATTRIBUTES, &LnmAttributes, 0 },
      { 255, LNM$_STRING, 0, &ValueLength },
      { 0,0,0,0 }
   };

   int  status;
   char  *cptr;

   /*********/
   /* begin */
   /*********/

   LnmIndex = IndexValue;

   LogNameDsc.dsc$a_pointer = LogName;
   LogNameDsc.dsc$w_length = strlen(LogName);
   if (LogValue)
      cptr = LnmItems[2].buf_addr = LogValue;
   else
      cptr = LnmItems[2].buf_addr = StaticLogValue;

   status = sys$trnlnm (0, &LnmTableDsc, &LogNameDsc, 0, &LnmItems);
   if (!(status & 1) || !(LnmAttributes & LNM$M_EXISTS))
   {
      if (LogValue) LogValue[0] = '\0';
      return (NULL);
   }

   cptr[ValueLength] = '\0';
   return (cptr);
}

/*****************************************************************************/
/*
Return true or false depending on whether the server is an equal or later
release to the supplied version string.
*/

int MinimumWASD (char *vstring)

{
   int  major[2], minor[2], tweak[2];
   char  *cptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "MinimumWASD() %s\n", vstring);

   major[0] = minor[0] = tweak[0] = major[1] = minor[1] = tweak[1] = 0;

   if (sscanf (vstring, "%d.%d.%d", &major[0], &minor[0], &tweak[0]) < 3)
      return (0);
   if (!(cptr = strstr (WsLibCgiVar("SERVER_SOFTWARE"), "WASD/")))
      return (0);
   if (sscanf (cptr+5, "%d.%d.%d", &major[1], &minor[1], &tweak[1]) < 3)
      return (0);

   if (major[1] > major[0]) return (1);
   if (major[1] == major[0])
   {
      if (minor[1] > minor[0]) return (1);
      if (minor[1] == minor[0] && tweak[1] >= tweak[0]) return (1);
   }

   return (0);
}

/*****************************************************************************/
/*
If WASD and the version >= 10.1 then WebSocket is potentially available.  The
multi-valued logical name ALAMODE_WEBSOCKET allows this to be overridden by
defining the first value to be zero, or one.  Further, additional values can
represent partial strings from the user-agent field.  A match returns the
complement of the first value.  So, 0,"Firefox" would disable WebSocket for all
but user-agents containing the string "Firefox".  If the leading character is
'!' then negates the match, so 0,"!Windows NT","Firefox" would disable
WebSocket for all but user-agents containing the string "Firefox" unless it
also contained the string "Windows NT".  Could be used with the first value
defined as 1, in which case it would function to exclude usage by specified
WebSocket agent strings.  With CGIplus-enabled WebSocket, as it is directly
accessing a CGI variable, can only be used before the request setup has
completed and the asynchronous WebSocket I/O has begun.  Bit of a kludge but
allows the flexibility currently required. 
*/

int WebSockAvailable ()

{
   static long  LnmIndex = -1,
                LnmAttributes;
   static unsigned short  LogValueLen;
   static char  LogValue [255+1];
   static struct {
      short  BufferLength;
      short  ItemCode;
      void  *BufferPtr;
      void  *LengthPtr;
   }
   LnmItems [] =
   {
      { sizeof(LnmIndex), LNM$_INDEX, &LnmIndex, 0 },
      { sizeof(LnmAttributes), LNM$_ATTRIBUTES, &LnmAttributes, 0 },
      { sizeof(LogValue)-1, LNM$_STRING, &LogValue, &LogValueLen },
      { 0,0,0,0 }
   };
   static $DESCRIPTOR (LogNameDsc, "ALAMODE_WEBSOCKET");
   static $DESCRIPTOR (LnmFileDevDsc, "LNM$FILE_DEV");

   int  cnt, status,
        WebSocketOk = 0;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "WebSockAvailable()\n");

   if (MinimumWASD ("10.1.0")) WebSocketOk = 1;

   if (!WebSocketOk) return (WebSocketOk);

   for (LnmIndex = 0; LnmIndex <= 127; LnmIndex++)
   {
      status = sys$trnlnm (0, &LnmFileDevDsc, &LogNameDsc, 0, &LnmItems);
      if (VMSnok (status) || !(LnmAttributes & LNM$M_EXISTS)) break;
      LogValue[LogValueLen] = '\0';
      if (!LnmIndex)
         WebSocketOk = atoi(LogValue);
      else
      if (LogValue[0] == '!') {
         if (strstr (CgiHttpUserAgentPtr, LogValue+1))
            return (WebSocketOk);
      }
      else
      if (strstr (CgiHttpUserAgentPtr, LogValue))
         return (!WebSocketOk);
   }

   return (WebSocketOk);
}

/*****************************************************************************/
/*
A little mitigation without a whole lot of work.
*/

#ifdef __VAX
#if NO_SNPRINTF

int snprintf
(
char *aptr,
__size_t nbyt,
const char *format,
...
)
{
   static int  bcnt;
   static char  *bptr;

   int  cnt;
   char  *cptr, *sptr;
   va_list  argptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "snprint()\n");

   if ((int)nbyt <= 0) return (-1);

   if (nbyt > bcnt)
   {
      bcnt = nbyt;
      if (bptr) free (bptr);
      if ((bptr = calloc (1, bcnt*2)) == NULL) EXIT_FI_LI (vaxc$errno);
   }

   va_start (argptr, format);
   cnt = vsprintf (bptr, format, argptr);
   if (cnt >= bcnt*2) EXIT_FI_LI (SS$_RESULTOVF);
   sptr = aptr;
   for (cptr = bptr; *cptr && --nbyt; *sptr++ = *cptr++);
   *sptr = '\0';
   return (sptr-aptr);
}

#endif /* NO_SNPRINTF */
#endif /* __VAX */

/*****************************************************************************/

